﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

using Zcu.Mve.Core;

namespace Zcu.Mve.Elementals
{
    /// <summary>
    /// A module for executing multiple branches of a map tree on a single request.
    /// </summary>
    [ModuleInfo("Oldřich Petřík", "A module for executing multiple branches of a map tree on a single request.\r\n" +
        "When executed, updates all preceding modules and passes through the output of the selected module.",
        CategoryType.Generic, CategoryData.Generic, IconName = "MultiPortEvaluator.ico")]
    public class MultiPortEvaluator : Module
    {
        private const string PORT_PREFIX = "Input";
        private const string OUT_PORT = "Output";

        private int _N;
        private int _throughPort;
        private int _throughPortBak;

        /// <summary>
        /// A zero-based index of the input port, whose input will be passed through to the output port.
        /// </summary>
        public int PutThroughPortId
        {
            get { return _throughPort; }
            set { _throughPort = value; }
        }

        /// <summary>
        /// Module constructor.
        /// </summary>
        public MultiPortEvaluator()
        {
            _N = 1;
            _throughPort = 0;

            AddInPort(GetPortName(0), typeof(IDataObject), false, true);
            AddOutPort(OUT_PORT, typeof(IDataObject), true);

            PortConnected += new PortConnectedEventHandler(MultiPortEvaluator_PortConnected);
            PortDisconnecting += new PortDisconnectingEventHandler(MultiPortEvaluator_PortDisconnecting);
            SetupDialogClosed += new SetupDialogClosedEventHandler(MultiPortEvaluator_SetupDialogClosed);
        }

        private void MultiPortEvaluator_SetupDialogClosed(object sender, bool cancelled)
        {
            if (cancelled) _throughPort = _throughPortBak;

            if (_throughPort >= _N - 1) _throughPort = _N - 2;

            OutPort op = GetOutPort(OUT_PORT);
            InPort ip = GetInPort(GetPortName(_throughPort));
            if (op.PortType != ip.PortType) op.PortType = ip.PortType;
        }

        private void MultiPortEvaluator_PortDisconnecting(object sender, OutPort portFrom, InPort portTo, out bool allowed)
        {
            allowed = (portTo.Name.Equals(GetPortName(_N - 2)));
            if (allowed)
            {
                portTo.PortType = typeof(IDataObject);
                if (_throughPort == _N - 2)
                {
                    _throughPort = Math.Max(0, _throughPort - 1);
                    GetOutPort(OUT_PORT).PortType = GetInPort(GetPortName(_throughPort)).PortType;
                }
                RemoveInPort(GetPortName(_N - 1));
                _N--;
            }
        }

        private void MultiPortEvaluator_PortConnected(object sender, OutPort portFrom, InPort portTo)
        {
            portTo.PortType = portFrom.PortType;
            if (_N > _throughPort && portTo == GetInPort(GetPortName(_throughPort)))
            {
                GetOutPort(OUT_PORT).PortType = portFrom.PortType;
            }
            AddInPort(GetPortName(_N++), typeof(IDataObject), false, true);
        }

        private string GetPortName(int id)
        {
            return PORT_PREFIX + (id + 1);
        }

        /// <summary>
        /// Returns the setup dialog for the module.
        /// </summary>
        /// <returns>setup dialog</returns>
        public override ModuleSetup InvokeSetup()
        {
            _throughPortBak = _throughPort;
            return base.InvokeSetup();
        }

        /// <summary>
        /// Updates all input ports and sends through the data from the selected port.
        /// </summary>
        public override void Execute()
        {
            if (GetInPort(GetPortName(0)).Link == null) return;

            SetOutput(OUT_PORT, GetInput(GetPortName(_throughPort)), IsDataSame(GetPortName(_throughPort)));
        }
    }
}
