using System;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Examples
{
	/// <summary>
	/// Module performs a convolution of a image.
	/// </summary>

    [ModuleInfo("Milan Frank", "Module performs a convolution of a image.", CategoryType.Filter, CategoryData.Image)]

  [PortInfo("Mask", "Input convolution mask.")]
  [PortInfo("Image", "An image, with which a convolution going to be performed.")]
  [PortInfo("Result", "Resulting image.")]
  public class Convolution : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Creates ports.
		/// </summary>
    public Convolution()
		{
      AddInPort("Mask", typeof (UniformDataArray2D));
      AddInPort("Image", typeof (RegGrid2D));
      AddOutPort("Result", typeof (RegGrid2D));
    }
	
    /// <summary>
    /// Method executes the convolution.
    /// </summary>
    public override void Execute()
    {
      UniformDataArray2D mask = (UniformDataArray2D) GetInput("Mask");
      RegGrid2D image = (RegGrid2D) GetInput("Image");
      RegGrid2D result = new RegGrid2D(image.Width, image.Height);

      UniformDataArray2D imageBmp = image.GetCellAttrs(typeof(ColorRGBA).Name);
      UniformDataArray2D resultBmp = new UniformDataArray2D(typeof(ColorRGBA), image.Width, image.Height);
      
      result.AddCellAttrs(typeof(ColorRGBA).Name, resultBmp);

      if (mask.Width != mask.Height || mask.Width % 2 != 1)
      {
        throw new MveException(this, "Incorrect mask format.");
      }

      int border = mask.Width / 2;
      int i, j;
      ColorRGBA imgColor;
      float maskVal;
      float r, g, b;

      for (int img_i = 0; img_i < image.Width; img_i++)
      {
        for (int img_j = 0; img_j < image.Height; img_j++)
        {
          r = g = b = 0.0f;

          for (int mask_i = 0; mask_i < mask.Width; mask_i++)
          {
            for (int mask_j = 0; mask_j < mask.Height; mask_j++)
            {
              i = img_i + mask_i - border;
              j = img_j + mask_j - border;
              
              if (i < 0 || i >= image.Width)
                continue;
              if (j < 0 || j >= image.Height)
                continue;

              imgColor = (ColorRGBA) imageBmp[i, j];
              maskVal = (Scalar) mask[mask_i, mask_j];

              r += imgColor.RedF * maskVal;
              g += imgColor.GreenF * maskVal;
              b += imgColor.BlueF * maskVal;
            }
          }

          resultBmp[img_i, img_j] = new ColorRGBA(r, g, b);
        }
      }

      SetOutput("Result", result);
    }
  }
}
