﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.ComponentModel;
using System.Windows.Forms;

using Zcu.Mve.Core;
using Zcu.Mve.Strings;

namespace Zcu.Mve.Strings
{
    /// <summary>
    /// Iterates through all items (files, directories, both) matching a specified mask in a selected directory.
    /// Sets the output to the current item name and waits for an input before proceeding to a next item.
    /// </summary>
    [ModuleInfo("Oldřich Petřík", "In a single map run, iterates through all items (files, directories, both) matching a specified mask in a selected directory.", CategoryType.Source, CategoryData.String)]
    [PortInfo("InUpdate", "Any IDataObject which causes the map to update.")]
    [PortInfo("InDirectory", "[optional] Base directory path.")]
    [PortInfo("OutItemName", "Name of the current file or directory.")]
    public class DirectoryIterator : Module
    {
        /// <summary>DirectoryIterator working modes</summary>
        public enum DirectoryIteratorModes
        {
            /// <summary>Select only files.</summary>
            Files,
            /// <summary>Select only directories.</summary>
            Directories,
            /// <summary>Select both files and directories.</summary>
            Both
        };

        private string _dir = ".";
        private string _mask = "*";
        private DirectoryIteratorModes _mode = DirectoryIteratorModes.Files;
        private InPort _updatePort, _dirPort;

        /// <summary>Full path to the base directory.</summary>
        [Editor(typeof(System.Windows.Forms.Design.FolderNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
        public string DirectoryPath { get { return _dir; } set { _dir = value; } }

        /// <summary>File/directory name mask</summary>
        public string NameMask { get { return _mask; } set { _mask = value; } }

        /// <summary>Item type specification</summary>
        public DirectoryIteratorModes Mode { get { return _mode; } set { _mode = value; } }


        /// <summary>
        /// Module constructor.
        /// </summary>
        public DirectoryIterator()
        {
            AddInPort("InUpdate", typeof(Zcu.Mve.Core.IDataObject), true, false);
            AddInPort("InDirectory", typeof(MveString), false, true);
            AddOutPort("OutItemName", typeof(MveString), false);

            _updatePort = GetInPort("InUpdate");
            _dirPort = GetInPort("InDirectory");

            SetupDialogClosed += new SetupDialogClosedEventHandler(DirectoryIterator_SetupDialogClosed);
        }


        private void DirectoryIterator_SetupDialogClosed(object sender, bool cancelled)
        {
            if (cancelled) return;

            if (!Directory.Exists(_dir))
            {
                MessageBox.Show(
                    "Directory " + _dir + " does not exist. Please specify an existing directory.",
                    "Directory Path Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                cancelled = true;
            }
        }


        /// <inheritdoc />
        public override void Execute()
        {
            if (_dirPort.Link != null) _dir = (string)(MveString)_dirPort.GetInput();

            DirectoryInfo di = new DirectoryInfo(_dir);
            if (!di.Exists) throw new MveException(this, "Directory " + _dir + " does not exist. Please specify an existing directory.");

            Array children = null;
            switch (_mode)
            {
                case DirectoryIteratorModes.Files: children = di.GetFiles(_mask); break;
                case DirectoryIteratorModes.Directories: children = di.GetDirectories(_mask); break;
                case DirectoryIteratorModes.Both: children = di.GetFileSystemInfos(_mask); break;
            }

            foreach (FileSystemInfo fsi in children)
            {
                SetOutput("OutItemName", (MveString)fsi.FullName, false);
                WriteLog("Item Name", fsi.FullName);

                UpdateInput(_updatePort);

                Globals.experimentLog.NewExperiment();
            }
        }
    }
}
