using System;
using Zcu.Mve.Core;
using System.Diagnostics;

namespace Zcu.Mve.Strings
{
	/// <summary>
	/// An MVE representation of a string. Encapsulates a single string and implements the IDataObject interface.
	/// </summary>
	[DataObject("An MVE representation of a string. Encapsulates a single string and implements the IDataObject interface.")]
	public class MveString:IDataObject
	{
		/// <summary>
		/// Value encapsualted by the class
		/// </summary>
		public string Value;


        /// <summary>
        /// Only used by XML serializer, don't call explicitely.
        /// </summary>
        public MveString()
        {
        }

		/// <summary>
		/// constructs a new instance of MveString that encapsulates a string passed as argument.
		/// </summary>
		/// <param name="val">String to be encapsulated by the MveString instance.</param>
		public MveString(string val)
		{
			this.Value = val;	
		}

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			xmlTextReader.Read();
			this.Value = xmlTextReader.Value;
            xmlTextReader.Read();
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file. CDATA is used for the case of escape characters in the string
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			if (this.Value.IndexOf("]]>")!=-1)
			{
				throw new MveException(this, "String that contains ']]>' as a substring cannot be serialized to XML");
			}
			else
			{
				xmlTextWriter.WriteCData(this.Value);
			}
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			return new MveString(this.Value);
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			if (Value == null) Debug.WriteLine("MveString containing null string encountered");
			return true;
		} // CheckConsistence()

		// ===== Methods of the class Object ==============================================
		#region Overriden methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			return "MveString: " + this.Value;
		} // ToString()

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if ((a is MveString) && (b is MveString))
			{
				return(((MveString)a).Value == ((MveString)b).Value);
			}
			return false;
		}

		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return MveString.Equals(this, obj);
		}

		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			return this.Value.GetHashCode();
		}

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		#endregion

		// ===== Conversions ==============================================================
		#region Conversions

		// ===== MveString <-> string ================
		/// <summary>
		/// Implicit conversion MveString -> string.
		/// </summary>
		/// <param name="s">MveString.</param>
		/// <returns>string.</returns>
		public static implicit operator string(MveString s) 
		{
			return s.Value;
		} // MveString -> string

		/// <summary>
		/// Implicit conversion string -> MveString
		/// </summary>
		/// <param name="s">string.</param>
		/// <returns>MveString.</returns>
		public static implicit operator MveString(string s)
		{
			return new MveString(s);
		} // string -> MveString

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(MveString a, MveString b)
		{
			return MveString.Equals(a, b);
		}

		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(MveString a, MveString b) 
		{
			return !(a == b);
		}

		#endregion
	}
}
