using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Loaders
{

    /// <summary>
    /// Library for validating and parsing module settings.
    /// </summary>
    public class VL_InputValidatorAndParser
    {

        #region private variables

        /// <summary>
        /// List of errors found while validating the module settings.
        /// </summary>
        private List<string> errors;

        /// <summary>
        /// Size of the loaded file;
        /// </summary>
        private long fileSize;

        /// <summary>
        /// Generated data description. - output
        /// </summary>
        private BL_DataDescription dataDescription;

        /// <summary>
        /// Settings of the module. - input
        /// </summary>
        private BL_ModuleSettings settings;

        /// <summary>
        /// Names of the sample attributes.
        /// </summary>
        private string[] attributeNames;

        /// <summary>
        /// Names of the sample according attribute datatypes.
        /// </summary>
        private string[] attributeDataTypeNames;

        /// <summary>
        /// Dimension of the data sample attributes.
        /// </summary>
        private int[] attributesDimension;

        /// <summary>
        /// Currently used number format (obtaind automatically form the machine settings).
        /// </summary>
        private NumberFormatInfo numberFormat;

        /// <summary>
        /// Numbers of samples per axis.
        /// </summary>
        private int[] sizes;

        /// <summary>
        /// Minimal indeces of the data subset.
        /// </summary>
        private int[] minIndeces;

        /// <summary>
        /// Maximal indeces of the data subset.
        /// </summary>
        private int[] maxIndeces;

        /// <summary>
        /// Minimal sample coordinates per each axis.
        /// </summary>
        private double[] mins;

        /// <summary>
        /// Maximal sample coordinates per each axis.
        /// </summary>
        private double[] maxs;

        /// <summary>
        /// String describing the data attributes.
        /// </summary>
        private string dataDesc;

        /// <summary>
        /// Dimension of the data grid.
        /// </summary>
        private int dataDimension;

        /// <summary>
        /// Wether to invert endian of each type or not.
        /// </summary>
        private bool[] invertEndian;

        #endregion


        ////////////////////
        // public METHODS //
        ////////////////////
        
        /// <summary>
        /// Validates and parses module settings.
        /// If any setting is not valid an MveException is thrown after the validation process.
        /// Actual NumberFormat is acquired on each call.
        /// </summary>
        /// <param name="settings">Actual settings of the module.</param>
        public BL_DataDescription ValidateAndParseInput(BL_ModuleSettings settings)
        {

            // Retrieve the actual number format.
            numberFormat = System.Threading.Thread.CurrentThread.CurrentCulture.NumberFormat;

            // Set the settings.
            this.settings = settings;

            // Clera all variables that are changed during the validation proccess.
            ClearVariables();

            // Validate the file readability.
            ValideteFileReadability();

            // Validate the data ranges.
            ValidateDataRanges();

            // Validat the data description.
            ValidateAndParseDataDesc();

            SetEndian();
            


            // If no errors occured so far, create an according instace of DataDesctiption and
            // check if size of the specified data is same as size of the loaded file.
            if (errors.Count == 0)
            {
                dataDescription = new BL_DataDescription(settings.DataFilePath, dataDimension, settings.HeaderSize,
                    sizes, mins, maxs, minIndeces, maxIndeces, attributeNames, attributeDataTypeNames,
                    attributesDimension, invertEndian,
                    settings.PointOrCells);

                ValidateFileSize();
            }
            
            
            // If any error has occured during the validation process, report all errors in an MveException.
            if (errors.Count != 0)
            {
                // String representing all errors.
                string errorString = "";

                // Add each error.
                for (int i = 0; i < errors.Count; i++)
                {
                    errorString += errors[i];
                    errorString += " ";
                }

                throw new MveException(this, errorString);
            }
            else
            {
                // return the resulting data description
                return dataDescription;
            }
        }




        //////////////////////////
        // MAIN private METHODS //
        //////////////////////////


        /// <summary>
        /// Clears all variables before validating process.
        /// </summary>
        private void ClearVariables()
        {
            // data description
            dataDescription = null;

            // data range
            mins = null;
            maxs = null;
            sizes = null;

            // data subset
            minIndeces = null;
            maxIndeces = null;

            // filesize
            fileSize = -1;

            // data dimension
            dataDimension = -1;

            // attributes
            attributeNames = null;
            attributesDimension = null;
            attributeDataTypeNames = null;

            // Clear or create a new list of found errors.
            if (errors == null)
            {
                errors = new List<string>();
            }
            else
            {
                errors.Clear();
            }

        }

        /// <summary>
        /// Checks if specified file is readable and contains enough data.
        /// </summary>
        private void ValideteFileReadability()
        {
           // Check filepath for null.
            if (settings.DataFilePath == null)
            {
                errors.Add("No file path sppecified.");
                return;
            }


            // Check file existance. 
            if (File.Exists(settings.DataFilePath) == false)
            {
                errors.Add("Specified file does not exist.");
                return;
            }


            // Try to open a filestream on the specified file and check its capabilities.
            // If somethig goes wrong while opening the file filesize will be set on -1B.
            try
            {

                fileSize = -1;


                // Open the file stream.
                FileStream fs = new FileStream(settings.DataFilePath, FileMode.Open);


                // save size of the file in bytes
                fileSize = fs.Length;


                // Check if the stream can bee readed.
                if (fs.CanRead == false)
                {
                    errors.Add("Unable to read the specified file.");
                }

                // Check if we can seek in the stream.
                if (fs.CanSeek == false)
                {
                    errors.Add("Unable to seek in the specified file.");
                }

                // Close the stream.
                fs.Close();

            }
            catch (IOException)
            {
                errors.Add("Unable to open specified file.");
            }
            catch (UnauthorizedAccessException)
            {
                errors.Add("You do not have access to the file.");
            }
        }

        /// <summary>
        /// Checks if data ranges are valid. If any cruicial step fails, the method ends after it.
        /// </summary>
        private void ValidateDataRanges()
        {

            // Check if grid dimension, data sizesS, minsS and maxsS are filled in.           
            if (CheckMinsMaxesSizesNotEmpty() == false)
            {
                return;
            }

            // Parse the data sizesS, minsS, maxsS.
            if (ParseMinsMaxsSizes() == false)
            {
                return;
            }

            // Check if the specified data dimension is allowed and if mins, max and dataSizes dimension is of the same value.
            if (CheckDataDimension() == false)
            {
                return;
            }


            // Check if all minsS < maxs and all dataSizes > 0;
            if (CheckMinsLoverMaxsAndSizesGraterZero() == false)
            {
                return;
            }


            // Check data subset settings are filled.
            if (CheckDataSubsetFilled() == true)
            {
                // If OK, try to parse the settings strings.
                if (ParseDataSubset() == true)
                {
                    // If OK, check the subset settings dimension.
                    if (CheckDataSubsetDimension() == true)
                    {
                        // If dimension OK, check their range.
                        CheckDataSubsetRange();
                    }
                    else
                    {
                        return;
                    }
                }
            }
        }

        /// <summary>
        /// Checks if the data description string is valid and creates
        /// according DataDescription instance as an output.
        /// </summary>
        private void ValidateAndParseDataDesc()
        {
            // Chcek if there is any data description string.
            if (settings.DataDesc == null || settings.DataDesc.CompareTo("") == 0)
            {
                errors.Add("No data description.");
                return;
            }
            else
            {
                dataDesc = settings.DataDesc;
            }



            // Split the data description string into attributes and create new instance of DataDecription (output).
            // "<attribute>;<attribute>; ... <attribute>" -> array of <atrribute>
            string[] attributes = settings.DataDesc.Trim().Split(';');

            // Create according arrays for attribute names and dimension.
            attributeNames = new string[attributes.Length];
            attributesDimension = new int[attributes.Length];
            attributeDataTypeNames = new string[attributes.Length];
            if (settings.InvertEndian == null) invertEndian = new bool[attributes.Length];


            // One attribute consisting of attribute name and type specification.
            string[] oneAttribute;

            // Parts of atrribute type specification - type name and dimension.
            string[] typeParts;
            
            // If the parsed attribute is really OK.
            bool attReallyOk;




            // Check each atrribute if its a correct atrribute.
            for (int i = 0; i < attributes.Length; i++)
            {
                
                // Split the current atrribute.
                // "<attribute>" -> "<name>", "<type>"
                oneAttribute = attributes[i].Split(',');


                // If the attribute string was not split into two parts, 
                // report an error and continue with the next one,
                // else set the attribute name.
                if (oneAttribute.Length != 2 && oneAttribute.Length != 3)
                {
                    errors.Add(attributes[i] + " is not a valid attribute setting.");
                    continue;
                }
                else
                {
                    attributeNames[i] = oneAttribute[0];
                    if (oneAttribute.Length == 3)
                    {
                        if (oneAttribute[2].CompareTo("inv") == 0)
                        {
                            invertEndian[i] = true;
                        }
                        else
                        {
                            errors.Add(oneAttribute[0] + "'s invert setting is not valid");
                        }
                    }
                    else
                    {
                        if (invertEndian != null)
                        {
                            invertEndian[i] = false;
                        }
                    }
                }


                // Check if the <type> part of attribute is correct.
                if (BinaryLoaderGlobalConstants.dataAttributeRegex.IsMatch(oneAttribute[1]) == true)
                {

                    // Extract the type name into typeParts[0].
                    // example: int32[2] -> int32, 2];
                    typeParts = oneAttribute[1].Split('[');

                    // We still do not know if the name is a known one.
                    attReallyOk = false;


                    // We compare the name with each of the known ones.
                    for (int j = 0; j < BinaryLoaderGlobalConstants.dataTypeNames.Length; j++)
                    {
                        // If the name is known set it ok, set according output and stop checking.
                        if (typeParts[0].CompareTo(BinaryLoaderGlobalConstants.dataTypeNames[j]) == 0)
                        {
                            attributeDataTypeNames[i] = typeParts[0];
                            attReallyOk = true;
                            break;
                        }
                    }



                    // If the name is not a known one, report an error.
                    if (attReallyOk == false)
                    {
                        errors.Add("Unknown data type: " + typeParts[0]);
                    }


                    // Extract the dimension of the attribute and check it on <= 0 and set the dimension at output.
                    // example: if the attribute is "int32[2]", then typeParts[1] = "2]"
                    typeParts = typeParts[1].Split(']');
                    if (int.Parse(typeParts[0]) <= 0)
                    {
                        errors.Add("Attribute dimension can not be <= 0: " + attributes[i]);
                    }
                    else
                    {
                        attributesDimension[i] = int.Parse(typeParts[0]);
                    }

                } // End of <type> parsing.
                
                // If the <type> part of attribute is incorrect.
                else
                {
                    errors.Add("Improper data type syntax: " + oneAttribute[1]);
                }

            } // End of cycle for each attribute.
        }

        /// <summary>
        /// Checks if the calculated data size is same as size of the loaded file.
        /// If not, an error will be added to the error list.
        /// </summary>
        private void ValidateFileSize()
        {
            try
            {
                // Open the file.
                FileStream fs = new FileStream(settings.DataFilePath, FileMode.Open);
                
                // Set the size of the file.
                long fileLength = fs.Length;

                // Close the file stream.
                fs.Close();

                // Set the size of the data as size of one data sample.
                long dataLength = dataDescription.SampleSizeInBytes;

                // Multiply the data size by number of data samples in each dimension.
                for (int i = 0; i < settings.DataDim; i++)
                {
                    dataLength = dataLength * sizes[i];
                }

                // Add the header size.
                dataLength = dataLength + settings.HeaderSize;

                // Check if the two dataSizes are equal.
                if (dataLength != fileLength)
                {
                    errors.Add("Non-equal data and file dataSizes. file: " + fileLength + "b data: " + dataLength);
                }
            }

            // If any IOException was thrown, propagate it as an MveException.
            catch (IOException e)
            {
                throw new MveException(this,"An error occured opening the specified file.", e);
            }
        }

        /// <summary>
        /// Sets the change of endian depending on input.
        /// </summary>
        private void SetEndian()
        {
            if (settings.InvertEndian != null)
            {
                invertEndian = settings.InvertEndian;
            }
        }





        //////////////////////////////
        // HELPFULL private METHODS //
        //////////////////////////////


        /// <summary>
        /// Checks if all the strings are filled in.
        /// If any string is blank or null, an error is added to the error list.
        /// </summary>
        /// <returns>TRUE if no error occured, othewise FALSE</returns>
        private bool CheckMinsMaxesSizesNotEmpty()
        {
            // Presum that the check will be ok.
            bool checkOK = true;

            // Check dimension.
            if (settings.DataDim == 0)
            {
                errors.Add("Grid dimension is not filled in.");
                checkOK = false;
            }

            // Check minsS.
            if (settings.Mins == null || settings.Mins.CompareTo("") == 0)
            {
                errors.Add("Minimal values are not filled in.");
                checkOK = false;
            }

            // Check maxs.
            if (settings.Maxs == null || settings.Maxs.CompareTo("") == 0)
            {
                errors.Add("Maximal values are not filled in.");
                checkOK = false;
            }

            // Check sizesS.
            if (settings.Sizes == null || settings.Sizes.CompareTo("") == 0)
            {
                errors.Add("NumberOfSamples, minsS or maxsS not filled in.");
                checkOK = false;
            }

            // Return the checks result
            return checkOK;
        }


        /// <summary>
        /// Parses the input strings for mins, maxs and dataSizes.
        /// If any value can not be parsed, an error is added to the error list.
        /// </summary>
        /// <returns>TRUE if no error occured, othewise FALSE</returns>
        private bool ParseMinsMaxsSizes()
        {

            // substring represnting each value of according array
            string[] minsS, maxsS, sizesS;

            // Split mins and create according double array.
            minsS = settings.Mins.Split(';');
            this.mins = new double[minsS.Length];

            // Split maxs and create according double array.
            maxsS = settings.Maxs.Split(';');
            this.maxs = new double[maxsS.Length];

            // Split dataSizes and create according int array.
            sizesS = settings.Sizes.Split(';');
            this.sizes = new int[sizesS.Length];

            // Presume that the parsing will not generate any error.
            bool parsingOK = true;



            // Parse all mins.
            for (int i = 0; i < minsS.Length; i++)
            {
                try
                {
                    this.mins[i] = double.Parse(minsS[i], numberFormat);
                }
                catch (FormatException)
                {
                    errors.Add(minsS[i] + " is not a double value");
                    parsingOK = false;
                }
            }


            // Parse all maxs.
            for (int i = 0; i < maxsS.Length; i++)
            {
                try
                {
                    this.maxs[i] = double.Parse(maxsS[i], numberFormat);
                }
                catch (FormatException)
                {
                    errors.Add(maxsS[i] + " is not a double value");
                    parsingOK = false;
                }
            }

            // Parse all dataSizes.
            for (int i = 0; i < sizesS.Length; i++)
            {
                try
                {
                    this.sizes[i] = int.Parse(sizesS[i], numberFormat);
                }
                catch (FormatException)
                {
                    errors.Add(maxsS[i] + " is not an integer value");
                    parsingOK = false;
                }
            }

            // return the result
            return parsingOK;
        }


        /// <summary>
        /// Checks if the specified data dimension is allowed.
        /// If any check fails, an error is added to the error list.
        /// </summary>
        /// <returns>TRUE if the dimension is allowed, otherwise FALSE</returns>
        /// <see cref="BinaryLoaderGlobalConstants.AllowedDimension"/>
        private bool CheckDataDimension()
        {
            // Presume that the dimension specification is wrong.
            bool dataDimensionOK = false;

            dataDimension = settings.DataDim;

            // Check with all allowed dimension values.
            for (int i = 0; i < BinaryLoaderGlobalConstants.allowedDimension.Length; i++)
            {

                // If the current value is allowed, set the result as ok and end the cycle.
                if (dataDimension == BinaryLoaderGlobalConstants.allowedDimension[i])
                {
                    dataDimensionOK = true;
                    break;
                }
            }

            // If data dimension is OK, check mins, maxs and dataSizes if they are the right dimension.
            if (dataDimensionOK == true)
            {
                // Check mins.
                if (mins.Length != dataDimension)
                {
                    errors.Add("MinimalValues dimension is not same as data dimension.");
                    dataDimensionOK = false;
                }

                // Check maxs.
                if (maxs.Length != dataDimension)
                {
                    errors.Add("MaximalValues dimension is not same as data dimension.");
                    dataDimensionOK = false;
                }

                // Check dataSizes.
                if (sizes.Length != dataDimension)
                {
                    errors.Add("NumberOfSamples dimension is not same as data dimension.");
                    dataDimensionOK = false;
                }
            }
            else
            {
                errors.Add(dataDimension.ToString() + " is not an allowed data dimension (allowed values are 2,3,4)");
            }

            // Return the result.
            return dataDimensionOK;
        }

        
        /// <summary>
        /// Checks if all mins are less than maxs and if all dataSizes are greater than 0.
        /// If any check fails, an error is added to the error list.
        /// </summary>
        /// <returns>TRUE if mins less maxs and dataSizes greater 0, otherwise FALSE</returns>
        private bool CheckMinsLoverMaxsAndSizesGraterZero()
        {
            // Presume that the check will be ok.
            bool checkOK = true;

            // For all mins, maxs and dataSizes.
            for (int i = 0; i < mins.Length; i++)
            {
                // Check mins < maxs.
                if (mins[i] >= maxs[i])
                {
                    errors.Add("mins[" + i + "] >= maxs[" + i + "]");
                    checkOK = false;
                }

                // Check dataSizes > 0.
                if (sizes[i] <= 0)
                {
                    errors.Add("dataSizes[" + i + "] <= 0");
                    checkOK = false;
                }
            }

            // Return the result.
            return checkOK;
        }


        /// <summary>
        /// Checks if each field of minIndeces, maxIndeces and dataSubsamplingSteps is filled in or
        /// each is not filled in. 
        /// </summary>
        /// <returns>TRUE if each field is filled in,
        /// FALSE if not each is filled in.</returns>
        private bool CheckDataSubsetFilled()
        {
            
            bool filled;

            // If each field of minIndeces, maxIndeces and dataSubsamplingSteps is blank, 
            // set according arryas as null.
            if ((settings.MinIndeces == null || settings.MinIndeces.CompareTo("") == 0) && 
                (settings.MaxIndeces == null || settings.MaxIndeces.CompareTo("") == 0 ))
            {
                minIndeces = null;
                maxIndeces = null;

                filled = false;
            }

            // Else if all the fileds are filled in.
            else if (settings.MinIndeces != null && settings.MinIndeces.CompareTo("") != 0 &&
                settings.MaxIndeces != null && settings.MaxIndeces.CompareTo("") != 0)
            {
                filled = true;
            }

            // Else if only some subset settings are filled in.
            else
            {
                errors.Add("Fill in all data subset settings or none.");
                filled = false;
            }

            return filled;
        }

    
        /// <summary>
        /// Parses strings for dataSubset selection.
        /// If any parsing fails, an error is added to the error list.
        /// </summary>
        /// <returns>TRUE if no error ocures, otherwise FALSE</returns>
        private bool ParseDataSubset()
        {
            // substring representing each falue of according arrays
            string[] minIndecesS, maxIndecesS;

            // Split minIndecesS and create according int array.
            minIndecesS = settings.MinIndeces.Split(';');
            minIndeces = new int[minIndecesS.Length];

            // Split maxIndecesS and create according int array.
            maxIndecesS = settings.MaxIndeces.Split(';');
            maxIndeces = new int[maxIndecesS.Length];

           

            // Presume that parsing will be ok;
            bool parsingOK = true;

            // Parse minIndeces
            for (int i = 0; i < minIndeces.Length; i++)
            {
                try
                {
                    minIndeces[i] = int.Parse(minIndecesS[i], numberFormat);
                }
                catch (FormatException)
                {
                    errors.Add(minIndecesS[i] + " is not a valid integer value.");
                    parsingOK = false;
                }
            }


            // Parse maxIndeces
            for (int i = 0; i < maxIndeces.Length; i++)
            {
                try
                {
                    maxIndeces[i] = int.Parse(maxIndecesS[i], numberFormat);
                }
                catch (FormatException)
                {
                    errors.Add(maxIndecesS[i] + " is not a valid integer value.");
                    parsingOK = false;
                }
            }

            // Return the result.
            return parsingOK;
        }


        /// <summary>
        /// Checks if the dataSubset dimension is all right.
        /// </summary>
        /// <returns>TRUE if is allright, otherwise FALSE</returns>
        private bool CheckDataSubsetDimension()
        {
            // Presume, that the check will be OK.
            bool checkOK = true;

            // Check minIndeces.
            if (minIndeces.Length != dataDimension)
            {
                errors.Add("MinIndeces are of wrong dimension.");
                checkOK = false;
            }

            
            // Check maxIndeces.
            if (maxIndeces.Length != dataDimension)
            {
                errors.Add("MaxIndeces are of wrong dimension.");
                checkOK = false;
            }
           
            // Return the result.
            return checkOK;
        }

        /// <summary>
        /// Checks if the dataSubset selection is not out of data range and minIndeces not > maxIndeces.
        /// </summary>
        /// <returns>TRUE if the subSet is valid, otherwise FALSE</returns>
        private bool CheckDataSubsetRange()
        {
            // Presume that the check will be OK.
            bool checkOK = true;

            // Check all minIndeces.
            for (int i = 0; i < minIndeces.Length; i++)
            {
                // Chceck if they are > 0;
                if (minIndeces[i] < 0)
                {
                    errors.Add("Minimal index " + minIndeces[i].ToString() + " is <= 0.");
                    checkOK = false;
                }

                // Chceck if they are < dataSizes.
                if (minIndeces[i] >= sizes[i])
                {
                    errors.Add("Minimal index " + minIndeces[i].ToString() + "is >= maximal allowed index.");
                    checkOK = false;
                }

                // Check if they are <= maxIndeces.
                if (minIndeces[i] > maxIndeces[i]) 
                {
                    errors.Add("Minimal index " + minIndeces[i].ToString() + " is > than maximal index.");
                    checkOK = false;
                }
            }


            // Check all maxIndeces.
            for (int i = 0; i < maxIndeces.Length; i++)
            {
                // Check if they are > 0.
                if (maxIndeces[i] < 0) 
                {
                    errors.Add("Maximal index " + maxIndeces[i].ToString() + " is < 0");
                    checkOK = false;
                }

                // Check if they are < dataSizes.
                if (maxIndeces[i] >= sizes[i])
                {
                    errors.Add("Maximal index " + maxIndeces[i].ToString() + " is >= maximal allowe index.");
                    checkOK = false;
                }
            }
   
            // Return the result.
            return checkOK;
        }
    }
}