using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Loaders
{

    /// <summary>
    /// Class for reading an attribute from a file to a MveArray.
    /// </summary>
    public class BL_TypeReader
    {

        #region properties


        private string attrName;
        /// <summary>
        /// Name of the data attribute readed by the instnce of the reader.
        /// </summary>
        public string AttrName
        {
            get { return attrName; }
        }



        private string attrType;
        /// <summary>
        /// String that specifies the attributes data type.
        /// </summary>
        public string AttrType
        {
            get { return attrType; }
        }



        private int attrTypeDimension;
        /// <summary>
        /// Dimension of the atrtibute data.
        /// </summary>
        public int AttrTypeDimension
        {
            get { return attrTypeDimension; }
        }



        private BinaryReader streamReaderToReadFrom;
        /// <summary>
        /// Stream binary reader used to read the data.
        /// </summary>
        public BinaryReader StreamReaderToReadFrom
        {
            get { return streamReaderToReadFrom; }
        }



        private MveArray MvearrayToWriteTo;
        /// <summary>
        /// Data array which will hold the readed data.
        /// </summary>
        public MveArray MveArrayToWriteTo
        {
            get { return MvearrayToWriteTo; }
        }
        

        
        private int[] indecesToWriteTo;
        /// <summary>
        /// Index in the array where the readed data will be written. Set to desired value before first reading.
        /// The index increments automatically.
        /// </summary>
        public int[] IndecesToWriteTo
        {
            get { return indecesToWriteTo; }
            set { indecesToWriteTo = value; }
        }


        private bool invertEndian;
        /// <summary>
        /// Specifies wether the reader shall chanda data endian on reading.
        /// </summary>
        public bool InvertEndian
        {
            get { return invertEndian; }
        }

        #endregion

        #region variables

        // variable for iterations
        private int i;
        // helpfull variable for reading Ndimensional vector
        private double[] vecN;      

        // varibales for temporal storage
        private Scalar scalarValue;
        private Integer integerValue;
        private Vector2D vector2DValue;
        private Vector3D vector3DValue;

        private byte[] twoBytesA;
        private byte[] twoBytesB = new byte[2];
        private byte[] fourBytesA;
        private byte[] fourBytesB = new byte[4];
        private byte[] sixBytesA;
        private byte[] sixBytesB = new byte[6];
        private byte[] eightBytesA;
        private byte[] eightBytesB = new byte[8];
        private byte[] twelveBytesA;
        private byte[] twelveBytesB = new byte[12];
        private byte[] sixteenBytesA;
        private byte[] sixteenBytesB = new byte[16];
        private byte[] twentyFourBytesA;
        private byte[] twentyFourBytesB = new byte[24];
        //</>
        #endregion

        /// <summary>
        /// Creates an instance of TypeReader.
        /// </summary>
        /// <param name="attrName">Name of the attribute the reader will read.</param>
        /// <param name="attrType">Type of the attribure the reader will read.</param>
        /// <param name="attrTypeDimension">Dimension of the attribute the reader will read.</param>
        /// <param name="arrayToWriteTo">MveArray in which the reader will write the readed data.</param>
        /// <param name="invertEndian">Specifies wether to change data endian on reading.</param>
        public BL_TypeReader(string attrName, string attrType, int attrTypeDimension, MveArray arrayToWriteTo,
            bool invertEndian)
        {

            // Porpagate the parameters to the instance variables.
            this.attrName = attrName;
            this.MvearrayToWriteTo = arrayToWriteTo;
            this.attrType = attrType;
            this.attrTypeDimension = attrTypeDimension;
            this.invertEndian = invertEndian;

            // Create a new array for Ndimensional attribute.
            vecN = new double[attrTypeDimension];

            // Allocate space for structures;
            scalarValue = new Scalar(0);
            integerValue = new Integer(0);
            vector2DValue = new Vector2D(0,0);
            vector3DValue = new Vector3D(0, 0, 0);

            
        }

        /// <summary>
        /// Sets the stream reader reference.
        /// </summary>
        /// <param name="reader">The stream reader.</param>
        public void SetTheStreamReader(BinaryReader reader)
        {
            this.streamReaderToReadFrom = reader;
        }


        //////////////////////////
        // 1 byte types reading //
        //////////////////////////
        #region uint8reading

        /// <summary>
        /// Read a Integer uint8 (byte) attribute.
        /// </summary>
        public void ReadUInt8scal()
        {
            // Read the Integer uint8 attribute.
            integerValue.Value = streamReaderToReadFrom.ReadByte();
            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector uint8 (byte) attribute.
        /// </summary>
        public void ReadUInt8vec2()
        {
            // Read the 2D vector uint8 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadByte();
            vector2DValue.Y = streamReaderToReadFrom.ReadByte();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector uint8 (byte) attribute.
        /// </summary>
        public void ReadUInt8vec3()
        {
            // Read the 3D vector uint8 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadByte();
            vector3DValue.Y = streamReaderToReadFrom.ReadByte();
            vector3DValue.Z = streamReaderToReadFrom.ReadByte();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector uint8 (byte) attribute.
        /// </summary>
        public void ReadUInt8vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint8 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadByte();
            }

            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region int8reading

        /// <summary>
        /// Read a Integer int8 (sbyte) attribute.
        /// </summary>
        public void ReadInt8scal()
        {
            // Read the Integer int8 attribute.
            integerValue = streamReaderToReadFrom.ReadSByte();

            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector int8 (sbyte) attribute.
        /// </summary>
        public void ReadInt8vec2()
        {
            // Read the 2D vector int8 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadSByte();
            vector2DValue.Y = streamReaderToReadFrom.ReadSByte();
            
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector int8 (sbyte) attribute.
        /// </summary>
        public void ReadInt8vec3()
        {
            // Read the 3D vector int8 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadSByte();
            vector3DValue.Y = streamReaderToReadFrom.ReadSByte();
            vector3DValue.Z = streamReaderToReadFrom.ReadSByte();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector int8 (sbyte) attribute.
        /// </summary>
        public void readInt8vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector int8 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadSByte();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion



        //////////////////////////
        // 2 byte types reading //
        //////////////////////////
        #region uint16reading

        /// <summary>
        /// Read a Integer uint16 (ushort) attribute.
        /// </summary>
        public void ReadUInt16scal()
        {
            // Read the Integer uint16 attribute.
            integerValue = streamReaderToReadFrom.ReadUInt16();
            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector uint16 (ushort) attribute.
        /// </summary>
        public void ReadUInt16vec2()
        {
            // Read the 2D vector uint16 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadUInt16();
            vector2DValue.Y = streamReaderToReadFrom.ReadUInt16();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector uint16 (ushort) attribute.
        /// </summary>
        public void ReadUInt16vec3()
        {
            // Read the 3D vector uint16 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadUInt16();
            vector3DValue.Y = streamReaderToReadFrom.ReadUInt16();
            vector3DValue.Z = streamReaderToReadFrom.ReadUInt16();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector uint16 (ushort)attribute.
        /// </summary>
        public void ReadUInt16vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadUInt16();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region uint16invertEndianReading

        /// <summary>
        /// Reads a Integer uint16(ushort) attrubite and changes its endian.
        /// </summary>
        public void ReadUInt16scalInv()
        {
            // Read the Integer and invert its endian. 
            twoBytesA = streamReaderToReadFrom.ReadBytes(2);
            twoBytesB[0] = twoBytesA[1];
            twoBytesB[1] = twoBytesA[0];
            
            // Assign the value to the array.
            integerValue.Value = BitConverter.ToUInt16(twoBytesB, 0);
            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector uint16 (ushort) attribute and changes its endian.
        /// </summary>
        public void ReadUInt16vec2Inv()
        {
            // Read the bytes from the stream.
            fourBytesA = streamReaderToReadFrom.ReadBytes(4);
            
            // Invert endian and assign x component.
            fourBytesB[1] = fourBytesA[0];
            fourBytesB[0] = fourBytesA[1];
            vector2DValue.X = BitConverter.ToUInt16(fourBytesB, 0);

            // Invert endian and assign y component.
            fourBytesB[3] = fourBytesA[2];
            fourBytesB[2] = fourBytesA[3];
            vector2DValue.Y = BitConverter.ToUInt16(fourBytesB, 2);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector uint16 (ushort) attribute and changes its endian.
        /// </summary>
        public void ReadUInt16vec3Inv()
        {
            // Read the bytes from the stream.
            sixBytesA = streamReaderToReadFrom.ReadBytes(6);

            // Invert endian and assign x component.
            sixBytesB[1] = sixBytesA[0];
            sixBytesB[0] = sixBytesA[1];
            vector3DValue.X = BitConverter.ToUInt16(sixBytesB, 0);

            // Invert endian and assign y component.
            sixBytesB[3] = sixBytesA[2];
            sixBytesB[2] = sixBytesA[3];
            vector3DValue.Y = BitConverter.ToUInt16(sixBytesB,  2);

            // Invert endian and assign z component.
            sixBytesB[5] = sixBytesA[4];
            sixBytesB[4] = sixBytesA[5];
            vector3DValue.Z = BitConverter.ToUInt16(sixBytesB, 4);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector uint16 (ushort) attribute and changes its endian.
        /// </summary>
        public void ReadUInt16vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                twoBytesA = streamReaderToReadFrom.ReadBytes(2);
                twoBytesB[0] = twoBytesA[1];
                twoBytesB[1] = twoBytesA[0];

                vecN[i] = BitConverter.ToUInt16(twoBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion


        #region int16reading

        /// <summary>
        /// Read a Integer int16 (short) attribute.
        /// </summary>
        public void ReadInt16scal()
        {
            // Read the Integer uint16 attribute.
            integerValue.Value = streamReaderToReadFrom.ReadInt16();
            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector int16 (short) attribute.
        /// </summary>
        public void ReadInt16vec2()
        {
            // Read the 2D vector int16 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadInt16();
            vector2DValue.Y = streamReaderToReadFrom.ReadInt16();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector int16 (short) attribute.
        /// </summary>
        public void ReadInt16vec3()
        {
            // Read the 3D vector int16 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadInt16();
            vector3DValue.Y = streamReaderToReadFrom.ReadInt16();
            vector3DValue.Z = streamReaderToReadFrom.ReadInt16();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector int16 (short)attribute.
        /// </summary>
        public void ReadInt16vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadInt16();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region int16invertEndianReading

        /// <summary>
        /// Reads a Integer int16(short) attrubite and changes its endian.
        /// </summary>
        public void ReadInt16scalInv()
        {
            // Read the Integer and invert its endian. 
            twoBytesA = streamReaderToReadFrom.ReadBytes(2);
            twoBytesB[0] = twoBytesA[1];
            twoBytesB[1] = twoBytesA[0];

            // Assign the value to the array.
            integerValue.Value = BitConverter.ToInt16(twoBytesB, 0);

            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector int16 (short) attribute and changes its endian.
        /// </summary>
        public void ReadInt16vec2Inv()
        {
            // Read the bytes from the stream.
            fourBytesA = streamReaderToReadFrom.ReadBytes(4);

            // Invert endian and assign x component.
            fourBytesB[1] = fourBytesA[0];
            fourBytesB[0] = fourBytesA[1];
            vector2DValue.X = BitConverter.ToInt16(fourBytesB, 0);

            // Invert endian and assign y component.
            fourBytesB[3] = fourBytesA[2];
            fourBytesB[2] = fourBytesA[3];
            vector2DValue.Y = BitConverter.ToInt16(fourBytesB, 2);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector int16 (short) attribute and changes its endian.
        /// </summary>
        public void ReadInt16vec3Inv()
        {
            // Read the bytes from the stream.
            sixBytesA = streamReaderToReadFrom.ReadBytes(4);

            // Invert endian and assign x component.
            sixBytesB[1] = sixBytesA[0];
            sixBytesB[0] = sixBytesA[1];
            vector3DValue.X = BitConverter.ToInt16(sixBytesB, 0);

            // Invert endian and assign y component.
            sixBytesB[3] = sixBytesA[2];
            sixBytesB[2] = sixBytesA[3];
            vector3DValue.Y = BitConverter.ToInt16(sixBytesB, 2);

            // Invert endian and assign z component.
            sixBytesB[5] = sixBytesA[4];
            sixBytesB[4] = sixBytesA[5];
            vector3DValue.Z = BitConverter.ToInt16(sixBytesB, 4);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector int16 (short) attribute and changes its endian.
        /// </summary>
        public void ReadInt16vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                twoBytesA = streamReaderToReadFrom.ReadBytes(2);
                twoBytesB[0] = twoBytesA[1];
                twoBytesB[1] = twoBytesA[0];

                vecN[i] = BitConverter.ToInt16(twoBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        
        
        //////////////////////////
        // 4 byte types reading //
        //////////////////////////
        #region uint32reading

        /// <summary>
        /// Read a scalar uint32 (uint) attribute.
        /// </summary>
        public void readUInt32scal()
        {
            // Read the scalar uint32 attribute.
            integerValue = new Integer((int)streamReaderToReadFrom.ReadUInt32());
            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector uint32 (uint) attribute.
        /// </summary>
        public void readUInt32vec2()
        {
            // Read the 2D vector uint32 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadUInt32();
            vector2DValue.Y = streamReaderToReadFrom.ReadUInt32();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector uint32 (uint) attribute.
        /// </summary>
        public void readUInt32vec3()
        {
            // Read the 3D vector uint32 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadUInt32();
            vector3DValue.Y = streamReaderToReadFrom.ReadUInt32();
            vector3DValue.Z = streamReaderToReadFrom.ReadUInt32();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector uint32 (uint)attribute.
        /// </summary>
        public void readUInt32vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint32 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadUInt32();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region uint32invertEndianReading

        /// <summary>
        /// Reads a scalar uint32(uint) attrubite and changes its endian.
        /// </summary>
        public void ReadUInt32scalInv()
        {
            // Read the scalar and invert its endian. 
            fourBytesA = streamReaderToReadFrom.ReadBytes(4);
            fourBytesB[0] = fourBytesA[3];
            fourBytesB[1] = fourBytesA[2];
            fourBytesB[2] = fourBytesA[1];
            fourBytesB[3] = fourBytesA[0];

            // Assign the value to the array.
            scalarValue.Value = BitConverter.ToUInt32(fourBytesB, 0);
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector uint32 (uint) attribute and changes its endian.
        /// </summary>
        public void ReadUInt32vec2Inv()
        {
            // Read the bytes from the stream.
            eightBytesA= streamReaderToReadFrom.ReadBytes(8);

            // Invert endian and assign x component.
            eightBytesB[0] = eightBytesA[3];
            eightBytesB[1] = eightBytesA[2];
            eightBytesB[2] = eightBytesA[1];
            eightBytesB[3] = eightBytesA[0];
            vector2DValue.X = BitConverter.ToUInt32(eightBytesB, 0);

            // Invert endian and assign y component.
            eightBytesB[4] = eightBytesA[7];
            eightBytesB[5] = eightBytesA[6];
            eightBytesB[6] = eightBytesA[5];
            eightBytesB[7] = eightBytesA[4];
            vector2DValue.Y = BitConverter.ToUInt32(eightBytesB, 4);       

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector uint32 (uint) attribute and changes its endian.
        /// </summary>
        public void ReadUInt32vec3Inv()
        {
            // Read the bytes from the stream.
            twelveBytesA = streamReaderToReadFrom.ReadBytes(12);

            // Invert endian and assign x component.
            twelveBytesB[0] = twelveBytesA[3];
            twelveBytesB[1] = twelveBytesA[2];
            twelveBytesB[2] = twelveBytesA[1];
            twelveBytesB[3] = twelveBytesA[0];
            vector3DValue.X = BitConverter.ToUInt32(twelveBytesB, 0);

            // Invert endian and assign y component.
            twelveBytesB[4] = twelveBytesA[7];
            twelveBytesB[5] = twelveBytesA[6];
            twelveBytesB[6] = twelveBytesA[5];
            twelveBytesB[7] = twelveBytesA[4];
            vector3DValue.Y = BitConverter.ToUInt32(twelveBytesB, 4);

            // Invert endian and assign z component.
            twelveBytesB[8] = twelveBytesA[11];
            twelveBytesB[9] = twelveBytesA[10];
            twelveBytesB[10] = twelveBytesA[9];
            twelveBytesB[11] = twelveBytesA[8];
            vector3DValue.Z = BitConverter.ToUInt32(twelveBytesB, 8);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector uint32 (uint) attribute and changes its endian.
        /// </summary>
        public void ReadUInt32vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                fourBytesA = streamReaderToReadFrom.ReadBytes(4);
                fourBytesB[0] = fourBytesA[3];
                fourBytesB[1] = fourBytesA[2];
                fourBytesB[2] = fourBytesA[1];
                fourBytesB[3] = fourBytesA[0];
                vecN[i] = BitConverter.ToUInt32(fourBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion


        #region int32reading

        /// <summary>
        /// Read a scalar int32 (int) attribute.
        /// </summary>
        public void ReadInt32scal()
        {
            // Read the scalar int32 attribute.
            integerValue = streamReaderToReadFrom.ReadInt32();
            MvearrayToWriteTo[indecesToWriteTo] = integerValue;
        }

        /// <summary>
        /// Read a 2D vector int32 (int) attribute.
        /// </summary>
        public void ReadInt32vec2()
        {
            // Read the 2D vector int32 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadInt32();
            vector2DValue.Y = streamReaderToReadFrom.ReadInt32();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector int32 (int) attribute.
        /// </summary>
        public void ReadInt32vec3()
        {
            // Read the 3D vector int32 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadInt32();
            vector3DValue.Y = streamReaderToReadFrom.ReadInt32();
            vector3DValue.Z = streamReaderToReadFrom.ReadInt32();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector int32 (int)attribute.
        /// </summary>
        public void ReadInt32vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector int32 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadInt32();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region uint32invertEndianReading

        /// <summary>
        /// Reads a scalar int32(int) attrubite and changes its endian.
        /// </summary>
        public void ReadInt32scalInv()
        {
            // Read the scalar and invert its endian. 
            fourBytesA = streamReaderToReadFrom.ReadBytes(4);
            fourBytesB[0] = fourBytesA[3];
            fourBytesB[1] = fourBytesA[2];
            fourBytesB[2] = fourBytesA[1];
            fourBytesB[3] = fourBytesA[0];

            // Assign the value to the array.
            scalarValue.Value = BitConverter.ToInt32(fourBytesB, 0);
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector int32 (int) attribute and changes its endian.
        /// </summary>
        public void ReadInt32vec2Inv()
        {
            // Read the bytes from the stream.
            eightBytesA = streamReaderToReadFrom.ReadBytes(8);

            // Invert endian and assign x component.
            eightBytesB[0] = eightBytesA[3];
            eightBytesB[1] = eightBytesA[2];
            eightBytesB[2] = eightBytesA[1];
            eightBytesB[3] = eightBytesA[0];
            vector2DValue.X = BitConverter.ToInt32(eightBytesB, 0);

            // Invert endian and assign y component.
            eightBytesB[4] = eightBytesA[7];
            eightBytesB[5] = eightBytesA[6];
            eightBytesB[6] = eightBytesA[5];
            eightBytesB[7] = eightBytesA[4];
            vector2DValue.Y = BitConverter.ToInt32(eightBytesB, 4);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector int32 (int) attribute and changes its endian.
        /// </summary>
        public void ReadInt32vec3Inv()
        {
            // Read the bytes from the stream.
            twelveBytesA = streamReaderToReadFrom.ReadBytes(12);

            // Invert endian and assign x component.
            twelveBytesB[0] = twelveBytesA[3];
            twelveBytesB[1] = twelveBytesA[2];
            twelveBytesB[2] = twelveBytesA[1];
            twelveBytesB[3] = twelveBytesA[0];
            vector3DValue.X = BitConverter.ToInt32(twelveBytesB, 0);

            // Invert endian and assign y component.
            twelveBytesB[4] = twelveBytesA[7];
            twelveBytesB[5] = twelveBytesA[6];
            twelveBytesB[6] = twelveBytesA[5];
            twelveBytesB[7] = twelveBytesA[4];
            vector3DValue.Y = BitConverter.ToInt32(twelveBytesB, 4);

            // Invert endian and assign z component.
            twelveBytesB[8] = twelveBytesA[11];
            twelveBytesB[9] = twelveBytesA[10];
            twelveBytesB[10] = twelveBytesA[9];
            twelveBytesB[11] = twelveBytesA[8];
            vector3DValue.Z = BitConverter.ToInt32(twelveBytesB, 8);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector int32 (int) attribute and changes its endian.
        /// </summary>
        public void ReadInt32vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                fourBytesA = streamReaderToReadFrom.ReadBytes(4);
                fourBytesB[0] = fourBytesA[3];
                fourBytesB[1] = fourBytesA[2];
                fourBytesB[2] = fourBytesA[1];
                fourBytesB[3] = fourBytesA[0];
                vecN[i] = BitConverter.ToInt32(fourBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion


        #region float32reading

        /// <summary>
        /// Read a scalar float32 (float) attribute.
        /// </summary>
        public void ReadFloat32scal()
        {
            // Read the scalar float32 attribute.
            scalarValue.Value = streamReaderToReadFrom.ReadSingle();
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector float32 (float) attribute.
        /// </summary>
        public void ReadFloat32vec2()
        {
            // Read the 2D vector float32 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadSingle();
            vector2DValue.Y = streamReaderToReadFrom.ReadSingle();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector float32 (float) attribute.
        /// </summary>
        public void ReadFloat32vec3()
        {
            // Read the 3D vector float32 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadSingle();
            vector3DValue.Y = streamReaderToReadFrom.ReadSingle();
            vector3DValue.Z = streamReaderToReadFrom.ReadSingle();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector float32 (float)attribute.
        /// </summary>
        public void readFloat32vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector float32 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadSingle();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region float32invertEndianReading

        /// <summary>
        /// Reads a scalar float32(float) attrubite and changes its endian.
        /// </summary>
        public void ReadFloat32scalInv()
        {
            // Read the scalar and invert its endian. 
            fourBytesA = streamReaderToReadFrom.ReadBytes(4);
            fourBytesB[0] = fourBytesA[3];
            fourBytesB[1] = fourBytesA[2];
            fourBytesB[2] = fourBytesA[1];
            fourBytesB[3] = fourBytesA[0];

            // Assign the value to the array.
            scalarValue.Value = BitConverter.ToSingle(fourBytesB, 0);
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector float32 (float) attribute and changes its endian.
        /// </summary>
        public void ReadFloat32vec2Inv()
        {
            // Read the bytes from the stream.
            eightBytesA = streamReaderToReadFrom.ReadBytes(8);

            // Invert endian and assign x component.
            eightBytesB[0] = eightBytesA[3];
            eightBytesB[1] = eightBytesA[2];
            eightBytesB[2] = eightBytesA[1];
            eightBytesB[3] = eightBytesA[0];
            vector2DValue.X = BitConverter.ToSingle(eightBytesB, 0);

            // Invert endian and assign y component.
            eightBytesB[4] = eightBytesA[7];
            eightBytesB[5] = eightBytesA[6];
            eightBytesB[6] = eightBytesA[5];
            eightBytesB[7] = eightBytesA[4];
            vector2DValue.Y = BitConverter.ToSingle(eightBytesB, 4);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector float32 (float) attribute and changes its endian.
        /// </summary>
        public void ReadFloat32vec3Inv()
        {
            // Read the bytes from the stream.
            twelveBytesA = streamReaderToReadFrom.ReadBytes(12);

            // Invert endian and assign x component.
            twelveBytesB[0] = twelveBytesA[3];
            twelveBytesB[1] = twelveBytesA[2];
            twelveBytesB[2] = twelveBytesA[1];
            twelveBytesB[3] = twelveBytesA[0];
            vector3DValue.X = BitConverter.ToSingle(twelveBytesB, 0);

            // Invert endian and assign y component.
            twelveBytesB[4] = twelveBytesA[7];
            twelveBytesB[5] = twelveBytesA[6];
            twelveBytesB[6] = twelveBytesA[5];
            twelveBytesB[7] = twelveBytesA[4];
            vector3DValue.Y = BitConverter.ToSingle(twelveBytesB, 4);

            // Invert endian and assign z component.
            twelveBytesB[8] = twelveBytesA[11];
            twelveBytesB[9] = twelveBytesA[10];
            twelveBytesB[10] = twelveBytesA[9];
            twelveBytesB[11] = twelveBytesA[8];
            vector3DValue.Z = BitConverter.ToSingle(twelveBytesB, 8);

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector float32 (float) attribute and changes its endian.
        /// </summary>
        public void ReadFloat32vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                fourBytesA = streamReaderToReadFrom.ReadBytes(4);
                fourBytesB[0] = fourBytesA[3];
                fourBytesB[1] = fourBytesA[2];
                fourBytesB[2] = fourBytesA[1];
                fourBytesB[3] = fourBytesA[0];
                vecN[i] = BitConverter.ToSingle(fourBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion


        //////////////////////////
        // 8 byte types reading //
        //////////////////////////
        #region uint64reading

        /// <summary>
        /// Read a scalar uint64 (ulong) attribute.
        /// </summary>
        public void readUInt64scal()
        {
            // Read the scalar uint64 attribute.
            scalarValue.Value = streamReaderToReadFrom.ReadUInt64();
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector uint64 (ulong) attribute.
        /// </summary>
        public void readUInt64vec2()
        {
            // Read the 2D vector uint64 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadUInt64();
            vector2DValue.Y = streamReaderToReadFrom.ReadUInt64();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector uint64 (ulong) attribute.
        /// </summary>
        public void readUInt64vec3()
        {
            // Read the 3D vector uint64 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadUInt64();
            vector3DValue.Y = streamReaderToReadFrom.ReadUInt64();
            vector3DValue.Z = streamReaderToReadFrom.ReadUInt64();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector uint64 (ulong)attribute.
        /// </summary>
        public void readUInt64vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint64 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadUInt64();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region uint64invertEndianReading

        /// <summary>
        /// Reads a scalar uint64(ulong) attrubite and changes its endian.
        /// </summary>
        public void ReadUInt64scalInv()
        {
            // Read the scalar and invert its endian. 
            eightBytesA = streamReaderToReadFrom.ReadBytes(8);
            eightBytesB[0] = eightBytesA[7];
            eightBytesB[1] = eightBytesA[6];
            eightBytesB[2] = eightBytesA[5];
            eightBytesB[3] = eightBytesA[4];
            eightBytesB[4] = eightBytesA[3];
            eightBytesB[5] = eightBytesA[2];
            eightBytesB[6] = eightBytesA[1];
            eightBytesB[7] = eightBytesA[0];

            // Assign the value to the array.
            scalarValue.Value = BitConverter.ToUInt64(eightBytesB, 0);
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector uint64 (ulong) attribute and changes its endian.
        /// </summary>
        public void ReadUInt64vec2Inv()
        {
            // Read the bytes from the stream.
            sixteenBytesA = streamReaderToReadFrom.ReadBytes(16);

            // Invert endian and assign x component.
            sixteenBytesB[0] = sixteenBytesA[7];
            sixteenBytesB[1] = sixteenBytesA[6];
            sixteenBytesB[2] = sixteenBytesA[5];
            sixteenBytesB[3] = sixteenBytesA[4];
            sixteenBytesB[4] = sixteenBytesA[3];
            sixteenBytesB[5] = sixteenBytesA[2];
            sixteenBytesB[6] = sixteenBytesA[1];
            sixteenBytesB[7] = sixteenBytesA[0];

            vector2DValue.X = BitConverter.ToUInt64(sixteenBytesB, 0);

            // Invert endian and assign y component.
            sixteenBytesB[8] = sixteenBytesA[15];
            sixteenBytesB[9] = sixteenBytesA[14];
            sixteenBytesB[10] = sixteenBytesA[13];
            sixteenBytesB[11] = sixteenBytesA[12];
            sixteenBytesB[12] = sixteenBytesA[11];
            sixteenBytesB[13] = sixteenBytesA[10];
            sixteenBytesB[14] = sixteenBytesA[9];
            sixteenBytesB[15] = sixteenBytesA[8];

            vector2DValue.Y = BitConverter.ToUInt64(sixteenBytesB, 8);


            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector uint64 (ulong) attribute and changes its endian.
        /// </summary>
        public void ReadUInt64vec3Inv()
        {
            // Read the bytes from the stream.
            twentyFourBytesA = streamReaderToReadFrom.ReadBytes(24);          

            // Invert endian and assign x component.
            twentyFourBytesB[0] = twentyFourBytesA[7];
            twentyFourBytesB[1] = twentyFourBytesA[6];
            twentyFourBytesB[2] = twentyFourBytesA[5];
            twentyFourBytesB[3] = twentyFourBytesA[4];
            twentyFourBytesB[4] = twentyFourBytesA[3];
            twentyFourBytesB[5] = twentyFourBytesA[2];
            twentyFourBytesB[6] = twentyFourBytesA[1];
            twentyFourBytesB[7] = twentyFourBytesA[0];
            
            vector3DValue.X = BitConverter.ToUInt64(twentyFourBytesB, 0);

            // Invert endian and assign y component.
            twentyFourBytesB[8] = twentyFourBytesA[15];
            twentyFourBytesB[9] = twentyFourBytesA[14];
            twentyFourBytesB[10] = twentyFourBytesA[13];
            twentyFourBytesB[11] = twentyFourBytesA[12];
            twentyFourBytesB[12] = twentyFourBytesA[11];
            twentyFourBytesB[13] = twentyFourBytesA[10];
            twentyFourBytesB[14] = twentyFourBytesA[9];
            twentyFourBytesB[15] = twentyFourBytesA[8];

            vector3DValue.Y = BitConverter.ToUInt64(twentyFourBytesB, 8);

            // Invert endian and assign z component.
            twentyFourBytesB[16] = twentyFourBytesA[23];
            twentyFourBytesB[17] = twentyFourBytesA[22];
            twentyFourBytesB[18] = twentyFourBytesA[21];
            twentyFourBytesB[19] = twentyFourBytesA[20];
            twentyFourBytesB[20] = twentyFourBytesA[19];
            twentyFourBytesB[21] = twentyFourBytesA[18];
            twentyFourBytesB[22] = twentyFourBytesA[17];
            twentyFourBytesB[23] = twentyFourBytesA[16];

            vector3DValue.Z = BitConverter.ToUInt64(twentyFourBytesB, 16);
            

            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector uint64 (ulong) attribute and changes its endian.
        /// </summary>
        public void ReadUInt64vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector uint16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                eightBytesA = streamReaderToReadFrom.ReadBytes(8);
                eightBytesB[0] = eightBytesA[7];
                eightBytesB[1] = eightBytesA[6];
                eightBytesB[2] = eightBytesA[5];
                eightBytesB[3] = eightBytesA[4];
                eightBytesB[4] = eightBytesA[3];
                eightBytesB[5] = eightBytesA[2];
                eightBytesB[6] = eightBytesA[1];
                eightBytesB[7] = eightBytesA[0];

                vecN[i] = BitConverter.ToUInt64(eightBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion


        #region uint64reading

        /// <summary>
        /// Read a scalar int64 (long) attribute.
        /// </summary>
        public void readInt64scal()
        {
            // Read the scalar int64 attribute.
            scalarValue.Value = streamReaderToReadFrom.ReadInt64();
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector int64 (long) attribute.
        /// </summary>
        public void readInt64vec2()
        {
            // Read the 2D vector int64 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadInt64();
            vector2DValue.Y = streamReaderToReadFrom.ReadInt64();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector int64 (long) attribute.
        /// </summary>
        public void readInt64vec3()
        {
            // Read the 3D vector uint64 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadInt64();
            vector3DValue.Y = streamReaderToReadFrom.ReadInt64();
            vector3DValue.Z = streamReaderToReadFrom.ReadInt64();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector int64 (long)attribute.
        /// </summary>
        public void readInt64vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector int64 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadInt64();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region int64invertEndianReading

        /// <summary>
        /// Reads a scalar int64(long) attrubite and changes its endian.
        /// </summary>
        public void ReadInt64scalInv()
        {
            // Read the scalar and invert its endian. 
            eightBytesA = streamReaderToReadFrom.ReadBytes(8);
            eightBytesB[0] = eightBytesA[7];
            eightBytesB[1] = eightBytesA[6];
            eightBytesB[2] = eightBytesA[5];
            eightBytesB[3] = eightBytesA[4];
            eightBytesB[4] = eightBytesA[3];
            eightBytesB[5] = eightBytesA[2];
            eightBytesB[6] = eightBytesA[1];
            eightBytesB[7] = eightBytesA[0];

            // Assign the value to the array.
            scalarValue.Value = BitConverter.ToInt64(eightBytesB, 0);
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector int64 (long) attribute and changes its endian.
        /// </summary>
        public void ReadInt64vec2Inv()
        {
            // Read the bytes from the stream.
            sixteenBytesA = streamReaderToReadFrom.ReadBytes(16);

            // Invert endian and assign x component.
            sixteenBytesB[0] = sixteenBytesA[7];
            sixteenBytesB[1] = sixteenBytesA[6];
            sixteenBytesB[2] = sixteenBytesA[5];
            sixteenBytesB[3] = sixteenBytesA[4];
            sixteenBytesB[4] = sixteenBytesA[3];
            sixteenBytesB[5] = sixteenBytesA[2];
            sixteenBytesB[6] = sixteenBytesA[1];
            sixteenBytesB[7] = sixteenBytesA[0];

            vector2DValue.X = BitConverter.ToInt64(sixteenBytesB, 0);

            // Invert endian and assign y component.
            sixteenBytesB[8] = sixteenBytesA[15];
            sixteenBytesB[9] = sixteenBytesA[14];
            sixteenBytesB[10] = sixteenBytesA[13];
            sixteenBytesB[11] = sixteenBytesA[12];
            sixteenBytesB[12] = sixteenBytesA[11];
            sixteenBytesB[13] = sixteenBytesA[10];
            sixteenBytesB[14] = sixteenBytesA[9];
            sixteenBytesB[15] = sixteenBytesA[8];

            vector2DValue.Y = BitConverter.ToInt64(sixteenBytesB, 8);


            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector int64 (long) attribute and changes its endian.
        /// </summary>
        public void ReadInt64vec3Inv()
        {
            // Read the bytes from the stream.
            twentyFourBytesA = streamReaderToReadFrom.ReadBytes(24);

            // Invert endian and assign x component.
            twentyFourBytesB[0] = twentyFourBytesA[7];
            twentyFourBytesB[1] = twentyFourBytesA[6];
            twentyFourBytesB[2] = twentyFourBytesA[5];
            twentyFourBytesB[3] = twentyFourBytesA[4];
            twentyFourBytesB[4] = twentyFourBytesA[3];
            twentyFourBytesB[5] = twentyFourBytesA[2];
            twentyFourBytesB[6] = twentyFourBytesA[1];
            twentyFourBytesB[7] = twentyFourBytesA[0];

            vector3DValue.X = BitConverter.ToInt64(twentyFourBytesB, 0);

            // Invert endian and assign y component.
            twentyFourBytesB[8] = twentyFourBytesA[15];
            twentyFourBytesB[9] = twentyFourBytesA[14];
            twentyFourBytesB[10] = twentyFourBytesA[13];
            twentyFourBytesB[11] = twentyFourBytesA[12];
            twentyFourBytesB[12] = twentyFourBytesA[11];
            twentyFourBytesB[13] = twentyFourBytesA[10];
            twentyFourBytesB[14] = twentyFourBytesA[9];
            twentyFourBytesB[15] = twentyFourBytesA[8];

            vector3DValue.Y = BitConverter.ToInt64(twentyFourBytesB, 8);

            // Invert endian and assign z component.
            twentyFourBytesB[16] = twentyFourBytesA[23];
            twentyFourBytesB[17] = twentyFourBytesA[22];
            twentyFourBytesB[18] = twentyFourBytesA[21];
            twentyFourBytesB[19] = twentyFourBytesA[20];
            twentyFourBytesB[20] = twentyFourBytesA[19];
            twentyFourBytesB[21] = twentyFourBytesA[18];
            twentyFourBytesB[22] = twentyFourBytesA[17];
            twentyFourBytesB[23] = twentyFourBytesA[16];

            vector3DValue.Z = BitConverter.ToInt64(twentyFourBytesB, 16);


            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector int64 (long) attribute and changes its endian.
        /// </summary>
        public void ReadInt64vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector int16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                eightBytesA = streamReaderToReadFrom.ReadBytes(8);
                eightBytesB[0] = eightBytesA[7];
                eightBytesB[1] = eightBytesA[6];
                eightBytesB[2] = eightBytesA[5];
                eightBytesB[3] = eightBytesA[4];
                eightBytesB[4] = eightBytesA[3];
                eightBytesB[5] = eightBytesA[2];
                eightBytesB[6] = eightBytesA[1];
                eightBytesB[7] = eightBytesA[0];

                vecN[i] = BitConverter.ToInt64(eightBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion
        

        #region float64reading

        /// <summary>
        /// Read a scalar float64 (double) attribute.
        /// </summary>
        public void ReadFloat64scal()
        {
            // Read the scalar float64 attribute.
            scalarValue.Value = streamReaderToReadFrom.ReadDouble();
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector float64 (double) attribute.
        /// </summary>
        public void ReadFloat64vec2()
        {
            // Read the 2D vector float64 attribute.
            vector2DValue.X = streamReaderToReadFrom.ReadDouble();
            vector2DValue.Y = streamReaderToReadFrom.ReadDouble();

            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector float64 (float) attribute.
        /// </summary>
        public void ReadFloat64vec3()
        {
            // Read the 3D vector float64 attribute.
            vector3DValue.X = streamReaderToReadFrom.ReadDouble();
            vector3DValue.Y = streamReaderToReadFrom.ReadDouble();
            vector3DValue.Z = streamReaderToReadFrom.ReadDouble();

            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector float64 (float)attribute.
        /// </summary>
        public void ReadFloat64vecN()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector float64 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                vecN[i] = streamReaderToReadFrom.ReadDouble();
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion

        #region float64invertEndianReading

        /// <summary>
        /// Reads a scalar float64(double) attrubite and changes its endian.
        /// </summary>
        public void ReadFloat64scalInv()
        {
            // Read the scalar and invert its endian. 
            eightBytesA = streamReaderToReadFrom.ReadBytes(8);
            eightBytesB[0] = eightBytesA[7];
            eightBytesB[1] = eightBytesA[6];
            eightBytesB[2] = eightBytesA[5];
            eightBytesB[3] = eightBytesA[4];
            eightBytesB[4] = eightBytesA[3];
            eightBytesB[5] = eightBytesA[2];
            eightBytesB[6] = eightBytesA[1];
            eightBytesB[7] = eightBytesA[0];

            // Assign the value to the array.
            scalarValue.Value = BitConverter.ToDouble(eightBytesB, 0);
            MvearrayToWriteTo[indecesToWriteTo] = scalarValue;
        }

        /// <summary>
        /// Read a 2D vector float64 (double) attribute and changes its endian.
        /// </summary>
        public void ReadFloat64vec2Inv()
        {
            // Read the bytes from the stream.
            sixteenBytesA = streamReaderToReadFrom.ReadBytes(16);

            // Invert endian and assign x component.
            sixteenBytesB[0] = sixteenBytesA[7];
            sixteenBytesB[1] = sixteenBytesA[6];
            sixteenBytesB[2] = sixteenBytesA[5];
            sixteenBytesB[3] = sixteenBytesA[4];
            sixteenBytesB[4] = sixteenBytesA[3];
            sixteenBytesB[5] = sixteenBytesA[2];
            sixteenBytesB[6] = sixteenBytesA[1];
            sixteenBytesB[7] = sixteenBytesA[0];

            vector2DValue.X = BitConverter.ToDouble(sixteenBytesB, 0);

            // Invert endian and assign y component.
            sixteenBytesB[8] = sixteenBytesA[15];
            sixteenBytesB[9] = sixteenBytesA[14];
            sixteenBytesB[10] = sixteenBytesA[13];
            sixteenBytesB[11] = sixteenBytesA[12];
            sixteenBytesB[12] = sixteenBytesA[11];
            sixteenBytesB[13] = sixteenBytesA[10];
            sixteenBytesB[14] = sixteenBytesA[9];
            sixteenBytesB[15] = sixteenBytesA[8];

            vector2DValue.Y = BitConverter.ToDouble(sixteenBytesB, 8);


            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector2DValue;
        }

        /// <summary>
        /// Read a 3D vector float64 (double) attribute and changes its endian.
        /// </summary>
        public void ReadFloat64vec3Inv()
        {
            // Read the bytes from the stream.
            twentyFourBytesA = streamReaderToReadFrom.ReadBytes(24);

            // Invert endian and assign x component.
            twentyFourBytesB[0] = twentyFourBytesA[7];
            twentyFourBytesB[1] = twentyFourBytesA[6];
            twentyFourBytesB[2] = twentyFourBytesA[5];
            twentyFourBytesB[3] = twentyFourBytesA[4];
            twentyFourBytesB[4] = twentyFourBytesA[3];
            twentyFourBytesB[5] = twentyFourBytesA[2];
            twentyFourBytesB[6] = twentyFourBytesA[1];
            twentyFourBytesB[7] = twentyFourBytesA[0];

            vector3DValue.X = BitConverter.ToDouble(twentyFourBytesB, 0);

            // Invert endian and assign y component.
            twentyFourBytesB[8] = twentyFourBytesA[15];
            twentyFourBytesB[9] = twentyFourBytesA[14];
            twentyFourBytesB[10] = twentyFourBytesA[13];
            twentyFourBytesB[11] = twentyFourBytesA[12];
            twentyFourBytesB[12] = twentyFourBytesA[11];
            twentyFourBytesB[13] = twentyFourBytesA[10];
            twentyFourBytesB[14] = twentyFourBytesA[9];
            twentyFourBytesB[15] = twentyFourBytesA[8];

            vector3DValue.Y = BitConverter.ToDouble(twentyFourBytesB, 8);

            // Invert endian and assign z component.
            twentyFourBytesB[16] = twentyFourBytesA[23];
            twentyFourBytesB[17] = twentyFourBytesA[22];
            twentyFourBytesB[18] = twentyFourBytesA[21];
            twentyFourBytesB[19] = twentyFourBytesA[20];
            twentyFourBytesB[20] = twentyFourBytesA[19];
            twentyFourBytesB[21] = twentyFourBytesA[18];
            twentyFourBytesB[22] = twentyFourBytesA[17];
            twentyFourBytesB[23] = twentyFourBytesA[16];

            vector3DValue.Z = BitConverter.ToDouble(twentyFourBytesB, 16);


            // Assign the value to the array.
            MvearrayToWriteTo[indecesToWriteTo] = vector3DValue;
        }

        /// <summary>
        /// Read a ND vector float64 (double) attribute and changes its endian.
        /// </summary>
        public void ReadFloat64vecNInv()
        {
            // Create a new array of values.
            vecN = new double[attrTypeDimension];

            // Read the ND vector int16 attribute.
            for (i = 0; i < attrTypeDimension; i++)
            {
                eightBytesA = streamReaderToReadFrom.ReadBytes(8);
                eightBytesB[0] = eightBytesA[7];
                eightBytesB[1] = eightBytesA[6];
                eightBytesB[2] = eightBytesA[5];
                eightBytesB[3] = eightBytesA[4];
                eightBytesB[4] = eightBytesA[3];
                eightBytesB[5] = eightBytesA[2];
                eightBytesB[6] = eightBytesA[1];
                eightBytesB[7] = eightBytesA[0];

                vecN[i] = BitConverter.ToDouble(eightBytesB, 0);
            }

            // Increment the attribute array index to write to.
            MvearrayToWriteTo[indecesToWriteTo] = new VectorND(vecN);
        }

        #endregion
    }
}