using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.Windows.Forms;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// This is a RegGridND version of a module that calculates new attribute values from existing attributes based on given expression.
    /// You can choose whether to work with point attributes or cell attributes.
    /// With point data, you can save the result as new coordinates or as a point attribute.
    /// For cell data, you have to specify the name of the cells array to be used.
    /// The results can be saved to a new attribute or an existing one, if an attribute with the specified name already exists.<br></br>
    /// <br></br>
    /// Expression syntax:<br></br>
    ///   <b>numbers</b> with '.' or ',' as decimal separator<br></br>
    ///   names of input <b>attributes</b><br></br>
    ///   basic operations '<b>+</b>', '<b>-</b>', '<b>*</b>', '<b>/</b>', '<b>%</b>'<br></br>
    ///   <b>point</b> ... coordinates of the current point (only while working with points)<br></br>
    ///   <b>index</b> ... index of the current point or cell<br></br>
    ///   <b>pow(x; y)</b> ... x to the power of y<br></br>
    ///   <b>log(x; y)</b> ... logarithm of x with the root y<br></br>
    ///   <b>exp(x)</b> ... exponential function of x<br></br>
    ///   <b>ln(x)</b> ... natural logarithm of x<br></br>
    ///   <b>sqrt(x)</b> ... square root of x<br></br>
    ///   <b>sin(x)</b>, <b>cos(x)</b>, <b>tan(x)</b> ... goniometric functions of x<br></br>
    ///   <b>asin(x)</b>, <b>acos(x)</b>, <b>atan(x)</b> ... inverse goniometric functions of x<br></br>
    ///   <b>sinh(x)</b>, <b>cosh(x)</b>, <b>tanh(x)</b> ... hyperbolic functions of x<br></br>
    ///   <b>abs(x)</b> ... absolute value of x<br></br>
    ///   <b>ceil(x)</b> ... x rounded to the next higher integer<br></br>
    ///   <b>floor(x)</b> ... x rounded to the next lower integer<br></br>
    ///   <b>ivec</b>, <b>jvec</b>, <b>kvec</b> ... unit vectors of coordinate axes<br></br>
    ///   <b>norm(x)</b> ... magnitude of vector x<br></br>
    ///   <b>smul(x; y)</b> ... dot product of vectors x and y<br></br>
    ///   <b>vmul(x; y)</b> ... cross product of vectors x and y<br></br>
    ///   <b>x[n]</b> ... value of the nth element of vector x<br></br>
    /// <br></br>
    /// Functions norm, smul, vmul and [] require vector operand(s).
    /// Addition and subtraction can be performed on either two vectors or two scalars,
    /// multiplication and division on one vector and one scalar.
    /// All other functions work only with scalar operand(s).
    /// You can create a vector by multiplying the coordinates by ivec, jvec and kvec and adding them together.
    /// </summary>
    [ModuleInfo("Jan Krej, Oldich Petk", "A module that performs scientific calculations on point or cell attributes of an RegGridND.", CategoryType.Filter, CategoryData.Volume, IconName="calc.ico")]
    [PortInfo("Input", "Input RegGridND data structure. Must contain attributes used in the expression.")]
    [PortInfo("Ouptut", "Output RegGridND containing resulting attribute of the specified name. If attribute with such name already exists, it will be overwritten.")]
    public class RegGridNDCalculator : Mve.Core.Module
    {
        private const string helpStr =
            "This is a RegGridND version of a module that calculates new attribute values from existing attributes based on given expression. " +
            "You can choose whether to work with point attributes or cell attributes using the combo box labeled \"Work with\". " +
            "With point data selected, you can use the checkbox on the right to save the result as new coordinates or as a point attribute. " +
            "For cell data, you have to specify the name of the cells array to be used, in the text box on the right. " +
            "The results can be saved to a new attribute or an existing one, if an attribute with the specified name already exists.\n\n" +
            "Expression syntax:\n" +
            "  numbers with '.' or ',' as decimal separator\n" +
            "  names of input attributes\n" +
            "  basic operations '+', '-', '*', '/', '%'\n" +
            "  point ... coordinates of the current point (only while working with points)\n" +
            "  index ... index of the current point or cell\n" +
            "  pow(x; y) ... x to the power of y\n" +
            "  log(x; y) ... logarithm of x with the root y\n" +
            "  exp(x) ... exponential function of x\n" +
            "  ln(x) ... natural logarithm of x\n" +
            "  sqrt(x) ... square root of x\n" +
            "  sin(x), cos(x), tan(x) ... goniometric functions of x\n" +
            "  asin(x), acos(x), atan(x) ... inverse goniometric functions of x\n" +
            "  sinh(x), cosh(x), tanh(x) ... hyperbolic functions of x\n" +
            "  abs(x) ... absolute value of x\n" +
            "  ceil(x) ... x rounded to the next higher integer\n" +
            "  floor(x) ... x rounded to the next lower integer\n" +
            "  ivec, jvec, kvec ... unit vectors of coordinate axes\n" +
            "  norm(x) ... magnitude of vector x\n" +
            "  smul(x; y) ... dot product of vectors x and y\n" +
            "  vmul(x; y) ... cross product of vectors x and y\n" +
            "  x[n] ... value of the nth element of vector x\n\n" +
            "Functions norm, smul, vmul and [] require vector operand(s). " +
            "Addition and subtraction can be performed on either two vectors or two scalars, " +
            "multiplication and division on one vector and one scalar. " +
            "All other functions work only with scalar operand(s). " +
            "You can create a vector by multiplying the coordinates by ivec, jvec and kvec and adding them together.";

        /// <summary>
        /// Expression analyzer and evaluator
        /// </summary>
        private Analyzer engine;

        /// <summary>
        /// Instance of a setup dialog
        /// </summary>
        private CalculatorSetup cs;

        /// <summary>
        /// List of attributes used in the expression along with their values for the current point or cell
        /// </summary>
        private Dictionary<string, double[]> var;

        /// <summary>
        /// List of attributes used in the expression
        /// </summary>
        private List<string> varNames;

        /// <summary>
        /// Number of elements in each dimension of the input data array (points or cells)
        /// </summary>
        private int[] inputDim;

        /// <summary>
        /// Multidimensional array index
        /// </summary>
        private int[] index;
        private double[] dIndex;

        /// <summary>
        /// Auxiliary variable for faster index incrementation
        /// </summary>
        private int idxIterator;

        /// <summary>
        /// Pointer to the input data structure
        /// </summary>
        private RegGridND data;

        /// <summary>
        /// Module constructor.
        /// </summary>
        public RegGridNDCalculator()
        {
            AddInPort("Input", typeof(RegGridND));
            AddOutPort("Output", typeof(RegGridND));

            cs = new CalculatorSetup(false, helpStr);
        }

        /// <summary>
        /// Writes module configuration to the map file.
        /// </summary>
        /// <param name="config">XML element with the module configuration</param>
        public override void WriteConfig(System.Xml.XmlElement config)
        {
            config.SetAttribute("cellData", cs.CellData.ToString());
            config.SetAttribute("outputName", cs.OutputName.ToString());
            config.InnerText = cs.Expression;
        }

        /// <summary>
        /// Reads saved module configuration.
        /// </summary>
        /// <param name="config">XML element with the module configuration</param>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            cs.SetConfig(
                config.InnerText,
                bool.Parse(config.GetAttribute("cellData")),
                "",
                config.GetAttribute("outputName"),
                false
            );
        }

        /// <summary>
        /// Increments the multidimensional index.
        /// </summary>
        /// <returns>[false] if the index reached the end of the array, [true] otherwise</returns>
        private bool Increment()
        {
            index[0]++;
            idxIterator = 0;
            for (; idxIterator < index.Length - 1;)
            {
                if (index[idxIterator] == inputDim[idxIterator])
                {
                    index[idxIterator] = 0;
                    dIndex[idxIterator] = 0;
                    index[++idxIterator]++;
                    dIndex[idxIterator] = index[idxIterator];
                }
                else idxIterator++;
            }
            if (index[idxIterator] == inputDim[idxIterator]) return false;
            return true;
        }

        /// <summary>
        /// Resets the index to its initial value.
        /// </summary>
        private void ResetIndex()
        {
            for (int i = 0; i < index.Length; i++)
            {
                index[i] = 0;
                dIndex[i] = 0;
            }
        }

        /// <summary>
        /// Executes the calculation.
        /// </summary>
        public override void Execute()
        {
            Dictionary<string, MveArray> attrs = new Dictionary<string,MveArray>();
            double[] result;
            MveArray results = null;
            MveArray current;
            int resLen = 0;
            ulong itemCount;
            IVector iv;
            int ivLen;
            double[] ivArr;
            long ticks;
            ulong item = 0;

            if (cs.Expression.Trim().Length == 0)
            {
                SetOutput("Output", GetInput("Input"), IsAllDataSame());
                return;
            }

            // some of the settings have changed since last run, we need to analyze the expression
            if (engine == null || cs.Changed)
            {
                engine = new Analyzer();
                varNames = engine.Analyze(cs.Expression);
            }

            // input or settings have changed, new calculation is needed
            if (data == null || !IsAllDataSame() || cs.Changed)
            {
                if (varNames == null) varNames = new List<string>();
                var = new Dictionary<string, double[]>(varNames.Count);

                ticks = DateTime.Now.Ticks;

                // make a shallow copy of the input
                data = ((RegGridND)GetInput("Input")).ShallowCopy();
                cs.Changed = false;

                inputDim = data.Dimensions;

                // now we can extract the arrays of attributes needed for calculation
                if (cs.CellData)
                {
                    foreach (string attr in varNames)
                    {
                        current = data.GetCellAttrs(attr);
                        if (current != null) attrs.Add(attr, current);
                        else throw new MveException(this, "Input data does not contain cell attributes named " + attr);
                    }
                }
                else
                {
                    varNames.Remove("point");
                    varNames.Remove("index");

                    foreach (string attr in varNames)
                    {
                        current = data.GetPointAttrs(attr);
                        if (current != null) attrs.Add(attr, current);
                        else throw new MveException(this, "Input data does not contain point attributes named " + attr);
                    }
                    // for points increase the counts by one in each dimension
                    for (int i = 0; i < inputDim.Length; i++) inputDim[i]++;
                }

                // initialize multidimensional index array
                index = new int[inputDim.Length];
                dIndex = new double[inputDim.Length];
                ResetIndex();

                // get the number of all items in the input (for progressbar display)
                itemCount = 1;
                for (int i = 0; i < inputDim.Length; i++)
                {
                    itemCount *= (ulong)inputDim[i];
                }

                for (item = 0; Increment(); item++)
                {
                    ProgressInfo((double)item / itemCount);

                    // we need to convert MVE types to double arrays for the calculator engine
                    var["index"] = dIndex;
                    if (!cs.CellData) var["point"] = data.GetPoint(index).Values;
                    foreach (string attr in varNames)
                    {
                        current = attrs[attr];
                        if (current.ElemType == typeof(Scalar)) var[attr] = new double[] { ((Scalar)attrs[attr][index]).Value };
                        else if (current.ElemType == typeof(Integer)) var[attr] = new double[] { ((Integer)attrs[attr][index]).Value };
                        else if (typeof(IVector).IsAssignableFrom(current.ElemType))
                        {
                            iv = (IVector)attrs[attr][index];
                            ivLen = iv.Length;
                            ivArr = new double[iv.Length];
                            for (int j = 0; j < ivLen; j++) ivArr[j] = iv[j];
                            var[attr] = ivArr;
                        }
                    }

                    if (item == 0)
                    {
                        // for the first item
                        // run the first pass (checks for all errors - array overflows, invalid vector operations etc.)
                        result = engine.FirstResult(var);

                        // create result arrays and convert result back to MVE datatypes
                        resLen = result.Length;
                        switch (resLen)
                        {
                            case 1:
                                results = new MveArray(typeof(Scalar), inputDim);
                                results[index] = new Scalar(result[0]);
                                break;
                            case 2:
                                results = new MveArray(typeof(Vector2D), inputDim);
                                results[index] = new Vector2D(result[0], result[1]);
                                break;
                            case 3:
                                results = new MveArray(typeof(Vector3D), inputDim);
                                results[index] = new Vector3D(result[0], result[1], result[2]);
                                break;
                            default:
                                results = new MveArray(typeof(VectorND), inputDim);
                                results[index] = new VectorND(result);
                                break;
                        }
                    }
                    else
                    {
                        // for each other item
                        // run the next pass (only runtime error checking - division by zero etc.)
                        result = engine.NextResult(var);

                        // convert result back to MVE datatypes
                        switch (resLen)
                        {
                            case 1:
                                results[index] = new Scalar(result[0]);
                                break;
                            case 2:
                                results[index] = new Vector2D(result[0], result[1]);
                                break;
                            case 3:
                                results[index] = new Vector3D(result[0], result[1], result[2]);
                                break;
                            default:
                                results[index] = new VectorND(result);
                                break;
                        }
                    }
                }

                // save resulting array
                if (cs.CellData)
                {
                    // add or replace attribute of the current cells
                    try
                    {
                        data.RemoveCellAttrs(cs.OutputName);
                    }
                    catch (Exception) { }
                    data.AddCellAttrs(cs.OutputName, results);
                }
                else
                {
                    try
                    {
                        data.RemovePointAttrs(cs.OutputName);
                    }
                    catch (Exception) { }
                    data.AddPointAttrs(cs.OutputName, results);
                }

                ticks = DateTime.Now.Ticks - ticks;
                Console.WriteLine("Calculation ended: {0} values in {1} ms.", item, (int)(ticks / 10000.0));

                // send resulting data structure to output
                SetOutput("Output", data, false);
            }
            // input data and module settings haven't changed since last run
            else
            {
                // we can send last data structure to output with dataSame indication
                SetOutput("Output", data, true);
            }
        }

        /// <summary>
        /// Shows the module setup dialog.
        /// </summary>
        /// <returns>instance of the setup dialog</returns>
        public override Zcu.Mve.Core.ModuleSetup InvokeSetup()
        {
            return cs;
        }
    }
}
