using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

//File with data types of ColorPointList and bod used by EditablePalette and PaletteGUI.

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Interpolation function delegate.
    /// </summary>
    /// <param name="left">left color</param>
    /// <param name="right">right color</param>
    /// <param name="leftY">left height (alpha, 0..1)</param>
    /// <param name="rightY">right height (alpha, 0..1)</param>
    /// <param name="pos">position (0..1, 0 = left, 1 = right)</param>
    /// <returns>interpolated color</returns>
    public delegate Color InterpolateFunc(Color left, Color right, double leftY, double rightY, double pos);

    /// <summary>
    /// Data type, it consists of other data type ColorPoint. It is bidirectional connected list.
    /// </summary>
    public class ColorPointList
    {
        /// <summary>
        /// Interpolates colors with linear transition.
        /// </summary>
        private static Color InterpolateLinear(Color left, Color right, double leftY, double rightY, double pos)
        {
            double ipos = 1 - pos;
            return Color.FromArgb(
                (int)Math.Round(255 * (pos * rightY + ipos * leftY)),
                (int)Math.Round(pos * right.R + ipos * left.R),
                (int)Math.Round(pos * right.G + ipos * left.G),
                (int)Math.Round(pos * right.B + ipos * left.B)
            );
        }

        /// <summary>
        /// Interpolates colors with cosine transition
        /// </summary>
        private static Color InterpolateCosine(Color left, Color right, double leftY, double rightY, double pos)
        {
            double cpos = 0.5 - 0.5 * Math.Cos(Math.PI * pos);
            return InterpolateLinear(left, right, leftY, rightY, cpos);
        }

        private static InterpolateFunc[] interps = { InterpolateLinear, InterpolateCosine };

        private ColorPoint first;
        private ColorPoint last;
        private int mode;

        /// <summary>
        /// Interpolation mode
        /// </summary>
        public int Mode
        {
            get { return mode; }
            set { mode = value; }
        }

        private InterpolateFunc interpolate { get { return interps[mode]; } }

        /// <summary>
        /// constructs a new instance of the list
        /// </summary>
        /// <param name="first">Starting element of the list.</param>
        /// <param name="last">Ending element of the list.</param>
        public ColorPointList(ColorPoint first, ColorPoint last)//for every point in seznam point on predecessor and succesor
        {
            this.first = first;
            this.last = last;
        }

        /// <summary>
        /// Method, which computes Color from input value.
        /// </summary>
        public Color GetColor(double value)
        {
            // start with the leftmost point
            ColorPoint point = first;

            // search for the first point, whose right neighbor is to the right of the value
            while (point.right != null && point.right.x <= value) point = point.right;

            // too much to the right - use the last point
            if (point.right == null)
            {
                return point.color;
            }

            // too much to the left - use the first point
            if (point.x >= value)
            {
                return point.color;
            }

            // run the interpolation
            return interpolate(point.color, point.right.color, point.y, point.right.y, (value - point.x) / (point.right.x - point.x));
        }

        /// <summary>
        /// Calculates interpolated color to the right of the specified ColorPoint.
        /// </summary>
        /// <param name="left">color point</param>
        /// <param name="value">global position value</param>
        /// <returns>interpolated color</returns>
        public Color GetColor(ColorPoint left, double value)
        {
            if (left.right == null) return left.color;

            return interpolate(left.color, left.right.color, left.y, left.right.y, (value - left.x) / (left.right.x - left.x));
        }

        /// <summary>
        /// Selects a ColorPoint at a given position, within given tolerance.
        /// </summary>
        /// <param name="x">X position of the selected point</param>
        /// <param name="y">X position of the selected point</param>
        /// <param name="epsilon">Tolerance</param>
        /// <returns>Selected ColorPoint</returns>
        public ColorPoint Select(double x, double y, double epsilon, bool add)
        {
            ColorPoint current = first;
            ColorPoint left = null;

            while (current != null)
            {
                if (Math.Abs(current.x - x) < epsilon)
                {
                    return current;
                }
                else if (current.x < x)
                {
                    left = current;
                }
                else break;

                current = current.right;
            }

            if (add)
            {
                if (left == null) return null;
                if (current == null) return null;

                Color c = interpolate(left.color, current.color, 1, 1, (x - left.x) / (current.x - left.x));
                ColorPoint n = new ColorPoint(x, y, c, left, current);
                left.right = n;
                current.left = n;
                return n;
            }
            return null;
        }

        /// <summary>
        /// returns the first element of the list
        /// </summary>
        public ColorPoint First
        {
            get
            {
                return first;
            }
        }
    }

    /// <summary>
    /// Data type for every control point(CP) in palette
    /// </summary>
    public class ColorPoint
    {
        /// <summary>
        /// x position
        /// </summary>
        public double x;

        /// <summary>
        /// y position
        /// </summary>
        public double y;

        /// <summary>
        /// color of CP
        /// </summary>
        public Color color;

        /// <summary>
        /// predecessor in the list
        /// </summary>
        public ColorPoint left;

        /// <summary>
        /// succesor in the list
        /// </summary>
        public ColorPoint right;

        /// <summary>
        /// Constructs a new instance of color point
        /// </summary>
        /// <param name="x">X position of the new ColorPoint</param>
        /// <param name="y">Y position of the new ColorPoint (represents opacity)</param>
        /// <param name="color">Color of the new ColorPoint</param>
        /// <param name="left">Predecessor in the list of ColorPoint instances</param>
        /// <param name="right">Successor in the list of ColorPoint instances</param>
        public ColorPoint(double x, double y, Color color, ColorPoint left, ColorPoint right)
        {
            this.x = x;
            this.y = y;
            this.color = color;
            this.left = left;
            this.right = right;
        }
    }
}
