using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using System.Drawing;
using System.Xml;
using System.Collections.Generic;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// A module that transforms a Scalar value into a ColorRGBA. Values lower or equal to zero will be computed as equal to zero, values greater or equal to one will be computed as equal to one , values in between will become linear interpolations among control points.*
    /// </summary>
    [ModuleInfo("Jan tefny", "A module that transforms a Scalar value into a ColorRGBA. Values lower or equal to zero will be computed as equal to zero, values greater or equal to one will be computed as equal to one , values in between will become linear interpolation or cosine interpolation among control points.", CategoryType.Filter, CategoryData.Number, IconName = "PaletteLinear.ico")]
    [PortInfo("Value", "Scalar value to be transformed into ColorRGBA")]
    [PortInfo("Color", "Resulting ColorRGBA value")]
    public class EditablePalette : Module
    {
        private PaletteGUI gui;//take palette and allows usage of its variables
        private ColorPointList list;
        bool stop = false, upToDate = false;

        /// <summary>
        /// Default constructor, adds input and output ports.
        /// </summary> 
        public EditablePalette()
        {
            AddInPort("Value", typeof(Scalar)); //input port
            AddOutPort("Color", typeof(ColorRGBA));//output port
            gui = new PaletteGUI();
        }

        /// <summary>
        /// WriteConfig method for saving of positions and values of CP on palette.
        /// </summary>
        public override void WriteConfig(System.Xml.XmlElement config)
        {
            XmlElement parent, child;
            XmlAttribute attr;
            XmlDocument doc = config.OwnerDocument;

            parent = doc.CreateElement("Palette");
            config.AppendChild(parent);

            attr = doc.CreateAttribute("Mode");
            attr.Value = gui.Function.ToString();
            parent.Attributes.Append(attr);

            ColorPoint point = gui.colorPointList.First;

            while (point != null)
            {
                child = config.OwnerDocument.CreateElement("Point");
                child.InnerText = point.x.ToString(Globals.nfi) + "," + point.y.ToString(Globals.nfi) + "," + point.color.R + "," + point.color.G + "," + point.color.B + "," + point.color.A;
                parent.AppendChild(child);
                point = point.right;
            }
        }

        /// <summary>
        /// ReadConfig method for loading of positions and values of CP which were saved by the user last time.
        /// </summary>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            XmlNode parent;
            XmlAttribute attr;

            if (config.FirstChild.LocalName.Equals("Palette")) parent = config.FirstChild;
            else parent = config; // backward compatibility mode, remove eventually

            attr = parent.Attributes["Mode"];
            if (attr != null) gui.Function = int.Parse(attr.Value);
            else gui.Function = 0; // backward compatibility mode, remove eventually

            XmlNodeList nodes = config.GetElementsByTagName("Point");
            SortedList<double, ColorPoint> points = new SortedList<double, ColorPoint>();
            ColorPoint p, prev;
            Color c;
            string[] parts;
            double xmax = 1;
            double ymax = 1;

            if (nodes.Count < 2) throw new MveException(this, "Wrong configuration. Not enough control points.");

            foreach (XmlNode node in nodes)
            {
                parts = node.InnerText.Split(',');
                if (parts.Length < 5) throw new MveException(this, "Wrong configuration. Not enough parameters for a control point.");

                if (parts.Length > 5) c = Color.FromArgb(int.Parse(parts[5]), int.Parse(parts[2]), int.Parse(parts[3]), int.Parse(parts[4]));
                else c = Color.FromArgb(int.Parse(parts[2]), int.Parse(parts[3]), int.Parse(parts[4]));

                p = new ColorPoint(double.Parse(parts[0], Globals.nfi), double.Parse(parts[1], Globals.nfi), c, null, null);
                xmax = Math.Max(xmax, p.x);
                ymax = Math.Max(ymax, p.y);
                points.Add(p.x, p);
            }

            prev = null;
            foreach (KeyValuePair<double, ColorPoint> kvp in points)
            {
                p = kvp.Value;
                p.x /= xmax;
                p.y /= ymax;
                p.left = prev;
                if (prev != null) prev.right = p;
                prev = p;
            }

            gui.colorPointList = new ColorPointList(points.Values[0], points.Values[points.Count - 1]);
        }

        /// <summary>
        /// Execution of module. It reloads palette if palette isn't actual and calls method which transfers scalar to color.
        /// </summary>
        public override void Execute()
        {
            if (!upToDate)
            {
                GetPalette();
            }

            Scalar input = (Scalar)GetInput("Value");
            double value = Math.Min(1, Math.Max(0, input.Value));

            SetOutput("Color", (ColorRGBA)list.GetColor(value));
        }

        /// <summary>
        /// Method, which gets control points from EditablePalette.
        /// </summary>
        private void GetPalette()
        {
            list = gui.colorPointList;
            upToDate = true;
        }

        /// <summary>
        /// Method, which invokes setup
        /// </summary>
        public override ModuleSetup InvokeSetup() //method for invoking setup
        {
            upToDate = false;
            return gui;
        }
    }
}
