using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Xml;

using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// A GUI for EditablePalette which allows creation of own RGB transition function. Stting of multiple controlpoints in color palette and their editting. Y value in palette represents alpha value.
    /// </summary>
    public partial class PaletteGUI : Zcu.Mve.Core.ModuleSetup
    {
        private ColorPoint selected = null;
        private int pointSize = 5; //surrounding of every point for drawing and draging by mouse
        private double epsilon;
        private bool mousedwn = false;
        private Image img;
        private Brush pattern;

        /// <summary>List of ColorPoint instances forming the palette</summary>
        public ColorPointList colorPointList;

        public int Function
        {
            get { return colorPointList.Mode; }
            set
            {
                colorPointList.Mode = value;
                comboBox1.SelectedIndex = value;
            }
        }

        /// <summary>width of the palette area</summary>
        public int width;
        /// <summary>height of the palette area</summary>
        public int height;

        /// <summary>
        /// Basic setting up of GUI on start.
        /// </summary>
        public PaletteGUI()
        {
            InitializeComponent();

            comboBox1.SelectedIndex = 0;//seting index of combobox to first one

            width = pictureBox1.Width - 2;
            height = pictureBox1.Height - 2;

            epsilon = (double)pointSize / width;

            ColorPoint left = new ColorPoint(0, 0.5, Color.Red, null, null);
            ColorPoint right = new ColorPoint(1, 0.5, Color.Blue, null, null);
            left.right = right;
            right.left = left;
            colorPointList = new ColorPointList(left, right);

            Bitmap pat = new Bitmap(16, 16, System.Drawing.Imaging.PixelFormat.Format24bppRgb);
            Graphics g = Graphics.FromImage(pat);
            g.FillRectangle(Brushes.White, 0, 0, 16, 16);
            g.FillRectangle(Brushes.LightGray, 0, 0, 8, 8);
            g.FillRectangle(Brushes.LightGray, 8, 8, 8, 8);
            g.Dispose();
            pattern = new TextureBrush(pat);

            this.Load += new EventHandler(PaletteGUI_Load);
        }

        private void PaletteGUI_Load(object sender, EventArgs e)
        {
            Repaint();
        }

        private int xToGUI(double x)
        {
            return (int)Math.Round(width * x);
        }

        private double guiToX(int x)
        {
            return (double)x / width;
        }

        private int yToGUI(double y)
        {
            return (int)Math.Round(height * y);
        }

        private double guiToY(int y)
        {
            return (double)y / height;
        }

        private void pictureBox1_MouseDown(object sender, MouseEventArgs e)//on mousedown click above GUI
        {
            if (e.Button == System.Windows.Forms.MouseButtons.Left)
            {
                selected = colorPointList.Select(guiToX(e.X), guiToY(e.Y), epsilon, true);
                mousedwn = true;
            }
            else if (e.Button == System.Windows.Forms.MouseButtons.Right)
            {
                selected = colorPointList.Select(guiToX(e.X), guiToY(e.Y), epsilon, false);
                if (selected != null)
                {
                    //
                }
            }
        }

        private void pictureBox1_MouseUp(object sender, MouseEventArgs e)//on mouse up event
        {
            mousedwn = false;

            Repaint();
        }

        private void pictureBox1_MouseMove(object sender, MouseEventArgs e)//method for moving of CP
        {
            int x, y;
            ColorPoint tmp;

            if (mousedwn)
            {
                if (selected.left == null || selected.right == null)
                {
                    x = xToGUI(selected.x);
                }
                else
                {
                    x = Math.Max(1, Math.Min(width - 1, e.X));
                    y = e.Y;

                    if (xToGUI(selected.left.x) == x) x++;
                    if (xToGUI(selected.right.x) == x) x--;

                    while (xToGUI(selected.left.x) > x)
                    {
                        tmp = selected.left;

                        selected.left = tmp.left;
                        tmp.left = selected;

                        tmp.right = selected.right;
                        selected.right = tmp;

                        if (selected.left != null) selected.left.right = selected;
                        if (tmp.right != null) tmp.right.left = tmp;
                    }

                    while (xToGUI(selected.right.x) < x)
                    {
                        tmp = selected.right;

                        selected.right = tmp.right;
                        tmp.right = selected;

                        tmp.left = selected.left;
                        selected.left = tmp;

                        if (selected.right != null) selected.right.left = selected;
                        if (tmp.left != null) tmp.left.right = tmp;
                    }

                }
                y = Math.Max(0, Math.Min(height, e.Y));

                selected.x = guiToX(x);
                selected.y = guiToY(y);
                Repaint();
            }
        }

        /// <summary>
        /// Method for repainting GUI.
        /// </summary>
        private void Repaint()
        {
            if (img == null)
            {
                img = new Bitmap(pictureBox1.Width, pictureBox1.Height, pictureBox1.CreateGraphics());
                pictureBox1.Image = img;
            }

            // get graphic context
            Graphics g = Graphics.FromImage(img);
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;

            // prepare brush and pens
            SolidBrush brush = new SolidBrush(Color.White);
            Pen penGrad = new Pen(Color.White);
            Pen penThin = Pens.Black;
            Pen penThick = new Pen(Brushes.Black, 2);
            Pen penOutline;

            g.FillRectangle(pattern, 0, 0, width, height);

            // start with the leftmost point and iterate to the right
            ColorPoint current = colorPointList.First;
            while (current != null)
            {
                // calculate the on-screen coordinates of the left and right points
                int leftX = xToGUI(current.x);
                int leftY = yToGUI(current.y);
                int rightX = width;
                int rightY = leftY;

                if (current.right != null)
                {
                    rightX = xToGUI(current.right.x);
                    rightY = yToGUI(current.right.y);
                }

                // fill the gradient between the points
                for (int x = leftX; x < rightX; x++)
                {
                    penGrad.Color = colorPointList.GetColor(current, guiToX(x));
                    g.DrawLine(penGrad, x, 0, x, height);
                }

                // draw a line connecting the points
                g.DrawLine(penThin, leftX, leftY, rightX, rightY);

                // draw circle around the left point
                brush.Color = current.color;
                penOutline = (current == selected) ? penThick : penThin;

                g.FillEllipse(brush, leftX - pointSize, leftY - pointSize, 2 * pointSize, 2 * pointSize);
                g.DrawEllipse(penOutline, leftX - pointSize, leftY - pointSize, 2 * pointSize, 2 * pointSize);

                // move one point to the right
                current = current.right;
            }

            g.Dispose();

            pictureBox1.Invalidate();
            //Refresh();
        }

        
        /// <summary>
        /// Method for deleting of CP.
        /// </summary>
        private void button1_Click(object sender, EventArgs e)
        {
            if (selected != null && !mousedwn)
            {
                if (selected.left != null && selected.right != null)
                {
                    selected.left.right = selected.right;
                    selected.right.left = selected.left;
                    selected = null;

                    Repaint();
                }
            }
        }

        /// <summary>
        /// Method for altering of CP color.
        /// </summary>
        private void button2_Click(object sender, EventArgs e)
        {
            if (selected != null)
            {
                colorDialog1.Color = selected.color;
                if (colorDialog1.ShowDialog() == DialogResult.OK)
                {
                    selected.color = colorDialog1.Color;
                    Repaint();
                }
            }
        }

        /// <summary>
        /// Saving configuration of CP.
        /// </summary>
        private void button3_Click(object sender, EventArgs e)
        {
            //WriteConfig();
        }

        /// <summary>
        /// Loading configuration of CP.
        /// </summary>
        private void button4_Click(object sender, EventArgs e)
        {
            //ReadConfig();
        }

        
        /// <summary>
        /// Method for chosing of color transtion.
        /// </summary>
        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (img != null)
            {
                Function = comboBox1.SelectedIndex;
                Repaint();
            }
        }

        /// <summary>
        /// Method for reseting of palettes CP.
        /// </summary>
        private void button3_Click_1(object sender, EventArgs e)
        {
            ColorPoint l = new ColorPoint(pictureBox1.Width - 2, pictureBox1.Height / 2, Color.Red, null, null);
            ColorPoint f = new ColorPoint(0, pictureBox1.Height / 2, Color.Blue, null, l);

            l.left = f;

            colorPointList = new ColorPointList(f, l);

            Repaint();

        }

        private void button4_Click_1(object sender, EventArgs e)
        {
            saveFileDialog1.ShowDialog();
            if (saveFileDialog1.FileName != null)
            {
                XmlDocument doc = new XmlDocument();

                XmlElement main = doc.CreateElement("Points");
                doc.AppendChild(main);

                ColorPoint pom = colorPointList.First;

                while (pom != null)
                {
                    XmlElement ec = doc.CreateElement("Point");
                    ec.InnerText = pom.x.ToString(Globals.nfi) + ", " + pom.y.ToString(Globals.nfi) + ", " + pom.color.R + ", " + pom.color.G + ", " + pom.color.B;
                    main.AppendChild(ec);
                    pom = pom.right;
                }

                doc.Save(saveFileDialog1.FileName);
            }
        }

        private void button5_Click(object sender, EventArgs e)
        {
            openFileDialog1.ShowDialog();
            if (openFileDialog1.FileName != null)
            {
                XmlTextReader r = new XmlTextReader(openFileDialog1.FileName);

                XmlDocument doc = new XmlDocument();

                doc.Load(openFileDialog1.FileName);

                XmlNodeList c = doc.GetElementsByTagName("Point");

                if (c.Count >= 2)
                {
                    ColorPoint b = null, first = null;

                    for (int i = 0; i < c.Count; i++)
                    {
                        XmlNode cn = c[i];

                        string[] line = cn.InnerText.Split(',');

                        ColorPoint n = new ColorPoint(double.Parse(line[0], Globals.nfi), double.Parse(line[1], Globals.nfi), Color.FromArgb(int.Parse(line[2]), int.Parse(line[3]), int.Parse(line[4])), null, null);

                        if (i != 0)
                        {
                            n.left = b;
                            b.right = n;
                        }
                        else
                        {
                            first = n;
                        }

                        b = n;
                    }

                    colorPointList = new ColorPointList(first, b);
                }

                r.Close();
            }
        }
    }
}