using System;
using System.IO;
using System.Threading;
using System.ComponentModel;

using Zcu.Mve.Core;
using Zcu.Mve.Strings;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

using openDicom.Registry;
using openDicom.File;
using openDicom.Image;

namespace Zcu.Mve.Visualization.Loaders
{
    /// <summary>
    /// Module loads set of DICOM images (.dcm) from selected directory.
    /// </summary>
    [ModuleInfo("Ji Zapletal", "Module loads set of DICOM images (.dcm) from selected directory.", CategoryType.Loader, CategoryData.Volume, IconName = "DICOM_Loader.ico")]
    [PortInfo("DICOMDirectory", "Directory containing DICOM images.")]
    [PortInfo("OutRegGridND", "Loaded data.")]
    public class DICOM_Loader : Zcu.Mve.Core.Module
    {
        private string pointAttributName = "Values";
        private string path = Directory.GetCurrentDirectory();
        private bool firstRun = true;
        private bool paramsChanged = true;
        private int bits = 0, bytesCount = 0, width = 0, height = 0;
        private double minValue = double.MaxValue, maxValue = double.MinValue;
        private MveArray data = null;
        private RegGridND outData = null;
        string[] dicomFiles = null;
        DataElementDictionary dataElementDictionary;
        UidDictionary uidDictionary;

        decimal sliceThickness = 1;
        decimal[] pixelSpacing = new decimal[2] { 1, 1 };

        /// <summary>
        /// Port management.
        /// </summary>
        public DICOM_Loader()
        {
            AddInPort("DICOMDirectory", typeof(MveString), false);
            AddOutPort("OutRegGridND", typeof(RegGridND));
        }

        /// <summary>
        /// Execution of the module.
        /// </summary>
        public override void Execute()
        {
            this.ProgressInfo(0f);
            if (!IsAllDataSame() || this.paramsChanged) // if I have to compute again 
            {
                LoadDicomFiles();
                this.paramsChanged = false;
                this.SetOutput("OutRegGridND", this.outData);
            }
            else
            {
                this.SetOutput("OutRegGridND", this.outData, true);

                GC.Collect();
                GC.WaitForPendingFinalizers();
                GC.Collect();
                Thread.Sleep(10);
            }
        }

        /// <summary>
        /// Initialize method, default values are set.
        /// </summary>
        private void Initialize()
        {
            paramsChanged = false;
            firstRun = true;

            minValue = double.MaxValue; maxValue = double.MinValue;

            if (this.GetInput("DICOMDirectory") != null)
                path = (MveString)this.GetInput("DICOMDirectory");
        }

        /// <summary>
        /// Loads DICOM files from selected directory.
        /// </summary>
        private void LoadDicomFiles()
        {
            Console.Write("Loading DICOM files...");
            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();
            Thread.Sleep(10);

            Initialize();

            // gets path to directory containing *.dcm files
            dicomFiles = Directory.GetFiles(path, "*.dcm");

            // reads dictionaries containing DICOM tags
            dataElementDictionary = new DataElementDictionary();
            uidDictionary = new UidDictionary();
            try
            {
                dataElementDictionary.LoadFrom("libs\\dicom-elements-2007.dic", DictionaryFileFormat.BinaryFile);
                uidDictionary.LoadFrom("libs\\dicom-uids-2007.dic", DictionaryFileFormat.BinaryFile);
            }
            catch (Exception dictionaryException)
            {
                throw new MveException(this, "Problems processing dictionaries." + dictionaryException);
            }

            // read DICOM files
            AcrNemaFile file = null;
            int counter = 0, time = Environment.TickCount;
            foreach (string df in dicomFiles)
            {
                try
                {
                    if (DicomFile.IsDicomFile(df))
                        file = new DicomFile(df, false);
                    else if (AcrNemaFile.IsAcrNemaFile(df))
                        file = new AcrNemaFile(df, false);
                    else
                        Console.Error.WriteLine("Selected file is wether a DICOM nor an ACR-NEMA file.");

                    // loads appropriate data from DICOM file on the disk
                    LoadData(file, counter);
                    counter++;
                }
                catch (Exception dicomFileException)
                {
                    throw new MveException(this, "Problems processing DICOM file." + dicomFileException);
                }
                this.ProgressInfo(((counter + 1) / (float)dicomFiles.Length));
            }
            Console.WriteLine("\tDone ({0} ms)", Environment.TickCount - time);

            // recalculate dimensions of the whole set
            decimal minSpacing = Math.Min(Math.Min(pixelSpacing[0], pixelSpacing[1]), sliceThickness);
            double[] mins = new double[] { 0, 0, 0 };
            double[] maxs = new double[] { (double)(minSpacing / pixelSpacing[0]), (double)(minSpacing / sliceThickness), (double)(minSpacing / pixelSpacing[1]) };

            // creates dataset
            this.outData = new RegGridND(mins, maxs, width - 1, dicomFiles.Length - 1, height - 1);
            this.outData.AddPointAttrs(this.pointAttributName, data);

            // add some tags
            this.outData.UserTags.Add("FileName", (MveString)path);                        
            this.outData.UserTags.Add("Min", (Scalar)minValue);
            this.outData.UserTags.Add("Max", (Scalar)maxValue);
            this.outData.UserTags.Add("Position", new Point3D(0, 0, 0));
            this.outData.UserTags.Add("Rotation", new Point3D(0, 0, 0));

            // clear memory
            data = null;
            dataElementDictionary = null;
            uidDictionary = null;

            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();
            Thread.Sleep(10);

            this.ProgressInfo(1f);
        }

        /// <summary>
        /// Loads appropriate data from each DICOM file
        /// </summary>
        /// <param name="img">Instance of representation of the file form the disk.</param>
        /// <param name="slice">Current number of the slice.</param>
        private void LoadData(AcrNemaFile img, int slice)
        {
            PixelData pd = img.PixelData;

            // read important (meta)data from DICOM files that are the same for all slices
            if (firstRun)
            {
                bits = img.PixelData.HighBit + 1;
                bytesCount = (int)Math.Ceiling(bits / 8.0);
                width = pd.Columns;
                height = pd.Rows;
                data = new MveArray(typeof(Integer), width, dicomFiles.Length, height);
                openDicom.DataStructure.Tag sliceThicknessTag = new openDicom.DataStructure.Tag(0x0018, 0x0050);
                openDicom.DataStructure.Tag pixelSpacingTag = new openDicom.DataStructure.Tag(0x0028, 0x0030);
                sliceThickness = (decimal)img.DataSet[sliceThicknessTag].Value[0];
                pixelSpacing = new decimal[2] { (decimal)img.DataSet[pixelSpacingTag].Value[0], (decimal)img.DataSet[pixelSpacingTag].Value[1] };

                firstRun = false;
            }

            // if the DICOM file contains really an image
            if (img.HasPixelData)
            {

                #region SLOW (Old version without pointer utilisation)
                /* byteArray = pd.ToBytesArray()[0];
                   int row, column, pixelCounter = 0;
                  for (int i = 0; i < byteArray.Length; i += bytesCount)
                  {
                      row = pixelCounter / width;
                      column = pixelCounter % width;

                      switch (bytesCount)
                      {
                          case 2:
                              value = BitConverter.ToInt16(byteArray, i);
                              break;

                          case 4: value = BitConverter.ToInt32(byteArray, i);
                              break;

                          default: value = byteArray[i];
                              break;
                      }

                      if (value < minValue)
                          minValue = value;
                      if (value > maxValue)
                          maxValue = value;

                      if (value != 0)
                          value = value;

                      this.data[column, slice, row] = (Integer)value;
                      pixelCounter++;
                  }//*/
                #endregion

                // switch loading values due to data type representing each value
                switch (bytesCount)
                {
                    case 2:
                        LoadAsShorts(pd, slice);
                        break;

                    case 4: LoadAsInts(pd, slice);
                        break;

                    default: LoadAsBytes(pd, slice);
                        break;
                }

                GC.Collect();
                GC.WaitForPendingFinalizers();
                GC.Collect();
                Thread.Sleep(10);
            }
            else throw new ArgumentException("No pixel data.");
        }

        /// <summary>
        /// Load data values represented by bytes.
        /// </summary>
        /// <param name="pd">Image data from DICOM file.</param>
        /// <param name="slice">Current number of the slice.</param>
        private void LoadAsBytes(PixelData pd, int slice)
        {
            byte[] byteArray = pd.ToBytesArray()[0];
            int value, index = 0, sliceHeight = slice * height, heightDicomFiles = height * dicomFiles.Length;
            Integer[, ,] d = (Integer[, ,])data.Array;

            unsafe
            {
                fixed (Integer* dataPtr = &d[0, 0, 0])
                {
                    for (int i = 0; i < byteArray.Length; i += bytesCount)
                    {
                        value = byteArray[i];

                        if (value < minValue)
                            minValue = value;
                        if (value > maxValue)
                            maxValue = value;

                        dataPtr[index / width + sliceHeight + heightDicomFiles * (index % width)] = (Integer)value;
                        index++;
                    }
                }
            }
        }

        /// <summary>
        /// Load data values represented by shorts.
        /// </summary>
        /// <param name="pd">Image data from DICOM file.</param>
        /// <param name="slice">Current number of the slice.</param>
        private void LoadAsShorts(PixelData pd, int slice)
        {
            byte[] byteArray = pd.ToBytesArray()[0];
            int value, index = 0, sliceHeight = slice * height, heightDicomFiles = height * dicomFiles.Length;
            Integer[, ,] d = (Integer[, ,])data.Array;

            unsafe
            {
                fixed (Integer* dataPtr = &d[0, 0, 0])
                {
                    for (int i = 0; i < byteArray.Length; i += bytesCount)
                    {
                        value = BitConverter.ToInt16(byteArray, i);

                        if (value < minValue)
                            minValue = value;
                        if (value > maxValue)
                            maxValue = value;

                        dataPtr[index / width + sliceHeight + heightDicomFiles * (index % width)] = (Integer)value;
                        index++;
                    }
                }
            }
        }

        /// <summary>
        /// Load data values represented by integers.
        /// </summary>
        /// <param name="pd">Image data from DICOM file.</param>
        /// <param name="slice">Current number of the slice.</param>
        private void LoadAsInts(PixelData pd, int slice)
        {
            byte[] byteArray = pd.ToBytesArray()[0];
            int value, index = 0, sliceHeight = slice * height, heightDicomFiles = height * dicomFiles.Length;
            Integer[, ,] d = (Integer[, ,])data.Array;

            unsafe
            {
                fixed (Integer* dataPtr = &d[0, 0, 0])
                {
                    for (int i = 0; i < byteArray.Length; i += bytesCount)
                    {
                        value = BitConverter.ToInt32(byteArray, i);

                        if (value < minValue)
                            minValue = value;
                        if (value > maxValue)
                            maxValue = value;

                        dataPtr[index / width + sliceHeight + heightDicomFiles * (index % width)] = (Integer)value;
                        index++;
                    }
                }
            }
        }

        #region Properties
        /// <summary>
        /// Path to the DICOM directory.
        /// </summary>
        [Editor(typeof(System.Windows.Forms.Design.FolderNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
        [Description("Path to the DICOM directory.")]
        public string Path
        {
            get { return this.path; }
            set
            {
                this.path = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Set name of gridND point attributes. Default set to Values.
        /// </summary>
        [Description("Set name of gridND point attributes. Default set to \"Values\"")]
        public string PointAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.pointAttributName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.pointAttributName; }
        }
        #endregion

    }
}