using System;
using System.IO;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;

using Zcu.Mve.Core;
using Zcu.Mve.Strings;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

using openDicom.Registry;
using openDicom.File;
using openDicom.Image;

namespace Zcu.Mve.Visualization.Renderers
{
    /// <summary>
    /// Data type enumeration.
    /// </summary>
    public enum ValueDataType
    {
        /// <summary>integer number</summary>
        Integer,
        /// <summary>floating point number</summary>
        Scalar
    };

    /// <summary>
    /// Axis-aligned planes
    /// </summary>
    public enum Plane
    {
        /// <summary>YZ plane</summary>
        YZ,
        /// <summary>XZ plane</summary>
        XZ,
        /// <summary>XY plane</summary>
        XY
    } ;

    /// <summary>
    /// Module renders volume data as separate slices.\n\nComment: Pixel values are normalized to 0-255 interval within all the dataset (i.e. not only each one image independently to the others in the dataset.).
    /// </summary>
    [ModuleInfo("Ji Zapletal", "Module renders volume data as separate slices.\n\nComment: Pixel values are normalized to 0-255 interval within all the dataset (i.e. not only each one image independently to the others in the dataset.).", CategoryType.Output, CategoryData.Volume, IconName = "VolumeSliceViewer.ico")]
    [PortInfo("InVolumeData", "Input volume data to be viewed.")]
    public class VolumeSliceViewer : Zcu.Mve.Core.Module
    {
        private bool findMinMax = false, paramsChanged = true;
        private int[] size = null;

        private double[, ,] dataArrayDouble = null;
        private int[, ,] dataArrayInt = null;
        private Rectangle rect;
        private ValueDataType valueDataType;

        /// <summary>zoom</summary>
        public float zoom = 0;
        /// <summary>inversion indicator</summary>
        public bool invert = false;

        /// <summary>number of slices</summary>
        public int slices = 0;
        /// <summary>width</summary>
        public int width = 0;
        /// <summary>height</summary>
        public int height = 0;
        /// <summary>image index</summary>
        public int imageIndex = 0;
        /// <summary>scale</summary>
        public int scale = 1;

        /// <summary>minimum value</summary>
        public double minValue = double.MaxValue;
        /// <summary>maximum value</summary>
        public double maxValue = double.MinValue;
        /// <summary>interval of the inversed values</summary>
        public double inverseDataValuesInterval = 0;

        /// <summary>name of the point attribute</summary>
        public string pointAttributName = "Values";
        /// <summary>name of the dataset</summary>
        public string datasetName = "";

        /// <summary>selected plane</summary>
        public Plane plane = Plane.XY;
        /// <summary>image</summary>
        public Bitmap bmp = null;

        /// <summary>
        /// Port management
        /// </summary>
        public VolumeSliceViewer()
        {
            AddInPort("InVolumeData", typeof(RegGridND));
        }

        /// <summary>
        /// Execution of the module.
        /// </summary>
        public override void Execute()
        {
            this.ProgressInfo(0f);

            if (!IsAllDataSame() || this.paramsChanged) // if I have to compute again 
                ViewImage();
        }

        /// <summary>
        /// View slices.
        /// </summary>
        private void ViewImage()
        {
            slices = 0; width = 0; height = 0; inverseDataValuesInterval = 0;
            minValue = double.MaxValue; maxValue = double.MinValue;
            dataArrayDouble = null; dataArrayInt = null;
            bmp = null;

            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();
            Thread.Sleep(10);

            RegGridND data = (RegGridND)this.GetInput("InVolumeData");
            if (data == null)
                throw new ArgumentException("Passed variable 'data' cannot be null.");

            if (data.UserTags.ContainsKey("Min") && data.UserTags.ContainsKey("Max"))
            {
                minValue = (Scalar)data.UserTags["Min"];
                maxValue = (Scalar)data.UserTags["Max"];
            }
            else
                findMinMax = true;

            if (data.UserTags.ContainsKey("FileName"))
            {
                datasetName = (MveString)data.UserTags["FileName"];
                if (datasetName.Contains("\\"))
                    datasetName = datasetName.Substring(datasetName.LastIndexOf('\\') + 1);
                else if (datasetName.Contains("/"))
                    datasetName = datasetName.Substring(datasetName.LastIndexOf('/') + 1);
            }
            else
                datasetName = "-- N/A --";

            MveArray dataArray = data.GetPointAttrs(this.pointAttributName);
            size = dataArray.Lengths;
            width = size[0]; slices = size[1]; height = size[2];

            #region Convert data from MVE data types to .NET data types for faster processing
            Console.Write("Converting data...");
            int ticks = Environment.TickCount;
            float progress = 0.9f / width;
            int index = 0;


            if (dataArray.ElemType == typeof(Zcu.Mve.Numerics.Integer))
            {
                valueDataType = ValueDataType.Integer;
                dataArrayInt = new int[width, height, slices];

                unsafe
                {
                    fixed (int* dataPtr = &dataArrayInt[0, 0, 0])
                    {
                        for (int i = 0; i < width; i++)
                        {
                            for (int j = 0; j < height; j++)
                                for (int k = 0; k < slices; k++)
                                {
                                    dataPtr[index] = (Integer)dataArray[i, k, j];
                                    if (findMinMax)
                                    {
                                        if (dataPtr[index] <= minValue)
                                            minValue = dataPtr[index];
                                        if (dataPtr[index] >= maxValue)
                                            maxValue = dataPtr[index];
                                    }
                                    index++;
                                }
                            this.ProgressInfo((i + 1) * progress);
                        }
                    }
                }
            }
            else if (dataArray.ElemType == typeof(Zcu.Mve.Numerics.Scalar))
            {
                valueDataType = ValueDataType.Scalar;
                dataArrayDouble = new double[width, height, slices];

                unsafe
                {
                    fixed (double* dataPtr = &dataArrayDouble[0, 0, 0])
                    {
                        for (int i = 0; i < width; i++)
                        {
                            for (int j = 0; j < height; j++)
                                for (int k = 0; k < slices; k++)
                                {
                                    dataPtr[index] = (Scalar)dataArray[i, k, j];
                                    if (findMinMax)
                                    {
                                        if (dataPtr[index] <= minValue)
                                            minValue = dataPtr[index];
                                        if (dataPtr[index] >= maxValue)
                                            maxValue = dataPtr[index];
                                    }
                                    index++;
                                }
                            this.ProgressInfo((i + 1) * progress);
                        }
                    }
                }
            }

            Console.WriteLine("\t\tDone ({0} ms)", Environment.TickCount - ticks);
            #endregion

            // clean memory
            dataArray = null;
            data = null;
            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();
            Thread.Sleep(10);

            inverseDataValuesInterval = 255 / ((double)(maxValue - minValue));

            bmp = RegGridNDtoBitmap();
            ImageForm form = new ImageForm(this);
            form.ShowDialog();

            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();
            Thread.Sleep(10);
            this.ProgressInfo(1f);
        }

        /// <summary>
        /// Convert appropriate slice of the dataset to bitmap.
        /// </summary>
        /// <returns>Bitmap of the slice.</returns>
        public Bitmap RegGridNDtoBitmap()
        {
            if (scale > 0)
                zoom = scale;
            else if (scale < 0)
                zoom = 1f / -scale;

            if (plane == Plane.YZ)
                bmp = new Bitmap((int)(height * zoom), (int)(slices * zoom));
            else if (plane == Plane.XZ)
                bmp = new Bitmap((int)(width * zoom), (int)(slices * zoom));
            else if (plane == Plane.XY)
                bmp = new Bitmap((int)(width * zoom), (int)(height * zoom));

            rect = new Rectangle(0, 0, bmp.Width, bmp.Height);
            BitmapData bmpData = bmp.LockBits(rect, ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb);

            if (valueDataType == ValueDataType.Integer)
            {
                #region Converting Integer input data
                #region Zoom in
                if (zoom >= 1) // zoom in
                {
                    if (plane == Plane.YZ)
                    {
                        for (int i = 0; i < height; i++)
                            for (int j = 0; j < slices; j++)
                                for (int k = 0; k < scale; k++)
                                    for (int l = 0; l < scale; l++)
                                        // get value of the pixel from appropriate slice, defined by imageIndex                     
                                        SetPixelPtr(bmpData, i * scale + k, j * scale + l,
                                            RemapValueToBrightnessI(dataArrayInt[imageIndex, i, j]));
                    }
                    else if (plane == Plane.XZ)
                    {
                        for (int i = 0; i < width; i++)
                            for (int j = 0; j < slices; j++)
                                for (int k = 0; k < scale; k++)
                                    for (int l = 0; l < scale; l++)
                                        // get value of the pixel from appropriate slice, defined by imageIndex                     
                                        SetPixelPtr(bmpData, i * scale + k, j * scale + l,
                                            RemapValueToBrightnessI(dataArrayInt[i, imageIndex, j]));
                    }
                    else if (plane == Plane.XY)
                    {
                        for (int i = 0; i < width; i++)
                            for (int j = 0; j < height; j++)
                                for (int k = 0; k < scale; k++)
                                    for (int l = 0; l < scale; l++)
                                        // get value of the pixel from appropriate slice, defined by imageIndex
                                        SetPixelPtr(bmpData, i * scale + k, j * scale + l,
                                            RemapValueToBrightnessI(dataArrayInt[i, j, imageIndex]));
                    }
                }
                #endregion
                #region Zoom out
                else
                {
                    if (plane == Plane.YZ)
                    {
                        for (int i = 0; i < height / Math.Abs(-scale); i++)
                            for (int j = 0; j < slices / Math.Abs(-scale); j++)
                                // get value of the pixel from appropriate slice, defined by imageIndex                     
                                SetPixelPtr(bmpData, i, j,
                                    RemapValueToBrightnessI(dataArrayInt[imageIndex, i * Math.Abs(-scale), j * Math.Abs(-scale)]));
                    }
                    else if (plane == Plane.XZ)
                    {
                        for (int i = 0; i < width / Math.Abs(-scale); i++)
                            for (int j = 0; j < slices / Math.Abs(-scale); j++)
                                // get value of the pixel from appropriate slice, defined by imageIndex                     
                                SetPixelPtr(bmpData, i, j,
                                    RemapValueToBrightnessI(dataArrayInt[i * Math.Abs(-scale), imageIndex, j * Math.Abs(-scale)]));
                    }
                    else if (plane == Plane.XY)
                    {
                        for (int i = 0; i < width / Math.Abs(-scale); i++)
                            for (int j = 0; j < height / Math.Abs(-scale); j++)
                                // get value of the pixel from appropriate slice, defined by imageIndex
                                SetPixelPtr(bmpData, i, j,
                                    RemapValueToBrightnessI(dataArrayInt[i * Math.Abs(-scale), j * Math.Abs(-scale), imageIndex]));
                    }
                }
                #endregion
                #endregion
            }
            else if (valueDataType == ValueDataType.Scalar)
            {
                #region Converting Scalar input data
                #region Zoom in
                if (zoom >= 1)
                {
                    if (plane == Plane.YZ)
                    {
                        for (int i = 0; i < height; i++)
                            for (int j = 0; j < slices; j++)
                                for (int k = 0; k < scale; k++)
                                    for (int l = 0; l < scale; l++)
                                        // get value of the pixel from appropriate slice, defined by imageIndex                                                 
                                        SetPixelPtr(bmpData, i * scale + k, j * scale + l,
                                            RemapValueToBrightnessD(dataArrayDouble[imageIndex, i, j]));
                    }
                    else if (plane == Plane.XZ)
                    {
                        for (int i = 0; i < width; i++)
                            for (int j = 0; j < slices; j++)
                                for (int k = 0; k < scale; k++)
                                    for (int l = 0; l < scale; l++)
                                        // get value of the pixel from appropriate slice, defined by imageIndex                                                         
                                        SetPixelPtr(bmpData, i * scale + k, j * scale + l,
                                            RemapValueToBrightnessD(dataArrayDouble[i, imageIndex, j]));
                    }
                    else if (plane == Plane.XY)
                    {
                        for (int i = 0; i < width; i++)
                            for (int j = 0; j < height; j++)
                                for (int k = 0; k < scale; k++)
                                    for (int l = 0; l < scale; l++)
                                        // get value of the pixel from appropriate slice, defined by imageIndex                                    
                                        SetPixelPtr(bmpData, i * scale + k, j * scale + l,
                                            RemapValueToBrightnessD(dataArrayDouble[i, j, imageIndex]));
                    }
                }
                #endregion
                #region Zoom out
                else
                {
                    if (plane == Plane.YZ)
                    {
                        for (int i = 0; i < height / Math.Abs(-scale); i++)
                            for (int j = 0; j < slices / Math.Abs(-scale); j++)
                                // get value of the pixel from appropriate slice, defined by imageIndex                     
                                SetPixelPtr(bmpData, i, j,
                                    RemapValueToBrightnessD(dataArrayDouble[imageIndex, i * Math.Abs(-scale), j * Math.Abs(-scale)]));
                    }
                    else if (plane == Plane.XZ)
                    {
                        for (int i = 0; i < width / Math.Abs(-scale); i++)
                            for (int j = 0; j < slices / Math.Abs(-scale); j++)
                                // get value of the pixel from appropriate slice, defined by imageIndex                     
                                SetPixelPtr(bmpData, i, j,
                                    RemapValueToBrightnessD(dataArrayDouble[i * Math.Abs(-scale), imageIndex, j * Math.Abs(-scale)]));
                    }
                    else if (plane == Plane.XY)
                    {
                        for (int i = 0; i < width / Math.Abs(-scale); i++)
                            for (int j = 0; j < height / Math.Abs(-scale); j++)
                                // get value of the pixel from appropriate slice, defined by imageIndex
                                SetPixelPtr(bmpData, i, j,
                                    RemapValueToBrightnessD(dataArrayDouble[i * Math.Abs(-scale), j * Math.Abs(-scale), imageIndex]));
                    }
                }
                #endregion
                #endregion
            }
            bmp.UnlockBits(bmpData);

            return bmp;
        }

        /// <summary>
        /// Remap integer value from dataset to real displayable value.
        /// </summary>
        /// <param name="value">Integer value from dataset.</param>
        /// <returns>Displayable value.</returns>
        private byte RemapValueToBrightnessI(int value)
        {
            byte brightness = (byte)((value - minValue) * inverseDataValuesInterval);
            return invert ? (byte)(255 - brightness) : brightness;
        }

        /// <summary>
        /// Remap double value from dataset to real displayable value.
        /// </summary>
        /// <param name="value">Double value from dataset.</param>
        /// <returns>Displayable value.</returns>
        private byte RemapValueToBrightnessD(double value)
        {
            byte brightness = (byte)((value - minValue) * inverseDataValuesInterval);
            return invert ? (byte)(255 - brightness) : brightness;
        }

        /// <summary>
        /// Set pixel via pointer (fast).
        /// </summary>
        /// <param name="bmData">Data from bitmap.</param>
        /// <param name="x">x-coordinate of the pixel.</param>
        /// <param name="y">y-coordinate of the pixel.</param>
        /// <param name="brightness">Value to set on coordinates.</param>
        private void SetPixelPtr(BitmapData bmData, int x, int y, byte brightness)
        {
            int bpp = 3;
            unsafe
            {
                byte* b = (byte*)bmData.Scan0 + (y * bmData.Stride) + (x * bpp);
                b[0] = brightness;
                b[1] = brightness;
                b[2] = brightness;
            }
        }

        #region Properties
        /// <summary>
        /// Cell attribute name.
        /// </summary>
        [Description("Cell attribute name.")]
        public string PointAttributName
        {
            get { return this.pointAttributName; }
            set
            {
                this.pointAttributName = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Inverted image?
        /// </summary>
        [Description("Inverted image?")]
        public bool Invert
        {
            get { return this.invert; }
            set
            {
                this.invert = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Plane to be sliced.
        /// </summary>
        [Description("Plane to be sliced.")]
        public Plane Plane
        {
            get { return this.plane; }
            set
            {
                this.plane = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Zoom factor of the viewed plane. Acceptable values for zoom:\n - OUT are smaller or equall to -2, defining how many times the plane will be reduced (e.g. -2 results in 1/2x = 50% of the original size, -3 in 1/3x = 33.333% size etc.).\n\n- IN are greater or equall to 1, defining how many times the plane will be magnified (e.g. 1 results in 1x = 100% = original size, 2 in 2x = 200% size etc.).
        /// </summary>
        [Description("Zoom factor of the viewed plane. Acceptable values for zoom:\n - OUT are smaller or equall to -2, defining how many times the plane will be reduced (e.g. -2 results in 1/2x = 50% of the original size, -3 in 1/3x = 33.333% size etc.).\n\n- IN are greater or equall to 1, defining how many times the plane will be magnified (e.g. 1 results in 1x = 100% = original size, 2 in 2x = 200% size etc.).")]
        public int ZoomFactor
        {
            get { return this.scale; }
            set
            {
                if ((value > -2) && (value < 1))
                {
                    this.scale = 1;
                    this.paramsChanged = true;
                }
                else
                {
                    this.scale = value;
                    this.paramsChanged = true;
                }
            }
        }

        /// <summary>
        /// Index of the slice in the selected plane (1 stands for the first slice). If the index exceeds number of the slices in the current plane, first slice will be selected.
        /// </summary>
        [Description("Index of the slice in the selected plane (1 stands for the first slice). If the index exceeds number of the slices in the current plane, first slice will be selected.")]
        public int Index
        {
            get { return this.imageIndex + 1; }
            set
            {
                this.imageIndex = 0;

                if (value > 0)
                {
                    if ((this.plane == Plane.YZ) && (value <= this.width))
                        this.imageIndex = value - 1;
                    else if ((this.plane == Plane.XZ) && (value <= this.height))
                        this.imageIndex = value - 1;
                    else if ((this.plane == Plane.XY) && (value <= this.slices))
                        this.imageIndex = value - 1;
                }

                this.paramsChanged = true;
            }
        }

        #endregion
    }
}