/*
 *		Jiri Skala		A01216
 *
 *		KIV / GSVD		2004/05
 */
using System;
using System.Xml;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
	/// <summary>
	/// Available cell color processing functions.
	/// </summary>
	internal enum ColorFunction {Gray, Red, Green, Blue};

	/// <summary>
	/// <p>Assigns scalar attributes to regular 2D grid points according to
	/// color attributes of adjacent cells. Scalar attributes will be in [0; 1] range.
	/// Supports alpha channel and allows RGB color channels masking.</p>
	/// <p><b>How it works:</b><br />
	/// Point colors are computed as a bilinear interpolation of adjacent cell colors.
	/// Module can process cell color intensity or one of RGB color components.</p>
	/// </summary>
	[ModuleInfo("Ji Skla", "Assigns scalar attributes to regular 2D grid points "
		+ "according to color attributes of adjacent cells. Scalar attributes "
		+ "will be in [0; 1] range. "
		+ "Supports alpha channel and allows RGB color channels masking.",
        CategoryType.Filter, CategoryData.Image,
		IconName = "CalculateDisplace.ico", Assemblied = "2005-05-12")]
	[PortInfo("InputGrid", "Regular 2D grid with cell color attributes. Typically a bitmap image.")]
	[PortInfo("OutputGrid", "Regular 2D grid with point color attributes.")]
	public class CalculateDisplace : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Color processing function delegate declaration.
		/// </summary>
		private delegate double ColorIntensityFunction(ColorRGBA color);

		/// <summary>
		/// Color processing function delegate instance.
		/// </summary>
		private ColorIntensityFunction Intensity;

		/// <summary>
		/// Color processing function name.
		/// </summary>
		private ColorFunction colorFunction;

		/// <summary>
		/// Regular 2D grid that is available at the module output port.
		/// </summary>
		private RegGrid2D outputGrid;

		/// <summary>
		/// Elevation bitmap cell color attribute name.
		/// </summary>
		private string colorAttrName;

		/// <summary>
		/// Output regular 2D grid point elevation attribute name.
		/// </summary>
		private string scalarAttrName;

		/// <summary>
		/// Determines whether elevation bitmap
		/// cell color attribute name is set to default.
		/// </summary>
		private bool colorAttrDefault;

		/// <summary>
		/// Determines whether output regular 2D grid
		/// point elevation attribute name is set to default.
		/// </summary>
		private bool scalarAttrDefault;

		/// <summary>
		/// Constructor. Creates input port for elevation bitmap
		/// and output port for regular 2D grid.
		/// Initializes internal state of this module.
		/// </summary>
		public CalculateDisplace()
		{
			// initialize color processing function
			Intensity = new ColorIntensityFunction(GetGrayIntensity);
			colorFunction = ColorFunction.Gray;

			// initialize attribute names
			colorAttrDefault = true;
			colorAttrName = typeof(ColorRGBA).Name;
			scalarAttrDefault = true;
			scalarAttrName = typeof(Scalar).Name;

			// there is no result from previous computation
			outputGrid = null;

			// add module ports
			AddInPort("InputGrid", typeof(RegGrid2D));
			AddOutPort("OutputGrid", typeof(RegGrid2D));
		}

		/// <summary>
		/// Creates a deep copy of this instance.
		/// </summary>
		/// <returns>Returns a deep copy of this instance.</returns>
		public override Module DeepCopy()
		{
			// create a new instance
			CalculateDisplace copy = (CalculateDisplace)base.DeepCopy();

			// set to the same state as this
			copy.SelectedColorFunction = this.colorFunction;	// also sets function delegate
			copy.SetColorAttrName(this.colorAttrName, this.colorAttrDefault);
			copy.SetScalarAttrName(this.scalarAttrName, this.scalarAttrDefault);

			return copy;
		}

		/// <summary>
		/// Gets color attribute name and specifies whether default name should be used.
		/// </summary>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		/// <returns>Returns color attribute name.</returns>
		internal string GetColorAttrName(out bool isDefault)
		{
			isDefault = colorAttrDefault;
			return colorAttrName;
		}

		/// <summary>
		/// Sets color attribute name to user or default value.
		/// </summary>
		/// <remarks>If default name should be set,
		/// <i>name</i> parameter is ignored.</remarks>
		/// <param name="name">Color attribute name.</param>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		internal void SetColorAttrName(string name, bool isDefault)
		{
			colorAttrDefault = isDefault;
			if (isDefault)
				// set default value
				colorAttrName = typeof(ColorRGBA).Name;
			else
				// set user value
				colorAttrName = name;
		}

		/// <summary>
		/// Gets scalar attribute name and specifies whether default name should be used.
		/// </summary>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		/// <returns>Returns scalar attribute name.</returns>
		internal string GetScalarAttrName(out bool isDefault)
		{
			isDefault = scalarAttrDefault;
			return scalarAttrName;
		}

		/// <summary>
		/// Sets scalar attribute name to user or default value.
		/// </summary>
		/// <remarks>If default name should be set,
		/// <i>name</i> parameter is ignored.</remarks>
		/// <param name="name">Scalar attribute name.</param>
		/// <param name="isDefault">Specifies whether default name should be used.</param>
		internal void SetScalarAttrName(string name, bool isDefault)
		{
			scalarAttrDefault = isDefault;
			if (isDefault)
				// set default value
				scalarAttrName = typeof(Scalar).Name;
			else
				// set user value
				scalarAttrName = name;
		}

		/// <summary>
		/// Gets or sets color processing function.
		/// </summary>
		internal ColorFunction SelectedColorFunction
		{
			get { return colorFunction; }

			set
			{
				// store color processing name
				colorFunction = value;
				// determine and chose appropriate function
				switch (value)
				{
					case ColorFunction.Gray:
						Intensity = new ColorIntensityFunction(GetGrayIntensity);
						break;

					case ColorFunction.Red:
						Intensity = new ColorIntensityFunction(GetRedIntensity);
						break;

					case ColorFunction.Green:
						Intensity = new ColorIntensityFunction(GetGreenIntensity);
						break;

					case ColorFunction.Blue:
						Intensity = new ColorIntensityFunction(GetBlueIntensity);
						break;
				}
			}
		}

		/// <summary>
		/// Saves current module settings.
		/// </summary>
		/// <param name="config">XML element where to save.</param>
		public override void WriteConfig(XmlElement config)
		{
			// save color processing function
			XmlElement colorFunc = config.OwnerDocument.CreateElement("colorFunction");
			switch (colorFunction)
			{
				case ColorFunction.Gray:
					colorFunc.InnerText = "Gray";
					break;

				case ColorFunction.Red:
					colorFunc.InnerText = "Red";
					break;

				case ColorFunction.Green:
					colorFunc.InnerText = "Green";
					break;

				case ColorFunction.Blue:
					colorFunc.InnerText = "Blue";
					break;
			}
			config.AppendChild(colorFunc);

			// save color attribute name setting
			XmlElement colorAttr = config.OwnerDocument.CreateElement("colorAttribute");
			colorAttr.SetAttribute("default", colorAttrDefault.ToString());
			colorAttr.InnerText = colorAttrName;
			config.AppendChild(colorAttr);

			// save scalar attribute name setting
			XmlElement scalarAttr = config.OwnerDocument.CreateElement("scalarAttribute");
			scalarAttr.SetAttribute("default", scalarAttrDefault.ToString());
			scalarAttr.InnerText = scalarAttrName;
			config.AppendChild(scalarAttr);
		}

		/// <summary>
		/// Loads module settings.
		/// </summary>
		/// <param name="config">XML element from where to load.</param>
		public override void ReadConfig(XmlElement config)
		{
			// load module setup
			foreach (XmlElement element in config.ChildNodes)
				switch (element.Name)
				{
					case "colorFunction":
						// color processing function definition
						switch (element.InnerText.Trim().ToLower())
						{
							case "gray":
								Intensity = new ColorIntensityFunction(GetGrayIntensity);
								colorFunction = ColorFunction.Gray;
								break;

							case "red":
								Intensity = new ColorIntensityFunction(GetRedIntensity);
								colorFunction = ColorFunction.Red;
								break;

							case "green":
								Intensity = new ColorIntensityFunction(GetGreenIntensity);
								colorFunction = ColorFunction.Green;
								break;

							case "blue":
								Intensity = new ColorIntensityFunction(GetBlueIntensity);
								colorFunction = ColorFunction.Blue;
								break;

							default:
								// unknown function keyword
								throw new MveException(this, "Error reading module configuration. \""
									+ element.InnerText.Trim()
									+ "\" is not recognized as known color processing function.");
								// break;	// unreachable code
						}
						break;

					case "colorAttribute":
						// get "default" attribute
						string attribColor = element.GetAttribute("default");
						// try to parse attribute value
						colorAttrDefault = Parser.BoolValue(attribColor, this);
						// get color attribute name
						colorAttrName = element.InnerText.Trim();
						break;

					case "scalarAttribute":
						// get "default" attribute
						string attribScalar = element.GetAttribute("default");
						// try to parse attribute value
						scalarAttrDefault = Parser.BoolValue(attribScalar, this);
						// get scalar attribute name
						scalarAttrName = element.InnerText.Trim();
						break;

					default:
						// unknown config tag
						Console.WriteLine(this.Name + " warning:");
						Console.WriteLine("\"" + element.Name + "\" is not recognized "
							+ "as known module configuration tag - ignoring.");
						break;
				}	// end of switch (element.Name)
			// end of foreach
		}

		/// <summary>
		/// <p><i>CalculateDisplace</i> module setup dialog.</p>
		/// <img src="CalculateDisplaceSetup.png" />
		/// <p><i>Elevation map color processing</i> determines how cell colors
		/// will be processed. You can select color lightness (also called intensity
		/// or value) or make the module use only one of RGB color components.</p>
		/// <p>In the <i>Attribute names</i> panel you can specify names for input
		/// cell color attributes and for output point elevation attributes.
		/// It is recomended to use default values unless you really need different
		/// attribute names.</p>
		/// </summary>
		/// <remarks>Creates module setup dialog.</remarks>
		/// <returns>Returns module setup dialog reference.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new CalculateDisplaceSetup(this);
		}

		/// <summary>
		/// Assigns scalar attributes to regular 2D grid points
		/// according to color attributes of adjacent cells.
		/// </summary>
		public override void Execute()
		{
			// check whether there is some new input
			if (IsDataSame("InputGrid") && outputGrid != null)
			{
				// publish old output
				SetOutput("OutputGrid", outputGrid, true);
				// no more work
				return;
			}

			// get input elevation bitmap
			RegGrid2D inputGrid = (RegGrid2D)GetInput("InputGrid");

			// make a shallow copy
			outputGrid = new RegGrid2D(inputGrid.Height, inputGrid.Width);
			// copy library names
			foreach (string libName in inputGrid.LibNames)
				outputGrid.AddLibName(libName);
			// copy point attribute references
			foreach (string attribID in inputGrid.GetPointAttrIds())
				outputGrid.AddPointAttrs(attribID, inputGrid.GetPointAttrs(attribID));
			// copy cell attribute references
			foreach (string attribID in inputGrid.GetCellAttrIds())
				outputGrid.AddCellAttrs(attribID, inputGrid.GetCellAttrs(attribID));

			string attribFound;
			// get cell color attributes
			UniformDataArray2D cellColorAttrs = AttributeFinder.FindCellAttrib(
				inputGrid, colorAttrName, typeof(ColorRGBA), out attribFound);

			// check success
			if (cellColorAttrs == null)
				// there are no cell color attributes in input grid
				throw new MveException(this, "Elevation bitmap does not contain any cell color attributes.");

			// check we found what we wanted
			if (attribFound != colorAttrName)
			{
				// show a warning that given attribute has not been found
				Console.WriteLine(this.Name + " warning:");
				Console.WriteLine("\"" + colorAttrName + "\" cell attribute not found or has wrong type.");
				Console.WriteLine("Using \"" + attribFound + "\" attribute instead.");
			}

			// compute point colors
			UniformDataArray2D pointColorAttrs = ComputePointColors(cellColorAttrs);

			// assign attributes to points
			outputGrid.AddPointAttrs(scalarAttrName, pointColorAttrs);

			// publish new output
			SetOutput("OutputGrid", outputGrid, false);
		}

		/// <summary>
		/// Computes point color attributes according to adjacent cell color attributes.
		/// </summary>
		/// <param name="cellColors">Array of cell color attributes.</param>
		/// <returns>Returns array of point color attributes.</returns>
		private UniformDataArray2D ComputePointColors(UniformDataArray2D cellColors)
		{
			// get point grid dimensions
			int width = cellColors.Width + 1;
			int height = cellColors.Height + 1;

			// check there are some data
			// at least one cell means at least 2 points
			if (width < 2 || height < 2)
				// grid is empty, has no cells
				throw new MveException(this, "Elevation bitmap is empty (size "
					+ (width-1).ToString() + "x" + (height-1).ToString() + " pixels).");

			// prepare array for point color attributes
			UniformDataArray2D pointColors = new UniformDataArray2D(typeof(Scalar), height, width);

			// corner points - absolutely special case
			pointColors[0, 0] = new Scalar( Intensity( (ColorRGBA)cellColors[0, 0] ));
			pointColors[height - 1, 0] = new Scalar( Intensity( (ColorRGBA)cellColors[height - 2, 0] ));
			pointColors[0, width - 1] = new Scalar( Intensity( (ColorRGBA)cellColors[0, width - 2] ));
			pointColors[height - 1, width - 1] = new Scalar( Intensity( (ColorRGBA)cellColors[height - 2, width - 2] ));

			// top and bottom edge points (excluding corner points) - special case
			for (int i = 1; i < width - 1; i++)
			{
				// point in the top row
				pointColors[0, i] = new Scalar( lerp(
					Intensity( (ColorRGBA)cellColors[0, i - 1] ),
					Intensity( (ColorRGBA)cellColors[0, i] )
					));
				// point in the bottom row
				pointColors[height - 1, i] = new Scalar( lerp(
					Intensity( (ColorRGBA)cellColors[height - 2, i - 1] ),
					Intensity( (ColorRGBA)cellColors[height - 2, i] )
					));
			}

			// left and right edge points (excluding corner points) - special case
			for (int j = 1; j < height - 1; j++)
			{
				// point in the left column
				pointColors[j, 0] = new Scalar( lerp(
					Intensity( (ColorRGBA)cellColors[j - 1, 0] ),
					Intensity( (ColorRGBA)cellColors[j, 0] )
					));
				// point in the right column
				pointColors[j, width - 1] = new Scalar( lerp(
					Intensity( (ColorRGBA)cellColors[j - 1, width - 2] ),
					Intensity( (ColorRGBA)cellColors[j, width - 2] )
					));
			}

			// inner points - normal case
			for (int i = 1; i < width - 1; i++)
				for (int j = 1; j < height - 1; j++)
				{
					pointColors[j, i] = new Scalar( blin(
						Intensity( (ColorRGBA)cellColors[j - 1, i - 1] ),
						Intensity( (ColorRGBA)cellColors[j - 1, i] ),
						Intensity( (ColorRGBA)cellColors[j, i - 1] ),
						Intensity( (ColorRGBA)cellColors[j, i] )
						));
				}

			return pointColors;
		}

		#region Color intensity computing functions

		/// <summary>
		/// Computes lightness (volume, luminosity) of given color.
		/// </summary>
		/// <param name="color">Input color sample.</param>
		/// <returns>Returns lightness of given color.</returns>
		/// <remarks>Lightness is computed as an average of all color channels.</remarks>
		private double GetGrayIntensity(ColorRGBA color)
		{
			// return average color intensity
			return (color.RedF + color.GreenF + color.BlueF) / 3.0 * color.AlphaF;
		}

		/// <summary>
		/// Extracts red channel from given color.
		/// </summary>
		/// <param name="color">Input color sample.</param>
		/// <returns>Returns red channel of given color.</returns>
		private double GetRedIntensity(ColorRGBA color)
		{
			// return red color intensity
			return color.RedF * color.AlphaF;
		}

		/// <summary>
		/// Extracts green channel from given color.
		/// </summary>
		/// <param name="color">Input color sample.</param>
		/// <returns>Returns green channel of given color.</returns>
		private double GetGreenIntensity(ColorRGBA color)
		{
			// return green color intensity
			return color.GreenF * color.AlphaF;
		}

		/// <summary>
		/// Extracts blue channel from given color.
		/// </summary>
		/// <param name="color">Input color sample.</param>
		/// <returns>Returns blue channel of given color.</returns>
		private double GetBlueIntensity(ColorRGBA color)
		{
			// return blue color intensity
			return color.BlueF * color.AlphaF;
		}

		#endregion

		/// <summary>
		/// Computes linear interpolation of two numbers.
		/// </summary>
		/// <param name="a">First number.</param>
		/// <param name="b">Second number.</param>
		/// <returns>Returns linear interpolation of two numbers.</returns>
		private double lerp(double a, double b)
		{
			// lineary interpolate
			return (a + b) / 2.0;
		}

		/// <summary>
		/// Computes bilinear interpolation of four numbers.
		/// </summary>
		/// <param name="a">First number.</param>
		/// <param name="b">Second number.</param>
		/// <param name="c">Third number.</param>
		/// <param name="d">Fourth number.</param>
		/// <returns>Returns bilinear interpolation of four numbers.</returns>
		private double blin(double a, double b, double c, double d)
		{
			// bilineary interpolate
			return (a + b + c + d) / 4.0;
		}
	}
}