using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Finds the longest distance between the points in the set.
    /// </summary>
    [ModuleInfo("Vit Ondracka", "Finds the longest distance between the points in the set.", CategoryType.Filter, CategoryData.Mesh)]
    [PortInfo("Points", "Set of points")]
    [PortInfo("Distance", "Maximum distance")]
    [PortInfo("Index0", "Index of the first point forming the maximum distance")]
    [PortInfo("Index1", "Index of the second point forming the maximum distance")]
    public class Distance : Zcu.Mve.Core.Module
    {
        private int[,] crossCubeIndices = { { 0, 7 }, { 1, 6 }, { 2, 5 }, { 3, 4 } };
        private int[,] crossFaceIndices = { { 0, 3 }, { 0, 5 }, { 0, 6 }, { 1, 2 }, { 1, 4 }, { 1, 7 }, { 2, 4 }, { 2, 7 }, { 3, 5 }, { 3, 6 }, { 4, 7 }, { 5, 6 } };
        private int[,] neighbIndices = { { 0, 1 }, { 0, 2 }, { 0, 4 }, { 1, 3 }, { 1, 5 }, { 2, 3 }, { 2, 6 }, { 3, 7 }, { 4, 5 }, { 4, 6 }, { 5, 7 }, { 6, 7 } };
        private Point3D[] p;
        private int[] c; // point count in the segments
        private int[] seg; // array storing the segment of the bounding box in which the point happens to be
        private double[] dst; // array storing the segment of the bounding box in which the point happens to be
        private double maxDist; // maximum distance found so far
        private int i0, i1; // indices of the points with the maximum distance
        private double[] d; // distance of the most distant points from the corners
        private int[] id;   //  indices  "  "    "     "       "     "   "     "

        private Scalar distance;
        private Integer index0, index1;

        /// <summary>
        /// Constructs a new isntance of the module, adds input and output ports.
        /// </summary>
        public Distance()
        {
            AddInPort("Points", typeof(UnstrGrid));
            AddOutPort("Distance", typeof(Scalar));
            AddOutPort("Index0", typeof(Integer));
            AddOutPort("Index1", typeof(Integer));
        }

        private void DistByIndices(ref int[] indices, int count)
        {
            int i, j;
            double xDiff, yDiff, zDiff;
            double p0x, p0y, p0z;
            Point3D p1;
            double dist;

            for (i = 0; i < count; i++)
            {
                // for each point...
                p0x = p[indices[i]].X;
                p0y = p[indices[i]].Y;
                p0z = p[indices[i]].Z;

                for (j = i + 1; j < count; j++)
                {
                    // ...and each other point
                    p1 = p[indices[j]];

                    xDiff = p0x - p1.X; yDiff = p0y - p1.Y; zDiff = p0z - p1.Z; // calculate the differences
                    dist = xDiff * xDiff + yDiff * yDiff + zDiff * zDiff; // calculate the distance
                    if (dist > maxDist)
                    {
                        // if it's the longest one, update
                        maxDist = dist;
                        i0 = indices[i]; i1 = indices[j];
                    }
                }
            }
        }

        private void DistPoints(ref int[] indices1, ref int[] indices2, int count1, int count2)
        {
            int i, j; // iterators
            Point3D p0, p1; // temporary stores for the points
            double dist; // actual distance, maximum found distance
            double xDiff, yDiff, zDiff; // coordinate differences

            // for each point in the first segment...
            for (i = 0; i < count1; i++)
            {
                p0 = p[indices1[i]];
                // for each point in the second segment...
                for (j = 0; j < count2; j++)
                {
                    p1 = p[indices2[j]];
                    // calculate the distance
                    xDiff = p0.X - p1.X; yDiff = p0.Y - p1.Y; zDiff = p0.Z - p1.Z;
                    dist = xDiff * xDiff + yDiff * yDiff + zDiff * zDiff;
                    // update the maximum distance and point indices if necessary
                    if (dist > maxDist)
                    {
                        maxDist = dist;
                        i0 = indices1[i]; i1 = indices2[j];
                    }
                }
            }
        }

        /// <summary>
        /// Finds the largets distance between the points in the two given segments
        /// </summary>
        /// <param name="segIndices">pairs of the indices of the segments</param>
        private void DistSegments(ref int[,] segIndices)
        {
            int i, j, segIndex0, segIndex1; // iterators and segment indices
            int count0, count1; // counters for the index arrays
            int[] indices0, indices1; // index arrays

            indices0 = indices1 = new int[p.Length]; // init the index arrays

            // for each pair of the segment indices...
            for (i = 0; i < (segIndices.Length >> 1); i++)
            {
                segIndex0 = segIndices[i, 0];
                segIndex1 = segIndices[i, 1];

                // if there are point in both segments...
                if (c[segIndex0] * c[segIndex1] > 0)
                {
                    count0 = count1 = 0;
                    // store the indices of the points belonging to one of the segments into the index arrays
                    for (j = 0; j < seg.Length; j++)
                        if (seg[j] == segIndex0) indices0[count0++] = j; // (first segment)
                        else if (seg[j] == segIndex1) indices1[count1++] = j; // (second segment)
                    
                    // update the maximum distance - search for the maximum distance between the given segments' points
                    DistPoints(ref indices0, ref indices1, c[segIndex0], c[segIndex1]);
                }
            }
        }

        /// <summary>
        /// Limit the data set by the distances between the most distant points of the given segments of the set 
        /// </summary>
        /// <param name="segArray">array with pairs of indices of the segments to check</param>
        private void LimitData(ref int[,] segArray)
        {
            int i, segIndex0, segIndex1; // iterator and the segment indices
            double xDiff, yDiff, zDiff, dist, mDist; // coordinate differences, the actual distance and the initial maximum distance
            Point3D p0, p1; // temporary store for the points

            mDist = maxDist;

            // for each pair of the segments...
            for (i = 0; i < (segArray.Length >> 1); i++)
            {
                segIndex0 = segArray[i, 0];
                segIndex1 = segArray[i, 1];

                // get the most distant points of these segments...
                p0 = p[id[segIndex0]];
                p1 = p[id[segIndex1]];

                // calculate the distance between them...
                xDiff = p0.X - p1.X; yDiff = p0.Y - p1.Y; zDiff = p0.Z - p1.Z;
                dist = xDiff * xDiff + yDiff * yDiff + zDiff * zDiff;
                // and if necessary...
                if (dist > maxDist)
                {
                    maxDist = dist; // increase the maximum distance found so far...
                    i0 = id[segIndex0]; // and store the indices of the points forming the maximum distance
                    i1 = id[segIndex1];
                }
            }
            if (maxDist > mDist) LimitSelection(); // reduce the data set by the new found maximum distance
        }

        /// <summary>
        /// Discards the points which are closer to the corresponding corners of the bounding box
        /// than the given distance
        /// </summary>
        private void LimitSelection()
        {
            for (int i = 0; i < c.Length; i++) c[i] = 0; // reset the counters of the segments
            // for each point...
            for (int i = 0; i < p.Length; i++)
            {
                if (seg[i] == -1) continue; // ...which is not already discarded...
                // ...check it's distance from the corner against the maximum distance...
                if (dst[i] < maxDist) seg[i] = -1; // ...if it's smaller, discard the point...
                else c[seg[i]]++; // otherwise increment the counter
            }
        }

        /// <summary>
        /// Finds the longest distance between the points in the given set
        /// </summary>
        /// <returns>array of indices of the two most distant points</returns>
        public void Dist()
        {
            #region local variables & initial setup
            int i; // iterator
            double xMin, xMax, yMin, yMax, zMin, zMax; // data set's extreme coordinate values
            int xMinI, xMaxI, yMinI, yMaxI, zMinI, zMaxI; // data set's extreme coordinate point indices
            double xDiff, yDiff, zDiff; // coordinate differences for calculating the distance
            double cX, cY, cZ; // center of the bounding box
            //double[] dst = new double[p.Length]; // distance of the points from the corresponding corner of the bounding box
            int[] ie = new int[6]; // index array for finding maximum distance between the points forming the bounding box
            Scalar[] res = new Scalar[2]; // return value

            // initial setup
            xMinI = xMaxI = yMinI = yMaxI = zMinI = zMaxI = 0;
            xMin = xMax = p[0].X;
            yMin = yMax = p[0].Y;
            zMin = zMax = p[0].Z;
            c[0] = c[1] = c[2] = c[3] = c[4] = c[5] = c[6] = c[7] = 0;
            #endregion

            /* find the points forming a bounding box, calculate its center
             * and calculate the largest distance between these points */
            #region bounding box and initial longest distance
            // find a bounding box of the points
            for (i = 1; i < p.Length; i++)
            {
                if (p[i].X > xMax) { xMax = p[i].X; xMaxI = i; }
                else if (p[i].X < xMin) { xMin = p[i].X; xMinI = i; }
                if (p[i].Y > yMax) { yMax = p[i].Y; yMaxI = i; }
                else if (p[i].Y < yMin) { yMin = p[i].Y; yMinI = i; }
                if (p[i].Z > zMax) { zMax = p[i].Z; zMaxI = i; }
                else if (p[i].Z < zMin) { zMin = p[i].Z; zMinI = i; }
            }

            // calculate the center of the bounding box
            cX = xMin + (xMax - xMin) * 0.5;
            cY = yMin + (yMax - yMin) * 0.5;
            cZ = zMin + (zMax - zMin) * 0.5;

            // find the longest distance between the extreme points
            ie[0] = xMaxI; ie[1] = xMinI; ie[2] = yMaxI; ie[3] = yMinI; ie[4] = zMaxI; ie[5] = zMinI;
            DistByIndices(ref ie, 6);
            #endregion

            /* divide the points into regions & calculate the distance from the corresponding corner
             * and find the points most distant from the corners */
            #region point partitioning & distance from the corners
            d[0] = d[1] = d[2] = d[3] = d[4] = d[5] = d[6] = d[7] = 0;
            id[0] = id[1] = id[2] = id[3] = id[4] = id[5] = id[6] = id[7] = 0;
            for (i = 0; i < p.Length; i++)
            {
                // determine the potential segment for the point
                if (p[i].X > cX)
                {
                    seg[i] += 1;
                    xDiff = p[i].X - xMin; // we're checking against xMin
                }
                else xDiff = p[i].X - xMax; // " xMax
                if (p[i].Y > cY)
                {
                    seg[i] += 2;
                    yDiff = p[i].Y - yMin; // " yMin
                }
                else yDiff = p[i].Y - yMax; // " yMax
                if (p[i].Z > cZ)
                {
                    seg[i] += 4;
                    zDiff = p[i].Z - zMin; // " zMin
                }
                else zDiff = p[i].Z - zMax; // " zMax

                // calculate the distance from the corresponding corner & possibly remove from the set
                dst[i] = xDiff * xDiff + yDiff * yDiff + zDiff * zDiff;
                if (dst[i] < maxDist) seg[i] = -1; // discard the point
                else
                {
                    c[seg[i]]++;
                    if (d[seg[i]] < dst[i]) // else check if we have the most distant point of the segment
                    {
                        // if so, update the distance
                        d[seg[i]] = dst[i];
                        id[seg[i]] = i;
                    }
                }
            }
            #endregion

            /* limit the data set and find the longest distance
             * first for the cross-cube segments, then for the cross-face segments and finally for the neighbouring segments */
            #region data set reduction and longest distance calculation
            // limit the data set by the cross cube distances
            LimitData(ref crossCubeIndices);
            // find the longest cross cube distance in the set
            DistSegments(ref crossCubeIndices);

            // is it necessary to check distances of cross-face segments?
            xDiff = xMax - xMin; yDiff = yMax - yMin; zDiff = zMax - zMin;
            if (xDiff < yDiff) xDiff = yDiff;
            if (xDiff < zDiff) xDiff = zDiff;
            if (maxDist < 2.25 * xDiff * xDiff)
            {
                // limit the data set by the cross gace distances
                LimitData(ref crossFaceIndices);
                // find the longest cross face distance in the set
                DistSegments(ref crossFaceIndices);
            }

            // is it necessary to check distances of neighbouring segments?
            if (maxDist < 1.25 * xDiff * xDiff)
            {
                // limit the data set by the cross gace distances
                LimitData(ref neighbIndices);
                // find the longest cross face distance in the set
                DistSegments(ref neighbIndices);
            }
            #endregion
        }

        /// <summary>
        /// Finds the most distant points in the set.
        /// </summary>
        public override void Execute()
        {
            if (IsAllDataSame())
            {
                SetOutput("Distance", distance,true);
                SetOutput("Index0", index0, true);
                SetOutput("Index1", index1, true);
            }
            else
            {
                UnstrGrid ug = (UnstrGrid)GetInput("Points");
                UniformDataArray points = ug.Points;

                p = new Point3D[points.Count];
                for (int i = 0; i < points.Count; i++)
                    p[i] = (Point3D)points[i];

                c = new int[8];
                dst = new double[p.Length];
                seg = new int[p.Length];
                d = new double[8];
                id = new int[8];

                maxDist = 0;
                i0 = i1 = -1;

                Dist();

                distance = new Scalar(Math.Sqrt(maxDist));
                index0 = new Integer(i0);
                index1 = new Integer(i1);

                SetOutput("Distance", distance);
                SetOutput("Index0", index0);
                SetOutput("Index1", index1);
            }
        }
    }
}
