using System;
using System.Collections;
using System.Xml;
using System.Diagnostics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
  /// <summary>
  /// N-dimensional structured grid.
  /// Contains two tables:
  /// <ul>
  ///  <li>table of cell (element) attributes with height x width resolution ND uniform arrays.</li>
  ///  <li>table of point (element vertices) attributes with height+1 x width+1 resolution ND uniform arrays.</li>
  ///  </ul>
  /// </summary>
  [DataObject("N-dimensional structured grid."
  + "Contains two tables:"
  + "<ul>"
  + "<li>table of cell (element) attributes with height x width resolution ND uniform arrays.</li>"
  + "<li>table of point (element vertices) attributes with height+1 x width+1 resolution ND uniform arrays.</li>"
  + "</ul>")]
  public abstract class GridND : Zcu.Mve.Visualization.StructGrid
  {
    /// <summary>
    /// 
    /// </summary>
    private Hashtable pointAttrs = new Hashtable();
    private Hashtable cellAttrs = new Hashtable();
    private int[] dimensions;


    /// <summary>
    /// Returns array containing grid dimensions (i.e. number of cells in each dimension)
    /// </summary>
    public int[] Dimensions
    {
      get
      {
        return ((int[]) (this.dimensions.Clone()));
      }
    }


    /// <summary>
    /// For XmlLoader use only. Use parametric constructor instead.
    /// </summary>
    public GridND()
    {
      UpdateModificationTime();
    } // GridND()


    /// <summary>
    /// Constructor - creates empty grid of given resolution
    /// </summary>
    /// <param name="dimensions">Dimensions of the grid to be created. You can pass either an array of integers, or a number of integer parameters. Created grid has given number of cells, i.e. number of vertices in each direction is higher by one!</param>
    /// <example>
    /// To create a 3D array of 128x128x128 voxels (i.e. 129x129x129 vertices) you can call either 
    /// <code>
    /// GridND volume = new GridND(128,128,128);
    /// </code>
    /// or
    /// <code>
    /// int[] dimensions = new int[] {128,128,128};
    /// GridND volume = new GridND(dimensions);
    /// </code>
    /// </example>
    public GridND(params int[] dimensions)
    {
      if (dimensions.Length == 0)
      {
        throw new MveException(this, "Number of GridND dimensions must be at least one.");
      }

      this.dimensions = new int[dimensions.Length];
      for (int i = 0; i < dimensions.Length; i++)
      {
        this.dimensions[i] = dimensions[i];
      }

      UpdateModificationTime();
    } // GridND()


    /// <summary>
    /// Add cell attributes.
    /// Array dimensions have to agree with GridND dimensions.
    /// </summary>
    /// <param name="attrid">Localy unique ID. Name of attribute type is recommended as default value.</param>
    /// <param name="attrs">ND array of attributes.</param>
    public void AddCellAttrs(string attrid, MveArray attrs)
    {
      if (attrs.Lengths.Length != this.dimensions.Length)
      {
        throw new MveException(this, "Number of cell attribute dimensions and GridND dimensions have to agree.");
      }

      for (int i = 0; i < this.dimensions.Length; i++)
        if (this.dimensions[i] != attrs.Lengths[i])
        {
          throw new MveException(this, "Cell attributes dimensions and GridND dimensions have to agree.");
        }

      if (this.cellAttrs.ContainsKey(attrid))
      {
        throw new MveException(this, "GridND already contains cell attribute id " + attrid + ".");
      }

      this.cellAttrs.Add(attrid, attrs);

      UpdateModificationTime();
    }

    /// <summary>
    /// Removes attribute associated with cells of the RegGrid. Throws MveException if the attribute of the given name does not exist.
    /// </summary>
    /// <param name="attrid">String identifier of the attribute to be removed.</param>
      public void RemoveCellAttrs(string attrid)
      {
          if (this.cellAttrs.ContainsKey(attrid))
          {
              this.cellAttrs.Remove(attrid);
              UpdateModificationTime();
          }
          else
          {
              throw new MveException(this, "GridND does not contain cell attribute with id "+attrid+".");
          }
      }


    /// <summary>
    /// Add point attributes.
    /// Array dimensions have to agree with GridND dimensions.
    /// </summary>
    /// <param name="attrid">Localy unique ID. Name of attribute type is recommended as default value.</param>
    /// <param name="attrs">ND array of attributes.</param>
    public void AddPointAttrs(string attrid, MveArray attrs)
    {

      if (attrs.Lengths.Length != this.dimensions.Length)
      {
        throw new MveException(this, "Number of point attribute dimensions and GridND dimensions have to agree.");
      }

      for (int i = 0; i < this.dimensions.Length; i++)
        if ((this.dimensions[i] + 1) != attrs.Lengths[i])
        {
          throw new MveException(this, "Point attributes dimensions and GridND dimensions have to agree.");
        }

      if (this.pointAttrs.ContainsKey(attrid))
      {
        throw new MveException(this, "GridND already contains point attribute id " + attrid + ".");
      }

      this.pointAttrs.Add(attrid, attrs);

      UpdateModificationTime();
    }

      /// <summary>
      /// Removes attribute associated with points of the RegGrid. Throws MveException if the attribute of the given name does not exist.
      /// </summary>
      /// <param name="attrid">String identifier of the attribute to be removed.</param>
      public void RemovePointAttrs(string attrid)
      {
          if (this.pointAttrs.ContainsKey(attrid))
          {
              this.pointAttrs.Remove(attrid);
              UpdateModificationTime();
          }
          else
          {
              throw new MveException(this, "GridND does not contain point attribute with id "+attrid+".");
          }
      }


    /// <summary>
    /// Return ND-array of pixel attributes.
    /// </summary>
    /// <param name="attrid"></param>
    /// <returns></returns>
    public MveArray GetCellAttrs(string attrid)
    {
      if (!this.cellAttrs.ContainsKey(attrid))
        throw new MveException(this, "GridND do not contain pixel attributes with name " + attrid + ".");

      return (MveArray) this.cellAttrs[attrid];
    }


    /// <summary>
    /// Return ND-array of point attributes.
    /// </summary>
    /// <param name="attrid"></param>
    /// <returns></returns>
    public MveArray GetPointAttrs(string attrid)
    {
      if (!this.pointAttrs.ContainsKey(attrid))
        throw new MveException(this, "GridND do not contain point attributes with name " + attrid + ".");

      return (MveArray) this.pointAttrs[attrid];
    }


    /// <summary>
    /// Returns collection of point attribute ids.
    /// </summary>
    /// <returns></returns>
    public ICollection GetPointAttrIds()
    {
      return this.pointAttrs.Keys;
    }


    /// <summary>
    /// Returns collection of cell attribute ids.
    /// </summary>
    /// <returns></returns>
    public ICollection GetCellAttrIds()
    {
      return this.cellAttrs.Keys;
    }

    /// <summary>
    /// Returns point in given position. 
    /// Point coordinates are given implicitly. Thus no points are stored in GridND data object.
    /// Point object is created and coordinates are calculated in implementation of getPoint function.
    /// </summary>
    /// <param name="coordinates">Array of coordinates, dimension must be equal to the dimension of the GridND.</param>
    /// <returns>Particular point</returns>
    public abstract PointND GetPoint(params int[] coordinates);


    /// <summary>
    /// Creates new instance of GridND ...
    /// </summary>
    /// <returns></returns>
    public override IDataObject DeepCopy()
    {
      GridND copy = (GridND) Activator.CreateInstance(this.GetType(), new object[] { this.dimensions });

      foreach (string cellAttrId in this.GetCellAttrIds())
      {
        copy.AddCellAttrs(cellAttrId, this.GetCellAttrs(cellAttrId));
      }

      foreach (string pointAttrId in this.GetPointAttrIds())
      {
        copy.AddCellAttrs(pointAttrId, this.GetPointAttrs(pointAttrId));
      }

      return copy;
    }


    /// <summary>
    /// Reads data for the GridND form an XML reader.
    /// </summary>
    /// <param name="xmlTextReader">XmlTextReader to read the data from</param>
    public override void ReadData(XmlTextReader xmlTextReader)
    {
        base.ReadData(xmlTextReader);
        xmlTextReader.WhitespaceHandling = WhitespaceHandling.None;
      ReadHeader(xmlTextReader);

      while (xmlTextReader.Read())
      {
        ReadNode(xmlTextReader);
      } // while

      UpdateModificationTime();
    }


    /// <summary>
    /// Reads width and height attributes
    /// </summary>
    /// <param name="xmlTextReader"></param>
    protected void ReadHeader(XmlTextReader xmlTextReader)
    {
      xmlTextReader.Read();
      if (xmlTextReader.Name == "dimensionCount")
      {
          xmlTextReader.Read();
        this.dimensions = new int[(int) Int32.Parse(xmlTextReader.Value.Trim())];
      }
      else
      {
        throw new MveException(this, "Number of dimensions expected in a dimensionCount element.");
      }

      xmlTextReader.Read();

      int i = 0;
      while (i < dimensions.Length)
      {
        xmlTextReader.Read();
        if (xmlTextReader.Name == "dimension")
        {
            xmlTextReader.Read();
          this.dimensions[i] = (int) Int32.Parse(xmlTextReader.Value.Trim());
          xmlTextReader.Read();
        }
        else
        {
          throw new MveException(this, i + ". dimension of GridND expected in dimension element.");
        }
        i++;
      }
    }

    /// <summary>
    /// Reads pointattrs and cellattrs nodes.
    /// </summary>
    /// <param name="xmlTextReader"></param>
    protected void ReadNode(XmlTextReader xmlTextReader)
    {
      MveArray array;
      string attrId;

      if (xmlTextReader.NodeType == XmlNodeType.Element)
      {
        switch (xmlTextReader.Name)
        {/*
            case "transform":
              this.transform = (AbstractTransform)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              this.transform.ReadData(xmlTextReader);
              break;
              */
          case "pointattrs":	// naten atribut bod
            attrId = xmlTextReader.GetAttribute("attrid").Trim();
            array = new MveArray();
            array.ReadData(xmlTextReader);
            this.AddPointAttrs(attrId, array);
            break;
          case "cellattrs":	// naten atribut bunk
            attrId = xmlTextReader.GetAttribute("attrid").Trim();
            array = new MveArray();
            array.ReadData(xmlTextReader);
            this.AddCellAttrs(attrId, array);
            break;
          default:
            Trace.WriteLine("Uknown node: " + xmlTextReader.Name + ".");
            break;
        } // switch
      }
    }


    /// <summary>
    /// Write all point attributes into XML file.
    /// </summary>
    /// <param name="xmlTextWriter"></param>
    protected void WritePointAttrs(XmlTextWriter xmlTextWriter)
    {
      // Write point attributes
      foreach (DictionaryEntry entry in this.pointAttrs)
      {
        UniformDataArrayND array = (UniformDataArrayND) entry.Value;
        xmlTextWriter.WriteStartElement("pointattrs");
        xmlTextWriter.WriteAttributeString("attrid", entry.Key.ToString());
        array.WriteData(xmlTextWriter);
        xmlTextWriter.WriteEndElement();
      }
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="xmlTextWriter"></param>
    protected void WriteCellAttrs(XmlTextWriter xmlTextWriter)
    {
      // Write cell attributes
      foreach (DictionaryEntry entry in this.cellAttrs)
      {
        UniformDataArrayND array = (UniformDataArrayND) entry.Value;
        xmlTextWriter.WriteStartElement("cellattrs");
        xmlTextWriter.WriteAttributeString("attrid", entry.Key.ToString());
        array.WriteData(xmlTextWriter);
        xmlTextWriter.WriteEndElement();
      }
    }

    /// <summary>
    /// Writes resolution of grid.
    /// </summary>
    /// <param name="xmlTextWriter"></param>
    protected void WriteHeader(XmlTextWriter xmlTextWriter)
    {
      xmlTextWriter.WriteElementString("dimensionCount", this.dimensions.Length.ToString());
      for (int i = 0; i < this.dimensions.Length; i++)
      {
        xmlTextWriter.WriteElementString("dimension", this.dimensions[i].ToString());
      }
    }


    /// <summary>
    /// Writes dimensions, point attrs and cell attrs.
    /// </summary>
    /// <param name="xmlTextWriter"></param>
    public override void WriteData(XmlTextWriter xmlTextWriter)
    {
        base.WriteData(xmlTextWriter);
      this.WriteHeader(xmlTextWriter);

      this.WritePointAttrs(xmlTextWriter);

      this.WriteCellAttrs(xmlTextWriter);
    }


    /// <summary>
    /// checks the consistence of the grid.
    /// </summary>
    /// <returns>Always returns true.</returns>
    public override bool CheckConsistence()
    {
      return true;
    }

    /// <summary>
    /// Transforms the contents of the GridND into a string
    /// </summary>
    /// <returns>String representation. Contains pixel values and attributa names</returns>
    public override string ToString()
    {
      string str = "Visualization.GridND:\r\n";

      str += "  dimension = " + dimensions.Length.ToString() + "\r\n";
      for (int i = 0; i < dimensions.Length; i++)
        str += "  " + i + ". length = " + dimensions[i].ToString() + "\r\n";

      str += "Point attributes:\r\n";
      foreach (DictionaryEntry entry in this.pointAttrs)
      {
          MveArray array = (MveArray)entry.Value;
        str += "  ID: " + entry.Key + " Type: " + array.ElemType.FullName + "\r\n";
      }
      str += "Cell attributes:\r\n";
      foreach (DictionaryEntry entry in this.cellAttrs)
      {
          MveArray array = (MveArray)entry.Value;
        str += "  ID: " + entry.Key + " Type: " + array.ElemType.FullName + "\r\n";
      }
      return str;
    }
  }
}
