using System;
using System.IO;
using System.Text;

namespace Zcu.Mve.Visualization.Loaders
{
	/// <summary>
	/// Class for reading TRI file format.
	/// </summary>
	public class ReaderTri
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private int actualLine = 0;                        // Actual line of read file.

		private string exception = null;

		private DataSet dataSet;

		/// <summary>
		/// Reads input file.
		/// </summary>
		/// <param name="fileName">Full filename.</param>
		/// <returns>Filled object of DataSet class.</returns>
		public DataSet ReadFile(string fileName)
		{
			StreamReader streamReader = new StreamReader(fileName);

			string line = null;
			byte   numberOf = 2;                           //  Counting parts of file head.

			try
			{
				while ((line = streamReader.ReadLine()) != null) 
				{
					this.actualLine++;
					line = line.Trim();
					line = line.ToUpper();

					if (line.IndexOf("NUMBER OF TRI") > 0)
					{                                      // Reads number of triangles.
						this.numberOfCells = this.IntNumber(line);
						numberOf--;
					}
					else if (line.IndexOf("NUMBER OF VER") > 0)
					{                                      // Reads number of vertices.
						this.numberOfVertices = this.IntNumber(line);
						numberOf--;
					}

					if (numberOf == 0) 
					{                                      // Numbers are readed - creates object of DataSet.
						this.dataSet = new DataSet(this.numberOfVertices, this.numberOfCells, 0);
						break;
					}
				}

				if (numberOf != 0)                             // if number of triangles or vertices wasn't found - throw exception
				{
					this.exception = "Invalid file format - can't find number of triangles and/or number of vertices.";
                    throw new Zcu.Mve.Core.MveException(typeof(MeshLoader), "Invalid file format - can't find number of triangles and/or number of vertices.");
				}

				while ((line = streamReader.ReadLine()) != null) 
				{
					this.actualLine++;

					line = line.Trim();
					line = line.ToUpper();

					if (line == "[VERTICES]") 
					{                                      // Reads vertices.
						this.ReadVertices(ref streamReader);
					}

					else if (line == "[TRIANGLES]") 
					{                                      // Reads triangles.
						this.ReadTriangles(ref streamReader);
					}

					else if (line == "[TRIANGLES' NORMALS]") 
					{                                      // Reads triangle normals.
						this.ReadTrianglesNormals(ref streamReader);
					}

					else if (line == "[VERTICES' NORMALS]") 
					{                                      // Reads vertex normals.
						this.ReadVerticesNormals(ref streamReader);
					}				

					else if (line == "[TRIANGLES' NEIGHBOURS]") 
					{                                      // Reads triangle neighbours.
						this.ReadTrianglesNeighbours(ref streamReader);
					}				
				}
			}
			catch (System.Exception)
			{
				streamReader.Close();
				if (this.exception == null) 
				{
					throw new Zcu.Mve.Core.MveException(typeof(MeshLoader), "Error in input file on line " + this.actualLine + ".");
				}
				else
				{
                    throw new Zcu.Mve.Core.MveException(typeof(MeshLoader), this.exception);
				}
			}

			streamReader.Close();

			return this.dataSet;
		}

		/// <summary>
		/// Reads one integer number.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Integer number.</returns>
		private int IntNumber(string line) 
		{
			string[] elements;

			elements = line.Split(" :\t".ToCharArray());

			return Int32.Parse(elements[elements.Length - 1]);
		}

		/// <summary>
		/// Reads tree double numbers.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Mve structure Point 3D.</returns>
		private Zcu.Mve.Visualization.Point3D ReadPoint3D(string line) 
		{
			string[] elements;
			int elementsLength = 0;
			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.Point3D(Double.Parse(elements[elementsLength - 3], nfi),
				Double.Parse(elements[elementsLength - 2], nfi),
				Double.Parse(elements[elementsLength - 1], nfi));
		}

		/// <summary>
		/// Reads tree double numbers.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Mve structure Normal 3D.</returns>
		private Zcu.Mve.Visualization.Normal3D ReadNormal3D(string line) 
		{
			string[] elements;
			int elementsLength = 0;
			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.Normal3D(Double.Parse(elements[elementsLength - 3], nfi),
				                                      Double.Parse(elements[elementsLength - 2], nfi),
				                                      Double.Parse(elements[elementsLength - 1], nfi));
		}

		/// <summary>
		/// Reads tree integer numbers.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Mve structure Triangle Neighbour.</returns>
		private Zcu.Mve.Visualization.TriangleNeighbor ReadTriangleNeighbor(string line) 
		{
			string[] elements;
			int elementsLength = 0;

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.TriangleNeighbor(Int32.Parse(elements[elementsLength - 3]),
				                                              Int32.Parse(elements[elementsLength - 2]),
				                                              Int32.Parse(elements[elementsLength - 1]));
		}

		/// <summary>
		/// Reads tree integer numbers.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Mve structure Triangle.</returns>
		private Zcu.Mve.Visualization.Triangle ReadTriangle(string line) 
		{
			string[] elements;
			int elementsLength = 0;

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.Triangle(Int32.Parse(elements[elementsLength - 3]),
				Int32.Parse(elements[elementsLength - 2]),
				Int32.Parse(elements[elementsLength - 1]));
		}

		/// <summary>
		/// Reads block of vertices in input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadVertices(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Point3D vertex;
			string   line;
			int    i = 0;                                  // Index of actual vertex.

			while ((line = streamReader.ReadLine()) != null) 
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skip free line.
				{
					continue;
				}
				else                                       // Try to read one vertex.
				{
					vertex = ReadPoint3D(line);
					this.dataSet.AddVertex(i, vertex);

					if (++i == numberOfVertices)           // Number of vertices reached - exit.
					{
						break;
					}
				}
			}
		}

		/// <summary>
		/// Reads block of triangles in input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadTriangles(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Triangle triangle;
			string line;
			int    i = 0;                                  // Index of actual triangle.

			while ((line = streamReader.ReadLine()) != null)
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skip free line.
				{
					continue;
				}
				else                                       // Try to read one triangle.
				{
					triangle = ReadTriangle(line);
					
					if ((triangle.V1 > this.numberOfVertices) ||
					    (triangle.V2 > this.numberOfVertices) ||
					    (triangle.V3 > this.numberOfVertices))
					{
						this.exception = "Error in input file on line " + this.actualLine + " - index is greater then number of vertices.";
                        throw new Zcu.Mve.Core.MveException(typeof(MeshLoader), "Error in input file on line " + this.actualLine + " - index is greater then number of vertices.");
					}

					this.dataSet.AddTriangle(i, triangle); // Add triangle to data set.

					if (++i == numberOfCells)              // Number of triangles reached - exit.
					{
						break;
					}
				}
			}
		}

		/// <summary>
		/// Reads block of triangle normals in input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadTrianglesNormals(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Normal3D normal;
			string   line;
			int    i = 0;                                  // Index of actual vertex.

			while ((line = streamReader.ReadLine()) != null) 
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skip free line.
				{
					continue;
				}
				else                                       // Try to read one triangle normal.
				{
					normal = ReadNormal3D(line);
					this.dataSet.AddCellNormal(i, normal);

					if (++i == numberOfCells)              // Number of vertices reached - exit.
					{
						break;
					}
				}
			}
		}

		/// <summary>
		/// Reads block of vertex normals in input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadVerticesNormals(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Normal3D normal;
			string   line;
			int    i = 0;                                  // Index of actual vertex.

			while ((line = streamReader.ReadLine()) != null) 
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skip free line.
				{
					continue;
				}
				else                                       // Try to read one vertex normal.
				{
					normal = this.ReadNormal3D(line);
					this.dataSet.AddVertexNormal(i, normal);

					if (++i == numberOfVertices)           // Number of vertices normals reached - exit.
					{
						break;
					}
				}
			}
		}
		
		/// <summary>
		/// Reads block of triangle neighbours in input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadTrianglesNeighbours(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.TriangleNeighbor neighbor;
			string line;
			int    i = 0;                                  // Index of actual triangle.

			while ((line = streamReader.ReadLine()) != null)
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skip free line.
				{
					continue;
				}
				else                                       // Try to read three indexes of neighbours.
				{
					neighbor = this.ReadTriangleNeighbor(line);

					if ((neighbor.Neighbors[0] > this.numberOfCells) ||
						(neighbor.Neighbors[1] > this.numberOfCells) ||
						(neighbor.Neighbors[2] > this.numberOfCells))
					{
						this.exception = "Error in input file on line " + this.actualLine + " - index is greater then number of vertices.";
                        throw new Zcu.Mve.Core.MveException(typeof(MeshLoader), "Error in input file on line " + this.actualLine + " - index is greater then number of vertices.");
					}

					this.dataSet.AddTriangleNeighbours(i, neighbor);
					                                       // Add triangle neighbours to data set.
					if (++i == numberOfCells)              // Number of triangles reached - exit.
					{
						break;
					}
				}
			}
		}	
	}
}
