using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using Zcu.Mve.Strings;
using System.IO;
using System.Globalization;
using System.ComponentModel;
using System.Windows.Forms.Design;

namespace Zcu.Mve.Visualization.Loaders
{
    /// <summary>
    /// Obj triangle mesh file loader module.
    /// </summary>
    [ModuleInfo("Libor Va", "Module for .obj format import. Produces Mesh.", CategoryType.Loader, CategoryData.Mesh, IconName = "objLoader.ico")]
    [PortInfo("Filename", "MveString that contains name of file to load. If nothing is connected, then the value of the FileName property is used")]
    [PortInfo("TriangleMesh", "Mesh instance that contains vertices and triangles loaded from the file")]
    public class ObjLoader : Module
    {

        class ObjFileNameEditor : FileNameEditor
        {
            string path;

            public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value)
            {
                this.path = (string)value;
                return base.EditValue(context, provider, value);
            }

            protected override void InitializeDialog(System.Windows.Forms.OpenFileDialog openFileDialog)
            {
                openFileDialog.Filter = "OBJ files (*.obj)|*.obj|All files (*.*)|*.*";   
                if (path.LastIndexOf("\\")>=0)
                    openFileDialog.InitialDirectory = path.Substring(0, path.LastIndexOf("\\"));
            }
        }

        private bool paramsChanged = true;

        private bool reload = false;

        /// <summary>
        /// Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.
        /// </summary>
        [Description("Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.")]
        public bool Reload
        {
            get { return (reload); }
            set { reload = value; }
        }

        bool flipNormals = false;

        public bool FlipNormals
        {
            get { return flipNormals; }
            set { flipNormals = value; }
        }

        bool setDataSame = true;

        public bool SetDataSame
        {
            get { return setDataSame; }
            set { setDataSame = value; }
        }

        bool flipZ = true;

        public bool FlipZ
        {
            get { return flipZ; }
            set { flipZ = value; }
        }

        private string fileName = "mesh.obj";

        private TriangleMesh output;
        private UniformDataArray textureCoordinates;

        private string trianglesCellName = "triangles";

        /// <summary>
        /// Sets the identifier for the triangle cells in the output TriangleMesh.
        /// </summary>
        [Description("Sets the identifier for the triangle cells in the output TriangleMesh")]
        public string TrianglesCellName
        {
            get { return (trianglesCellName); }
            set
            {
                if (value == "")
                {
                    throw new MveException(this, "Cell identifier of triangle in the ObjLoader module cannot be an empty string.");
                }
                else
                {
                    this.trianglesCellName = value;
                    paramsChanged = true;
                }
            }
        }

        bool loadNormals = true;

        public bool LoadNormals
        {
            get { return loadNormals; }
            set { loadNormals = value; }
        }

        /// <summary>
        /// File to load data from. Should be in the .obj format. The property has no effect when there's a connection at the Filename port
        /// </summary>
        [Editor(typeof(ObjFileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
        [Description("File to load data from. Should be in the .obj format. The property has no effect when there's a connection at the Filename port.")]
        public string FileName
        {
            get { return (fileName); }
            set
            {
                if (value != "")
                {
                    fileName = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "FileName property cannot be set to an empty string!");
                }
            }
        }

        /// <summary>
        /// Default constructor.
        /// </summary>
        public ObjLoader()
        {
            AddOutPort("TriangleMesh", typeof(TriangleMesh));
            AddOutPort("TextureCoordinates", typeof(UniformDataArray));
            AddInPort("Filename", typeof(MveString), false);
        }

        bool loadTextureCoordinates = false;

        public bool LoadTextureCoordinates
        {
            get { return loadTextureCoordinates; }
            set { loadTextureCoordinates = value; }
        }

        /// <summary>
        /// Loads the input file.
        /// I property values remain unchanged since the last run, then the file is not reloaded, unless the Realod property is set to true.
        /// </summary>
        public override void Execute()
        {
            if ((paramsChanged) || (reload) || (output == null) || (!IsAllDataSame()))
            {
                // data is loaded only when needed
                string fn;

                if (GetInPort("Filename").Link != null)
                {
                    MveString inFileName = (MveString)GetInput("Filename");
                    fn = inFileName.Value;
                }
                else
                {
                    fn = fileName;
                }
                WriteLog("FileName", fn);

                // first pass determines the number of vertices and the number of triangles
                StreamReader sr;
                try
                {
                    sr = new StreamReader(fn);
                }
                catch (FileNotFoundException fnfe)
                {
                    throw new MveException(this, fnfe.Message);
                }

                int triangleCount = 0;
                int vertexCount = 0;
                int lineCount = 0;
                int textureCoordinateCount = 0;
                string line = sr.ReadLine();

                bool normalsFound = false;

                while (line != null)
                {
                    line = line.Trim();

                    if (line.StartsWith("v ")) vertexCount++;
                    if (line.StartsWith("vt ")) textureCoordinateCount++;
                    if (line.StartsWith("vn ")) normalsFound = true;
                    if ((line.StartsWith("f ")) || (line.StartsWith("fo ")) || line.StartsWith("f\t"))
                    {
                        triangleCount++;
                        if (line.Split(new char[]{' ','\t'}).Length == 5)
                            triangleCount++;
                    }
                    line = sr.ReadLine();
                    lineCount++;
                }

                sr.Close();

                // second pass performs the actual parsing
                sr = new StreamReader(fn);
                UniformDataArray vertices = new UniformDataArray(typeof(Point3D), vertexCount);
                UniformDataArray normals = null;
                if (normalsFound&&loadNormals)
                    normals = new UniformDataArray(typeof(Normal3D), vertexCount);
                UniformDataArray triangles = new UniformDataArray(typeof(Triangle), triangleCount);
                UniformDataArray textureIndices = null;
                textureCoordinates = null;
                if (loadTextureCoordinates)
                {
                    textureIndices = new UniformDataArray(typeof(Triangle), triangleCount);
                    textureCoordinates = new UniformDataArray(typeof(Point2D), textureCoordinateCount);
                }

                int vi = 0;
                int ti = 0;
                int ni = 0;
                int tci = 0;
                Point3D point;
                Normal3D normal;
                Triangle triangle, textureTriangle;
                NumberFormatInfo nfi = new NumberFormatInfo();
                nfi.NumberDecimalSeparator = ".";
                nfi.NumberGroupSeparator = ",";

                int linePos = 0;
                line = sr.ReadLine();
                while (line != null)
                {
                    line = line.Trim();

                    //parsing of a vertex
                    if (line.StartsWith("v "))
                    {
                        string[] coords = line.Split(new char[] { ' ' }, 4, StringSplitOptions.RemoveEmptyEntries);

                        Scalar c1 = double.Parse(coords[1], nfi);
                        Scalar c2 = double.Parse(coords[2], nfi);
                        Scalar c3 = double.Parse(coords[3], nfi);

                        if (flipZ)
                            point = new Point3D(c1, c2, -c3);
                        else
                            point = new Point3D(c1, c2, c3);
                        vertices[vi] = point;
                        vi++;
                    }

                    //parsing of a vertex normal
                    if (line.StartsWith("vn "))
                    {
                        if (loadNormals)
                        {
                            string[] coords = line.Split(new char[] { ' ' }, 4, StringSplitOptions.RemoveEmptyEntries);
                            Scalar c1, c2, c3;
                            try
                            {
                                c1 = double.Parse(coords[1], nfi);
                                c2 = double.Parse(coords[2], nfi);
                                c3 = double.Parse(coords[3], nfi);
                            }
                            catch (FormatException)
                            {
                                c1 = 0;
                                c2 = 0;
                                c3 = 1;
                            }

                            if (flipZ)
                                normal = new Normal3D(c1, c2, -c3);
                            else
                                normal = new Normal3D(c1, c2, c3);
                            if (flipNormals)
                                normal = -normal;
                            normals[ni] = normal;
                            ni++;
                        }
                    }

                    //parsing of a vertex normal
                    if (line.StartsWith("vt "))
                    {
                        if (loadTextureCoordinates)
                        {
                            string[] coords = line.Split(new char[] { ' ' }, 4, StringSplitOptions.RemoveEmptyEntries);
                            Scalar c1, c2;
                            try
                            {
                                c1 = double.Parse(coords[1], nfi);
                                c2 = double.Parse(coords[2], nfi);                                
                            }
                            catch (FormatException)
                            {
                                c1 = 0;
                                c2 = 0;                                
                            }

                            Point2D tc = new Point2D(c1, c2);
                            textureCoordinates[tci] = tc;
                            tci++;
                        }
                    }

                    // parsing of a triangle
                    if ((line.StartsWith("f ")) || (line.StartsWith("fo ")) || (line.StartsWith("f\t")))
                    {
                        string[] indices = line.Split(new char[] { ' ', '\t' }, 5, StringSplitOptions.RemoveEmptyEntries);

                        int t1 = 0;
                        int t2 = 0;
                        int t3 = 0;

                        string[] parts = indices[1].Split(new char[] {'/'}, StringSplitOptions.RemoveEmptyEntries);                        
                        int v1 = int.Parse(parts[0]) - 1;
                        if (parts.Length > 1)
                            t1 = int.Parse(parts[1]) - 1;

                        parts = indices[2].Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);                                                
                        int v2 = int.Parse(parts[0]) - 1;
                        if (parts.Length > 1)
                            t2 = int.Parse(parts[1]) - 1;

                        parts = indices[3].Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
                        int v3 = int.Parse(parts[0]) - 1;
                        if (parts.Length > 1)
                            t3 = int.Parse(parts[1]) - 1;

                        if (flipNormals)
                        {
                            triangle = new Triangle(v1, v3, v2);
                            textureTriangle = new Triangle(t1, t3, t2);
                        }
                        else
                        {
                            triangle = new Triangle(v1, v2, v3);
                            textureTriangle = new Triangle(t1, t2, t3);
                        }
                        triangles[ti] = triangle;
                        if (loadTextureCoordinates)
                            textureIndices[ti] = textureTriangle;
                        ti++;

                        if (indices.Length == 5)
                        {
                            parts = indices[4].Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
                            v2 = int.Parse(parts[0]) - 1;
                            if (parts.Length > 1)
                                t2 = int.Parse(parts[1]) - 1;

                            if (flipNormals)
                            {
                                triangle = new Triangle(v1, v2, v3);
                                textureTriangle = new Triangle(t1, t2, t3);
                            }
                            else
                            {
                                triangle = new Triangle(v1, v3, v2);
                                textureTriangle = new Triangle(t1, t3, t2);
                            }
                            triangles[ti] = triangle;
                            if (loadTextureCoordinates)
                            {
                                textureIndices[ti] = textureTriangle;
                            }
                            ti++;
                        }
                    }
                    line = sr.ReadLine();
                    linePos++;
                    if ((linePos % 1000) == 0)
                        ProgressInfo((float)linePos / lineCount);
                }

                // creating an output Mesh instance
                output = new TriangleMesh();
                output.Points = vertices;
                output.AddCells("triangles", triangles);
                if (normalsFound&&loadNormals)
                    output.AddPointAttrs("Normal3D", normals);
                if (loadTextureCoordinates)
                {
                    output.AddCellAttrs("triangles", "textureIndices", textureIndices);
                    SetOutput("TextureCoordinates", textureCoordinates);
                }
                SetOutput("TriangleMesh", output);
                paramsChanged = false;
            }
            else
            {
                // nothing's changed, we just pass current data
                if (setDataSame)
                {
                    if (loadTextureCoordinates) SetOutput("TextureCoordinates", textureCoordinates, true);
                    SetOutput("TriangleMesh", output, true);
                }
                else
                {
                    if (loadTextureCoordinates) SetOutput("TextureCoordinates", textureCoordinates, false);
                    SetOutput("TriangleMesh", output, false);
                }
            }
        }// Execute

    }//ObjLoader
}
