using System;
using System.Collections;
using System.Xml;
using System.Diagnostics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
  /// <summary>
  /// Object represent orthogonal, regular N dimensional grid.
  /// Structure consists of points only. 
  /// Point location is given implicitly.   
  /// </summary>
  [DataObject("Object represent orthogonal, regular N dimensional grid. Structure consists of points only. Point location is given implicitly.")]
  public class RegGridND : Visualization.GridND
  {
      double[] mins;

      /// <summary>
      /// Gets/sets positions of minima in all dimensions
      /// </summary>
      public double[] Mins
      {
          get { return mins; }
          set {
              if (value.Length == mins.Length)
                  mins = value;
              else throw new MveException(this, "Incorrect number of dimensions of the mins array");
          }
      }

      double[] maxs;

      /// <summary>
      /// Gets/sets positions of maxima in all dimensions
      /// </summary>
      public double[] Maxs
      {
          get { return maxs; }
          set {
              if (value.Length == maxs.Length)
                  maxs = value;
              else throw new MveException(this, "Incorrect number of dimensions of the maxs array");
          }
      }

    double[] spans;

    /// <summary>
    /// For XmlLoader use only. Use parametric constructor instead.
    /// </summary>
    public RegGridND()
    {
      UpdateModificationTime();
    } // RegGridND()


    /// <summary>
    /// Creates empty grid
    /// </summary>
    /// <param name="sizes">Structure dimensions.</param>
    /// <param name="mins">Position of minima in all dimensions.</param>
    /// <param name="maxs">Position of maxima in all dimensions.</param>
    public RegGridND(double[] mins, double[] maxs, params int[] sizes)
      : base(sizes)
    {
      if ((mins.Length != sizes.Length) || (mins.Length != maxs.Length))
      {
        throw new MveException(this, "Number of mins, maxs and sizes must be equal for regGridND constructor");
      }
      this.mins = (double[]) mins.Clone();
      this.maxs = (double[]) maxs.Clone();

      this.spans = new double[mins.Length];
      for (int i = 0; i < mins.Length; i++)
      {
        spans[i] = maxs[i] - mins[i];
        if (spans[i] < 0)
        {
          throw new MveException(this, "Max value must be greater than min value for each direction");
        }
      }

      UpdateModificationTime();
    } // RegGridND()

    /// <summary>
    /// Creates empty grid
    /// </summary>
    /// <param name="sizes">Structure dimensions.</param>
    public RegGridND(params int[] sizes)
      : base(sizes)
    {
      this.mins = new double[sizes.Length];
      this.maxs = new double[sizes.Length];
      this.spans = new double[sizes.Length];

      for (int i = 0; i < sizes.Length; i++)
      {
        mins[i] = 0;
        maxs[i] = 1;
        spans[i] = 1;
      }

      UpdateModificationTime();
    } // RegGridND()



    /// <summary>
    /// Returns x, y coordinates of point with i, j indices.
    /// </summary>
    /// <param name="coords">Point coordinates. Count must be equal to structure dimension</param>
    /// <returns>new  instance of point</returns>
    public override PointND GetPoint(params int[] coords)
    {
      if (coords.Length != this.Dimensions.Length)
      {
        throw new MveException(this, "Coordinate count and RegGridND dimension must agree.");
      }
      PointND result = new PointND(coords.Length);

      for (int i = 0; i < coords.Length; i++)
        result.Values[i] = mins[i] + spans[i] * (double) coords[i] / (Dimensions[i]);

      return (result);
    }

      /// <summary>
      /// Creates a shallow copy of the reggridND instance. The cell attributes and point attributes are referenced. Useful when adding new attributes, working with separate attributes etc.
      /// </summary>
      /// <returns>Shallow copy of the instance</returns>
      public RegGridND ShallowCopy()
      {
          RegGridND result = new RegGridND(this.mins, this.maxs, this.Dimensions);
          
          ICollection attrIds = this.GetPointAttrIds();
          foreach (string s in attrIds)
          {
              MveArray attr = this.GetPointAttrs(s);
              result.AddPointAttrs(s, attr);
          }

          attrIds = this.GetCellAttrIds();
          foreach (string s in attrIds)
          {
              MveArray attr = this.GetCellAttrs(s);
              result.AddCellAttrs(s, attr);
          }

          foreach (string s in this.UserTags.Keys)
          {
              result.UserTags.Add(s, this.UserTags[s]);
          }

          return (result);
      }
  } // class RegGridND
} // namespace