using System;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.ComponentModel;
using System.Globalization;
using System.Xml;
using System.Diagnostics;

namespace Zcu.Mve.Visualization.Sources
{
  /// <summary>
  /// This enum is primarily intended to be used within RotationTransformSourceModule
  /// to allow user to choose if entered angle was given in radians or degrees.
  /// </summary>
  public enum AngleUnits
  {
    /// <summary>
    /// Radians angle metric.
    /// </summary>
    Radians = 0,
    /// <summary>
    /// Degrees angle metric.
    /// </summary>
    Degrees = 1
  }
  
  /// <summary>
  /// This enum is primarily intended to be used within RotationTransformSourceModule
  /// to allow user to choose the rotation axis .
  /// </summary>
  public enum Axis
  {
    /// <summary>
    /// X axis.
    /// </summary>
    X = 0,
    /// <summary>
    /// Y axis.
    /// </summary>
    Y = 1,
    /// <summary>
    /// Z axis.
    /// </summary>
    Z = 2
  }

  /// <summary>
  /// Source of Transform3D rotation matrix. 
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Source of Transform3D rotation matrix.", CategoryType.Source, CategoryData.Number, IconName = "RotationTransformSource.ico")]
  [PortInfo("outTransform3D", "Resulting transformation matrix.")]
  public class RotationTransformSource : Module
  {
    private Transform3D transform3D;
    bool dataChanged;

    #region Properties
    private AngleUnits unit;
    /// <summary>
    /// Units in which the rotation angle will be provided.
    /// User can choose between degrees and radians.
    /// </summary>
    [DefaultValue(AngleUnits.Degrees)]
    [Description("Units in which the rotation angle will be provided. User can choose between degrees and radians.")]
    public AngleUnits Unit
    {
      get
      {
        return unit;
      }
      set
      {
        unit = value;

        dataChanged = true;
      }
    }

    private Axis rotationAxis;
    /// <summary>
    /// Rotation axis.
    /// </summary>
    [DefaultValue(Axis.X)]
    [Description("Rotation axis.")]
    public Axis RotationAxis
    {
      get
      {
        return rotationAxis;
      }
      set
      {
        rotationAxis = value;

        dataChanged = true;
      }
    }

    private float angle;
    /// <summary>
    /// Rotation angle.
    /// </summary>
    [DefaultValue(0)]
    [Description("Rotation angle.")]
    public float Angle
    {
      get
      {
        return angle;
      }
      set
      {
        angle = value;

        dataChanged = true;
      }
    }

    private bool showInAngle;
    /// <summary>
    /// Show/hide input port. When angle input port is shown
    /// angle value is read from input port and value set in module properties is ignored.
    /// </summary>
    [DefaultValue(false)]
    [Description("Show/hide input port. When angle input port is shown,"
                +"angle value is read from input port and value set in module properties is ignored.")]
    public bool ShowInAngle
    {
      get
      {
        return showInAngle;
      }
      set
      {
        if (value)
        {
          if (!showInAngle)
          {
            AddInPort("inAngle", typeof(Scalar), true, true);
          }
        }
        else
        {
          if (showInAngle)
          {
            RemoveInPort("inAngle");
          }
        }

        showInAngle = value;

        dataChanged = true;
      }
    }
    #endregion

    /// <summary>
    /// Default constructor, adds input and output ports.
    /// </summary>
    public RotationTransformSource()
    {
      AddOutPort("outTransform3D", typeof(Transform3D));

      showInAngle = false;
      unit = AngleUnits.Degrees;
      rotationAxis = Axis.X;
      dataChanged = true;

      transform3D = new Transform3D();
    }

    /// <summary>
    /// Creates the transform matrix. If data are unchanged, then the last result is returned immediately.
    /// </summary>
    public override void Execute()
    {
      if (!IsAllDataSame() || dataChanged)
      {
        float angleInRadians;

        if (ShowInAngle)
        {
          Scalar a = (Scalar) GetInput("inAngle");
          Angle = (float) a.Value;
        }

        if (Unit == AngleUnits.Degrees)
          angleInRadians = degreesToRadians(Angle);
        else
          angleInRadians = Angle;

        Debug.WriteLine("angleInRadians" + angleInRadians);

        switch (RotationAxis)
        {
          case Axis.X:
            transform3D.SetRotaionX(angleInRadians);
            break;
          case Axis.Y:
            transform3D.SetRotaionY(angleInRadians);
            break;
          case Axis.Z:
            transform3D.SetRotaionZ(angleInRadians);
            break;
        }

        //Console.WriteLine(transform3D);

        SetOutput("outTransform3D", transform3D, false);
        dataChanged = false;
      }
      else
        SetOutput("outTransform3D", transform3D, true);
    }

    /// <summary>
    /// Reads Xml configuration of current module.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void ReadConfig(System.Xml.XmlElement config)
    {
      foreach (XmlElement xeConfig in config.ChildNodes)
      {
        if (xeConfig.Name == "Properties")
        {
          foreach (XmlElement xeProperties in xeConfig.ChildNodes)
          {
            switch (xeProperties.Name)
            {
              case "Unit":
                switch (xeProperties.InnerText.Trim())
                {
                  case "Degrees":
                    this.unit = AngleUnits.Degrees;
                    break;
                  case "Radians":
                    this.unit = AngleUnits.Radians;
                    break;
                  default:
                    this.unit = AngleUnits.Degrees;
                    break;
                }
                break;
              case "RotationAxis":
                switch (xeProperties.InnerText.Trim())
                {
                  case "X":
                    this.rotationAxis = Axis.X;
                    break;
                  case "Y":
                    this.rotationAxis = Axis.Y;
                    break;
                  case "Z":
                    this.rotationAxis = Axis.Z;
                    break;
                  default:
                    this.rotationAxis = Axis.X;
                    break;
                }
                break;
              case "Angle":
                if (!float.TryParse(xeProperties.InnerText.Trim(), NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out this.angle))
                {
                  this.angle = 0f;
                }
                break;
              case "ShowInAngle":
                bool result;
                if (!bool.TryParse(xeProperties.InnerText.Trim(), out result))
                {
                  result = false;
                }

                ShowInAngle = result;
                break;
            }
          }
        }
      }
    }

    /// <summary>
    /// Convert value in degrees to radians.
    /// </summary>
    /// <param name="degrees">Value in degrees.</param>
    /// <returns>Returned value in radians.</returns>
    private float degreesToRadians(float degrees)
    {
      return (float) (degrees * Math.PI / 180f);
    }

    /// <summary>
    /// Creates new instance of module and copy all necessary member variables.
    /// This method is used by MapEditor in Copy/Paste process.
    /// Without overriding it returns implicit instance of current module.
    /// </summary>
    public override Module DeepCopy()
    {
      RotationTransformSource rotationTransformSource = (RotationTransformSource) base.DeepCopy();

      rotationTransformSource.angle = this.angle;
      rotationTransformSource.rotationAxis = this.rotationAxis;
      rotationTransformSource.unit = this.unit; 

      return rotationTransformSource;
    }
  }
}