using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.ComponentModel;

namespace Zcu.Mve.Visualization.Sources
{
	/// <summary>
	/// A source of a translation transform matrix.
	/// </summary>
    [ModuleInfo("Libor Va", "A source of a translation transform matrix.", CategoryType.Source, CategoryData.Number)]
	[PortInfo("X", "Scalar value of the translation in the X axis")]
	[PortInfo("Y", "Scalar value of the translation in the Y axis")]
	[PortInfo("Z", "Scalar value of the translation in the Z axis")]
	[PortInfo("Transform", "Resulting transformation matrix.")]
    class TransformTranslationSource:Module
    {
        private double x, y, z;

		private bool paramsChanged = true;

		private Transform3D result;

		/// <summary>
		/// Translation coefficient in the X direction. Only works when nothing is connected to the X input port.
		/// </summary>
		[Description("Translation coefficient in the X direction. Only works when nothing is connected to the X input port.")]
        public double X
        {
            get { return (x); }
            set { x = value; }
        }

		/// <summary>
		/// Translation coefficient in the Y direction. Only works when nothing is connected to the Y input port.
		/// </summary>
		[Description("Translation coefficient in the Y direction. Only works when nothing is connected to the Y input port.")]
        public double Y
        {
            get { return (y); }
            set { y = value; }
        }

		/// <summary>
		/// Translation coefficient in the Z direction. Only works when nothing is connected to the Z input port.
		/// </summary>
		[Description("Translation coefficient in the Z direction. Only works when nothing is connected to the Z input port.")]
		public double Z
        {
            get { return (z); }
            set { z = value; }
        }

		/// <summary>
		/// default constructor. Adds the input and output ports.
		/// </summary>
        public TransformTranslationSource()
        {
            AddInPort("X", typeof(Scalar), false);
            AddInPort("Y", typeof(Scalar), false);
            AddInPort("Z", typeof(Scalar), false);
            AddOutPort("Transform", typeof(Transform3D));
        }

		/// <summary>
		/// Constructs the translation matrix. If no inputs are changed since the last run, then the last result is returned immediately.
		/// </summary>
        public override void Execute()
        {            
            if (GetInPort("X").Link != null)
            {
                if (!IsDataSame("X"))
                {
                    x = ((Scalar)GetInput("X")).Value;
                    paramsChanged = true;
                }
            }

            if (GetInPort("Y").Link != null)
            {
                if (!IsDataSame("Y"))
                {
                    y = ((Scalar)GetInput("Y")).Value;
                    paramsChanged = true;
                }
            }

            if (GetInPort("Z").Link != null)
            {
                if (!IsDataSame("Z"))
                {
                    z = ((Scalar)GetInput("Z")).Value;
                    paramsChanged = true;
                }
            }

            if ((paramsChanged) || (result == null))
            {
                result = new Transform3D();
                result.SetTranslation((float)x, (float)y, (float)z);

                SetOutput("Transform", result);
                paramsChanged = false;
            }
            else
            {
                SetOutput("Transform", result, true);
            }
        }
    }
}
