using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    class TriangleMeshFixer:Module
    {
        public TriangleMeshFixer()
        {
            AddInPort("Input mesh", typeof(TriangleMesh));
            AddOutPort("Output mesh", typeof(TriangleMesh));
        }

        bool removeUnusedVertices = true;

        public bool RemoveUnusedVertices
        {
            get { return removeUnusedVertices; }
            set { removeUnusedVertices = value; }
        }

        bool removeNonManifoldTriangles = true;

        public bool RemoveNonManifoldTriangles
        {
          get { return removeNonManifoldTriangles; }
          set { removeNonManifoldTriangles = value; }
        }

        bool unifyOrientations = true;

        public bool UnifyOrientations
        {
            get { return unifyOrientations; }
            set { unifyOrientations = value; }
        }

        bool removeDegenerateTriangles = false;

        public bool RemoveDegenerateTriangles
        {
            get { return removeDegenerateTriangles; }
            set { removeDegenerateTriangles = value; }
        }

        bool collocatedVertices = false;

        public bool CollocatedVertices
        {
            get { return collocatedVertices; }
            set { collocatedVertices = value; }
        }

        string id;

        public override void Execute()
        {
            TriangleMesh input = (TriangleMesh)GetInput("Input mesh");
            UniformDataArray points = input.Points;
            id = input.GetCellIds(typeof(Triangle))[0];
            UniformDataArray triangles = input.GetCells(id);
            TriangleMesh result = (TriangleMesh)input.ShallowCopy();

            if (collocatedVertices)
            {
                for (int i = 0; i < (points.Count - 1); i++)
                {
                    for (int j = i + 1; j < points.Count; j++)
                    {
                        Point3D p1 = (Point3D)points[i];
                        Point3D p2 = (Point3D)points[j];
                        if (p1.Equals(p2))
                            WriteConsole("Collocated vertices " + i + " and " + j);                        
                    }
                    ProgressInfo((float)i / (points.Count - 1));
                }
            }

            if (removeUnusedVertices)
            {
                WriteDebug("Removing unused vertices");
                bool[] usages = new bool[points.Count];
                foreach (Triangle t in triangles)
                {
                    usages[t.V1] = true;
                    usages[t.V2] = true;
                    usages[t.V3] = true;
                }

                int shift = 0;
                int[] shifts = new int[points.Count];
                for (int i = 0; i < points.Count; i++)
                {
                    if (usages[i] == false)
                    {
                        shift++;
                    }
                    shifts[i] = shift;
                }
                
                if (shift > 0)
                {
                    UniformDataArray newTriangles = new UniformDataArray(typeof(Triangle), triangles.Count);
                    for (int i = 0; i < triangles.Count; i++)
                    {
                        Triangle t = (Triangle)triangles[i];
                        t.V1 -= shifts[t.V1];
                        t.V2 -= shifts[t.V2];
                        t.V3 -= shifts[t.V3];
                        newTriangles[i] = t;
                    }
                    UniformDataArray newPoints = new UniformDataArray(points.ElemType, points.Count - shift);
                    int index = 0;
                    for (int i = 0; i < points.Count; i++)
                    {
                        if (usages[i])
                        {
                            newPoints[index] = points[i];
                            index++;
                        }
                    }
                    
                    result.RemoveCells(id);
                    result.AddCells(id, newTriangles);
                    result.Points = newPoints;
                    triangles = newTriangles;
                    WriteLog("UnusedVertices", shift);
                }
            }

            if (removeNonManifoldTriangles)
            {
                // compute edge incidence
                Dictionary<Line, List<int>> edges = new Dictionary<Line, List<int>>();
                int[] usage = new int[points.Count];

                for (int i = 0; i < triangles.Count; i++)
                {
                    Triangle t = (Triangle)triangles[i];
                    usage[t.V1]++;
                    usage[t.V2]++;
                    usage[t.V3]++;
                    
                    Line l = (t.V1 < t.V2) ? (new Line(t.V1, t.V2)) : (new Line(t.V2, t.V1));
                    if (edges.ContainsKey(l)) edges[l].Add(i);
                    else
                    {
                        edges[l] = new List<int>();
                        edges[l].Add(i);
                    }
                    l = (t.V2 < t.V3) ? (new Line(t.V2, t.V3)) : (new Line(t.V3, t.V2));
                    if (edges.ContainsKey(l)) edges[l].Add(i);
                    else
                    {
                        edges[l] = new List<int>();
                        edges[l].Add(i);
                    }
                    l = (t.V1 < t.V3) ? (new Line(t.V1, t.V3)) : (new Line(t.V3, t.V1));
                    if (edges.ContainsKey(l)) edges[l].Add(i);
                    else
                    {
                        edges[l] = new List<int>();
                        edges[l].Add(i);
                    }
                }

                Dictionary<int, int> nmt = new Dictionary<int, int>();
                List<List<int>> nme = new List<List<int>>();
                
                foreach (Line line in edges.Keys)
                {
                    if (edges[line].Count <= 2)
                        continue;                    
                    else
                    {
                        string s = "Non-manifold edge " + line.V1.ToString()+" - " + line.V2.ToString() + " with incidence " + edges[line].Count.ToString() + " found, shared by triangles ";
                        nme.Add(edges[line]);
                        for (int i = 0; i < edges[line].Count; i++)
                        {
                            int t = edges[line][i];
                            s += " " + t.ToString();
                            if (nmt.ContainsKey(t))
                                nmt[t] += 1;
                            else
                                nmt[t] = 1;
                        }
                        WriteConsole(s);
                    }
                }

                List<int> remt = new List<int>();
                
                // first remove triangles with more than one non manifold edge
                foreach (int i in nmt.Keys)
                {
                    if (nmt[i] > 1)
                    {
                        remt.Add(i);
                        foreach (List<int> n in nme)
                        {
                            if (n.Contains(i))
                                n.Remove(i);
                        }
                    }
                }

                // if there are any more non manifold edges, then remove more triangles
                for (int i = 0; i < nme.Count; i++)
                {
                    if (nme[i].Count > 2)
                    {
                        for (int j = 2; j < nme[i].Count; j++)
                        {
                            int rt = nme[i][j];
                            remt.Add(rt);
                            foreach (List<int> n in nme)
                            {
                                if (n.Contains(rt))
                                    n.Remove(rt);
                            }
                        }
                    }
                }

                // construct new triangles
                UniformDataArray newTriangles = new UniformDataArray(typeof(Triangle), triangles.Count - remt.Count);
                int index = 0;
                for (int i = 0; i < triangles.Count; i++)
                {
                    if (remt.Contains(i))
                        continue;
                    newTriangles[index] = triangles[i];
                    index++;
                }
                result.RemoveCells(id);
                result.AddCells(id, newTriangles);
                triangles = newTriangles;
                WriteLog("NonManifoldTriangles", remt.Count);
            }

            if (removeDegenerateTriangles)
            {
                bool changes = true;
                int degTriangleCount = 0;
                while (changes)
                {
                    changes = false;
                    triangles = result.GetCells(id);
                    foreach (Triangle t in triangles)
                    {
                        Point3D p1 = (Point3D)result.Points[t.V1];
                        Point3D p2 = (Point3D)result.Points[t.V2];
                        Point3D p3 = (Point3D)result.Points[t.V3];
                        if (pointEquals(p1, p2))
                        {
                            if (t.V1 < t.V2)
                                contractEdge(t.V1, t.V2, result);
                            else
                                contractEdge(t.V2, t.V1, result);
                            degTriangleCount++;
                            changes = true;
                            break;
                        }

                        if (pointEquals(p2, p3))
                        {
                            if (t.V2 < t.V3)
                                contractEdge(t.V2, t.V3, result);
                            else
                                contractEdge(t.V3, t.V2, result);
                            degTriangleCount++;
                            changes = true;
                            break;
                        }
                        if (pointEquals(p1, p3))
                        {
                            if (t.V1 < t.V3)
                                contractEdge(t.V1, t.V3, result);
                            else
                                contractEdge(t.V3, t.V1, result);
                            degTriangleCount++;
                            changes = true;
                            break;
                        }
                    }                    
                }
                triangles = result.GetCells(id);
                WriteLog("DegenerateTriangles", degTriangleCount);
            }

            if (unifyOrientations)
            {
                UniformDataArray newTriangles = new UniformDataArray(typeof(Triangle), triangles.Count);

                Dictionary<Line, object> edges = new Dictionary<Line, object>();

                int processedTriangleCount = 0;
                bool[] processed = new bool[triangles.Count];

                while (processedTriangleCount < triangles.Count)
                {
                    // select one triangle and set its orientation
                    for (int i = 0; i < triangles.Count; i++)
                    {
                        if (processed[i])
                            continue;
                        processed[i] = true;
                        processedTriangleCount++;
                        Triangle it = (Triangle)triangles[i];
                        Triangle nt = new Triangle(it.V1, it.V2, it.V3);

                        newTriangles[i] = nt;

                        edges[new Line(nt.V1, nt.V2)] = null;
                        edges[new Line(nt.V2, nt.V3)] = null;
                        edges[new Line(nt.V3, nt.V1)] = null;
                        break;
                    }

                    // add more triangles, whose orientation follows from the selected one
                    bool added = true;
                    while (added)
                    {
                        added = false;
                        for (int i = 0; i < triangles.Count; i++)
                        {
                            if (processed[i])
                                continue;
                            Triangle it = (Triangle)triangles[i];
                            int cw = 0;
                            int ccw = 0;
                            Line l1cw = new Line(it.V1, it.V2);
                            Line l1ccw = new Line(it.V2, it.V1);
                            if (edges.ContainsKey(l1cw))
                                ccw++;
                            if (edges.ContainsKey(l1ccw))
                                cw++;
                            Line l2cw = new Line(it.V2, it.V3);
                            Line l2ccw = new Line(it.V3, it.V2);
                            if (edges.ContainsKey(l2cw))
                                ccw++;
                            if (edges.ContainsKey(l2ccw))
                                cw++;
                            Line l3cw = new Line(it.V3, it.V1);
                            Line l3ccw = new Line(it.V1, it.V3);
                            if (edges.ContainsKey(l3cw))
                                ccw++;
                            if (edges.ContainsKey(l3ccw))
                                cw++;

                            Triangle nt;
                            if (cw > 0)
                            {
                                added = true;
                                if (ccw > 0)
                                    WriteConsole("Mesh unification failed");
                                nt = new Triangle(it.V1, it.V2, it.V3);
                                newTriangles[i] = nt;
                                processed[i] = true;
                                processedTriangleCount++;

                                edges[new Line(nt.V1, nt.V2)] = null;
                                edges[new Line(nt.V2, nt.V3)] = null;
                                edges[new Line(nt.V3, nt.V1)] = null;
                            }
                            else if (ccw > 0)
                            {
                                added = true;
                                if (cw > 0)
                                    WriteConsole("Mesh unification failed");
                                nt = new Triangle(it.V1, it.V3, it.V2);
                                newTriangles[i] = nt;
                                processed[i] = true;
                                processedTriangleCount++;

                                edges[new Line(nt.V1, nt.V2)] = null;
                                edges[new Line(nt.V2, nt.V3)] = null;
                                edges[new Line(nt.V3, nt.V1)] = null;
                            }
                        }
                    }
                }

                result.RemoveCells(id);
                result.AddCells(id, newTriangles);
            }

            SetOutput("Output mesh", result);
        }

        private void contractEdge(int p, int p_2, TriangleMesh result)
        {
            UniformDataArray newPoints = new UniformDataArray(typeof(Point3D), result.Points.Count - 1);
            List<Triangle> newTriangles = new List<Triangle>();
            UniformDataArray triangles = result.GetCells(id);
            for (int i = 0;i<triangles.Count;i++)
            {
                Triangle t = (Triangle)triangles[i];
                if (t.V1 == p_2)
                {
                    t.V1 = p;
                    if ((t.V2 == p) || (t.V3 == p))
                        continue;
                }
                if (t.V2 == p_2)
                {
                    t.V2 = p;
                    if ((t.V1 == p) || (t.V3 == p))
                        continue;
                }
                if (t.V3 == p_2)
                {
                    t.V3 = p;
                    if ((t.V1 == p) || (t.V2 == p))
                        continue;
                }
                if (t.V1 > p_2)
                    t.V1 -= 1;
                if (t.V2 > p_2)
                    t.V2 -= 1;
                if (t.V3 > p_2)
                    t.V3 -= 1;
                newTriangles.Add(t);
            }
            for (int i = 0; i < newPoints.Count; i++)
            {
                if (i < p_2)
                    newPoints[i] = result.Points[i];
                else
                    newPoints[i] = result.Points[i + 1];
            }
            result.RemoveCells(id);
            result.AddCells(id, UniformDataArray.FromList(newTriangles));
            result.Points = newPoints;

        }

        private bool pointEquals(Point3D p1, Point3D p2)
        {
            if ((p1.X == p2.X) && (p1.Y == p2.Y) && (p1.Z == p2.Z))
                return (true);
            else
                return (false);
        }
    }
}
