using System;
using Zcu.Mve.Core;
using System.Xml;
using System.Collections;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Clas represent strip of triangels defined by atatic array of point indices.
	/// </summary>
	public struct TriangleStrip : ICell
	{
		/// <summary>
		/// Array of point indices.
		/// </summary>
		public int [] Points;

		/// <summary>
		/// Create a TriangleStrip object.
		/// </summary>
		/// <param name="points">Indices of points defined TriangleStrip.</param>
		public TriangleStrip(params int[] points)
		{
			Points = (int []) points.Clone();
		} // TriangleStrip()

		/// <summary>
		/// Return number of points of cell.
		/// </summary>
		public int PointCount
		{
			get
			{
				if (this.Points == null)
					return 0;
				else
					return this.Points.Length;
			}
		} // PointCount
		
		/// <summary>
		/// Add index of the point into TriangleStrip.
		/// </summary>
		/// <param name="index">Index of the point (in reference to DataSet).</param>
		/// <returns>Index of the point (in reference to TriangleStrip).</returns>
		public int AddPoint(int index)
		{
			int [] tmp = new int [this.PointCount + 1];
			if (this.Points != null)
				this.Points.CopyTo(tmp, 0);
			tmp[this.PointCount] = index;
			this.Points = tmp;
			return this.PointCount - 1;
		} // AddPoint()

		/// <summary>
		/// Copy elements from given ArrayList to current TriangleStrip.
		/// Original indices are deleted.
		/// </summary>
		/// <param name="points">Array of point indices.</param>
		public void FromArrayList(ArrayList points)
		{
			this.Points = (int []) points.ToArray(typeof(int));
		}

		/// <summary>
		/// Get index of the point.
		/// </summary>
		/// <param name="index">Index of the point in reference to TriangleStrip (0, 1, 2, ...).</param>
		/// <returns>Index of the point in reference to DataSet.</returns>
		public int GetPoint(int index)
		{
			return this[index];
		} // GetPoint()

		/// <summary>
		/// Get/set index of the point.
		/// </summary>
		public int this[int index]
		{
			get
			{
				if (index < this.Points.Length && index >= 0)
					return (int) this.Points[index];
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.Points.Length - 1) + ">");
			}
			set
			{
				if (index < this.Points.Length && index >= 0)
					this.Points[index] = value;
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.Points.Length - 1) + ">");
			}
		} // this[int]

		/// <summary>
		/// Read data from XML file.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			string [] points = xmlTextReader.ReadString().Trim().Split(' ');
			ArrayList pointList = new ArrayList();

			for (int i = 0; i < points.Length; i++)
				pointList.Add(Int32.Parse(points[i]));

			this.Points = (int []) pointList.ToArray(typeof(int));
		} // ReadData()

		/// <summary>
		/// Write data to XML file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			string pom = "";
			for (int i = 0; i < this.Points.Length; i++)
				pom += this.Points[i].ToString() + " ";
			pom = pom.Substring(0, pom.Length - 1);	 // odstraneni prebytecne mezery na konci
			xmlTextWriter.WriteString(pom);
		} // WriteData()

		/// <summary>
		/// Deep copy of curent object
		/// </summary>
		/// <returns>Deep copy (same data, different memory place).</returns>
		public IDataObject DeepCopy()
		{
			return new TriangleStrip(this.Points);
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// TriangleStrip with negative index is considered inconsistent.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			if (this.Points == null)
			{
				Console.WriteLine("Error! " + this.GetType().Name + ": contains no indices.");
				return false;
			}

			for (int i = 0; i < this.Points.Length; i++)
				if (this[i] < 0) 
				{
					Console.WriteLine("Error! " + this.GetType().Name + ": point No. " + i + " has negative index.");
					return false;
				}
			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			string pom = "TriangleStrip (" + this.PointCount + " vertices): {";
			if (this.PointCount == 0)
				return pom += "}";
			foreach (int index in this.Points)
				pom += index + ", ";
			return pom.Substring(0, pom.Length - 2) + "}";
		} // ToString()	

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

	} // class TriangleStrip
} // namespace
