﻿using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module for cutting by polyhedron
    /// Polyhedron is in UnstrGrid type and contains attribute with triangles, where one triangle is one cutting plane.
    /// Value -1 is inside, value 0 is on the surface, value 1 is outside the polyhedron.
    /// </summary>    
    [ModuleInfo("Roman Soukal", "Module for cutting by polyhedron", CategoryType.Filter, CategoryData.Volume, IconName = "VolumeCut3.ico")]
    [PortInfo("Input Grid", "Input RegGridND data")]
    [PortInfo("Input Polyhedron", "Input UnstrGrid")]
    [PortInfo("Output Grid", "Output RegGridND data")]
    public class ConvexPolyhedronCutting : Module
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public ConvexPolyhedronCutting()
        {
            AddInPort("Input Grid", typeof(RegGridND));
            AddInPort("Input Polyhedron", typeof(UnstrGrid));
            AddOutPort("Output Grid", typeof(RegGridND));
        }

        #region Variables
        private RegGridND data = null;
        private string pointAttributesName = "Values";
        private string cuttingAttributesName = "CuttingValues";
        private string planeAttributesName = "Planes";
        private bool paramsChanged = false;
        #endregion

        /// <summary>
        /// Overrided
        /// </summary> 
        public override void Execute()
        {
            if (!IsAllDataSame() || paramsChanged)
            {
                RegGridND old = (RegGridND)this.GetInput("Input Grid");
                data = old.ShallowCopy();
                //IDataObject val = null;
                //if (!data.UserTags.TryGetValue("Min", out val))
                //{
                //    old.UserTags.TryGetValue("Min", out val);
                //    data.UserTags.Add("Min", val);
                //}
                //if (!data.UserTags.TryGetValue("Max", out val))
                //{
                //    old.UserTags.TryGetValue("Max", out val);
                //    data.UserTags.Add("Max", val);
                //}
                //if (!data.UserTags.TryGetValue("Distortion", out val))
                //{
                //    old.UserTags.TryGetValue("Distortion", out val);
                //    data.UserTags.Add("Distortion", val);
                //}

                UnstrGrid polyhedron = (UnstrGrid)this.GetInput("Input Polyhedron");
                MveArray values = data.GetPointAttrs(this.PointAttributeName);
                MveArray newValues = new MveArray(typeof(Scalar), values.Lengths);
                int[] sizes = values.Lengths;

                UniformDataArray triangles = polyhedron.GetCells(planeAttributesName);
                Point3D point1, point2, point3;
                double[,] planeFunctions = new double[triangles.Count, 4];
                for (int i = 0; i < triangles.Count; i++)
                {
                    Triangle triangle = (Triangle)triangles.Array.GetValue(i);
                    point1 = (Point3D)polyhedron.Points.Array.GetValue(triangle.V1);
                    point2 = (Point3D)polyhedron.Points.Array.GetValue(triangle.V2);
                    point3 = (Point3D)polyhedron.Points.Array.GetValue(triangle.V3);
                    double detB1C2 = point2.X * point3.Y - point2.Y * point3.X;
                    double detB1C3 = point2.X * point3.Z - point2.Z * point3.X;
                    double detB1C4 = point2.X - point3.X;
                    double detB2C3 = point2.Y * point3.Z - point2.Z * point3.Y;
                    double detB2C4 = point2.Y - point3.Y;
                    double detB3C4 = point2.Z - point3.Z;
                    planeFunctions[i, 0] = -(point1.Y * detB3C4 - point1.Z * detB2C4 + detB2C3);
                    planeFunctions[i, 1] = point1.X * detB3C4 - point1.Z * detB1C4 + detB1C3;
                    planeFunctions[i, 2] = -(point1.X * detB2C4 - point1.Y * detB1C4 + detB1C2);
                    planeFunctions[i, 3] = point1.X * detB2C3 - point1.Y * detB1C3 + point1.Z * detB1C2;
                }
                for (int i = 0; i < sizes[0]; i++)
                {
                    this.ProgressInfo(((float)i / sizes[0]));
                    for (int j = 0; j < sizes[1]; j++)
                        for (int k = 0; k < sizes[2]; k++)
                        {
                            int[] coords = { i, j, k };
                            PointND pValue = data.GetPoint(coords);
                            double value = -1;
                            for (int l = 0; l < triangles.Count; l++)
                            {
                                double orient = Math.Sign(pValue.Values[0] * planeFunctions[l, 0] + pValue.Values[1] * planeFunctions[l, 1] + pValue.Values[2] * planeFunctions[l, 2] + planeFunctions[l, 3]); 
                                if (orient > value)
                                {
                                    value = orient;
                                    if (value == 1)
                                        break;
                                }
                            }
                            /*foreach (Triangle triangle in triangles.Array)
                            {
                                point1 = (Point3D)polyhedron.Points.Array.GetValue(triangle.V1);
                                point2 = (Point3D)polyhedron.Points.Array.GetValue(triangle.V2);
                                point3 = (Point3D)polyhedron.Points.Array.GetValue(triangle.V3);
                                double a11 = point1.X - pValue.Values[0];
                                double a12 = point1.Y - pValue.Values[1];
                                double a13 = point1.Z - pValue.Values[2];
                                double a21 = point2.X - pValue.Values[0];
                                double a22 = point2.Y - pValue.Values[1];
                                double a23 = point2.Z - pValue.Values[2];
                                double a31 = point3.X - pValue.Values[0];
                                double a32 = point3.Y - pValue.Values[1];
                                double a33 = point3.Z - pValue.Values[2];
                                double distance = Math.Sign(a11 * (a22 * a33 - a23 * a32) - a12 * (a21 * a33 - a31 * a23) + a13
                                        * (a21 * a32 - a22 * a31));
                                if (distance > value)
                                {
                                    value = distance;
                                    if (value == 1)
                                        break;
                                }
                            }*/
                            newValues.Array.SetValue(new Scalar(value), coords);
                        }
                }
                data.AddPointAttrs(this.cuttingAttributesName, newValues);
                this.SetOutput("Output Grid", this.data, false);
                paramsChanged = false;
            }
            else
            {
                this.SetOutput("Output Grid", data, true);
            }
        }

        #region Properties

        /// <summary>
        /// Set name of RegGridND point attribute. Default set to Values.
        /// </summary>
        [Description("Set name of RegGridND point attribute. Default set to \"Values\"")]
        public string PointAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.pointAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.pointAttributesName; }
        }

        /// <summary>
        /// Set name of RegGridND attribute with cutting information. Default set to CuttingValues.
        /// </summary>
        [Description("Set name of RegGridND attribute with cutting information. Default set to\"CuttingValues\"")]
        public string CuttingAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.cuttingAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.cuttingAttributesName; }
        }

        /// <summary>
        /// Set name of RegGridND attribute with each plane information. Default set to Planes.
        /// </summary>
        [Description("Set name of RegGridND attribute with each plane information. Default set to\"Planes\"")]
        public string PlaneAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.planeAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.planeAttributesName; }
        }

        #endregion
    }
}
