using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;


namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Abstract class for marching tetrahedra methods.
    /// </summary>
    public abstract class MarchingTetrahedra : IsoSurfaceExtractor
    {
        private int hashIndex = 0;
        private Line edge;
        private Vertex newVertex = new Vertex();
        private Dictionary<Line, int> dict;

        /// <summary> Constructor. Creates input and output ports. </summary>
        public MarchingTetrahedra() : base() { } // of MarchingCubes        

        /// <summary> Executes module functionality. </summary>
        public override void Execute()
        {
            ProgressInfo(0f);
            if (!IsAllDataSame() || paramsChanged) // if I have to compute again 
            {
                ComputeIsoSurface();
                SetOutput("Out", mesh);
            }
            else
            {
                SetOutput("Out", mesh, true);
                IsoSurfaceExtractor.GCClean();
            }
        } // of Execute

        /// <summary> This method initializes variables and call cube extracting method. </summary>
        protected override void ComputeIsoSurface() { base.ComputeIsoSurface(); }

        /// <summary> Initializes variables.  </summary>
        protected override void Initialize()
        {
            dict = null;
            base.Initialize();
            indexList = new int[3];
            dict = new Dictionary<Line, int>();
        }

        /// <summary> This method makes cubes from values in two neghbouring slices. The convention of cubes vertices numbering is shown below. 
        /// After each cube is created, it is commited to the next step - checked whether it is intersected by searched iso-surface and eventually appropriate triangles are generated.
        ///   4 ------ 5
        ///  /|       /|
        /// 7 ------ 6 |
        /// | |      | |
        /// | 0 -----| 1
        /// |/       |/ 
        /// 3 ------ 2      </summary> 
        protected override void DecomposeVolumeIntoCubes() { base.DecomposeVolumeIntoCubes(); } // of DecomposeVolumeIntoCubes        

        /// <summary> Checks whether a tetrahedron is intersected by searched iso-surface. </summary>
        protected virtual void CalculateTetraIndex(Vertex p0, Vertex p1, Vertex p2, Vertex p3)
        {
            int tetraIndex = 0;

            // create index which vertices of the cube are under iso-surface
            if (p0.value < isoValue) tetraIndex |= 1;
            if (p1.value < isoValue) tetraIndex |= 2;
            if (p2.value < isoValue) tetraIndex |= 4;
            if (p3.value < isoValue) tetraIndex |= 8;

            // stop if there is no intersection 
            if ((tetraIndex == 0x00) || (tetraIndex == 0x0F))
                return;

            if (valueDataType == ValueDataType.Integer)
            {
                // calculate normals for all tetrahedras vertices
                p0.normal = CalcNormalI(p0, dataI, size);
                p1.normal = CalcNormalI(p1, dataI, size);
                p2.normal = CalcNormalI(p2, dataI, size);
                p3.normal = CalcNormalI(p3, dataI, size);
            }
            else if (valueDataType == ValueDataType.Scalar)
            {
                // calculate normals for all tetrahedras vertices
                p0.normal = CalcNormalD(p0, dataD, size);
                p1.normal = CalcNormalD(p1, dataD, size);
                p2.normal = CalcNormalD(p2, dataD, size);
                p3.normal = CalcNormalD(p3, dataD, size);
            }

            GenerateTrianglesForTetrahedron(tetraIndex, p0, p1, p2, p3);
        } // CalculateTetraIndexI      

        /// <summary> Find edges where we found intersection and calculate positions of triangles vertices. </summary>
        /// <param name="tetraIndex">Rate index for each tetrahedron.</param>
        /// <param name="p0">First vertex of the tetrahedron.</param>
        /// <param name="p1">Second vertex of the tetrahedron.</param>
        /// <param name="p2">Third vertex of the tetrahedron.</param>
        /// <param name="p3">Fourth vertex of the tetrahedron.</param>
        private void GenerateTrianglesForTetrahedron(int tetraIndex, Vertex p0, Vertex p1, Vertex p2, Vertex p3)
        {
            Triangle triangle;
            switch (tetraIndex)
            {
                case 0x00:
                case 0x0F:
                    break;
                case 0x0E:
                case 0x01:
                    triangle = new Triangle(GetPointIndex(p0, p1), GetPointIndex(p0, p2), GetPointIndex(p0, p3));
                    indices.Add(triangle);
                    break;
                case 0x0D:
                case 0x02:
                    triangle = new Triangle(GetPointIndex(p1, p0), GetPointIndex(p1, p3), GetPointIndex(p1, p2));
                    indices.Add(triangle);
                    break;
                case 0x0C:
                case 0x03:
                    triangle = new Triangle(GetPointIndex(p0, p3), GetPointIndex(p0, p2), GetPointIndex(p1, p3));
                    indices.Add(triangle);

                    triangle = new Triangle(triangle.V3, GetPointIndex(p1, p2), triangle.V2);
                    indices.Add(triangle);
                    break;
                case 0x0B:
                case 0x04:
                    triangle = new Triangle(GetPointIndex(p2, p0), GetPointIndex(p2, p1), GetPointIndex(p2, p3));
                    indices.Add(triangle);
                    break;
                case 0x0A:
                case 0x05:
                    triangle = new Triangle(GetPointIndex(p0, p1), GetPointIndex(p2, p3), GetPointIndex(p0, p3));
                    indices.Add(triangle);

                    triangle = new Triangle(triangle.V1, GetPointIndex(p1, p2), triangle.V2);
                    indices.Add(triangle);
                    break;
                case 0x09:
                case 0x06:
                    triangle = new Triangle(GetPointIndex(p0, p1), GetPointIndex(p1, p3), GetPointIndex(p2, p3));
                    indices.Add(triangle);

                    triangle = new Triangle(triangle.V1, GetPointIndex(p0, p2), triangle.V3);
                    indices.Add(triangle);
                    break;
                case 0x07:
                case 0x08:
                    triangle = new Triangle(GetPointIndex(p3, p0), GetPointIndex(p3, p2), GetPointIndex(p3, p1));
                    indices.Add(triangle);
                    break;
            }
        } // GenerateTrianglesForTetrahedron

        /// <summary> Gets index of the point form array of already computed points or adds new point into this array. </summary>
        /// <params name="index">Index of intersected edge.</params>
        /// <params name="a">First vertex of the edge.</params>
        /// <params name="b">Second vertex of the edge.</params>
        private int GetPointIndex(Vertex a, Vertex b)
        {
            int value;
            // construct edge
            if (a.value < b.value)
            {
                edge.V1 = a.index;
                edge.V2 = b.index;
            }
            else
            {
                edge.V1 = b.index;
                edge.V2 = a.index;
            }

            // check if on this edge was already computed vertex (on each edge can be max. 1 vertex)
            if (dict.TryGetValue(edge, out value))
                return value;
            else
            {
                newVertex = Interpolate(a, b, isoValue, inverseDistortion, linearInterpolation);
                dict.Add(edge, hashIndex);
                pts.Add(newVertex);
                hashIndex++;
                return hashIndex - 1;
            }
        } // of GetPointIndex

        /// <summary> Creates from computed data output data </summary>
        protected override void CreateOutput()
        {
            base.CreateOutput();

            hashIndex = 0;
            dict = null;
            IsoSurfaceExtractor.GCClean();

            ProgressInfo(1f);
        } // of CreateOutput

    } // of MarchingTetrahedra class
} // of namespace Zcu.Gsvd.Extractions